#!/usr/bin/python3

# Look for files tmp/piece_*.sh and run them "threads" at a time.
# Each piece creates a wait* file, and deletes this wait* file when it is done.
# As the 'wait' files disappear, we run more of the tmp/piece_*.sh scripts, until we are done.

import sys
import os
import glob
import time
import multiprocessing

version = 1.0

if len( sys.argv ) == 2 and sys.argv[1] == "-v" :
    print( "fracture_feeder.py v%1.1f" % version )
    sys.exit()


threads = 2 * multiprocessing.cpu_count() - 2

def check_for_crash():
    if os.path.exists( "fullstop.txt" ) :
        print( "\n\nabort\n\n" )
        os.system( "killconv" )
        sys.exit()

file = glob.glob( "tmp/piece_*" )

numleft = len( file )
total = numleft
nhash = 60
p = "%"

print( "\nThere are %d pieces.\n" % numleft );

first = True

while numleft > 0 :                     # numleft is the number of files we have left to launch
    wait = glob.glob( "tmp/wait*" )
    launchnum = threads - len(wait)
    if launchnum > numleft : launchnum = numleft

    if numleft > 0 and launchnum == 0 :
        time.sleep( 2 )
        wait = glob.glob( "tmp/wait*" )
        launchnum = threads - len(wait)
        if launchnum > numleft : launchnum = numleft
    
    #if launchnum > 0 : print( "\nLaunching %d tasks...\n" % launchnum )

    numwaiting = numleft + len( wait )
    statusbar  = float(total - numwaiting) / total
    if statusbar < 0 : statusbar = 0
        
    done = nhash * statusbar
    perc = 100   * statusbar
    print( "   %3d%s    [" % (perc, p), end="" )
    for i in range( round(done) ) : print( "#", end="" )

    if perc < 99.5 :
        for i in range( nhash - round(done) + 1 ) : print( " ", end="" )
        print( "] ", end="\015" )
    else:
        print( "] ", end="\015" )

    sys.stdout.flush()

    for i in range( launchnum ) :
        check_for_crash()
        os.system( "mv %s tmp/done/%s" % (file[i], file[i]) )    # first move it to tmp/done/tmp/piece_*.sh
        os.system( "tmp/done/%s &" % file[i] )                   # then run it. that way it's not in the glob.

    file = glob.glob( "tmp/piece_*" )   
    numleft = len( file ) 

time.sleep(2)

# now wait for the wait files to disappear

wait = glob.glob( "tmp/wait*" )

while len( wait ) > 0 :
    numwaiting = len( wait )
    check_for_crash()
    statusbar  = float(total - numwaiting) / total
    if statusbar < 0 : statusbar = 0
        
    done = nhash * statusbar
    perc = 100   * statusbar
    print( "   %3d%s    [" % (perc, p), end="" )
    for i in range( round(done) ) : print( "#", end="" )
    if perc < 99.5 :
        for i in range( nhash - round(done) + 1 ) : print( " ", end="" )
        print( "] ", end="\015" )
    else:
        print( "] ", end="\015" )

    sys.stdout.flush()
    time.sleep( 2 )
    wait = glob.glob( "tmp/wait*" )


done = nhash
perc = 100
print( "   %3d%s    [" % (perc, p), end="" )
for i in range( round(done) ) : print( "#", end="" )
print( "] ", end="\015" )
print( "\n\n...\n" )
sys.stdout.flush()





