#!/bin/bash

VERSION="2.5"
MAXGDSFILESIZE=10000  # MB
THISISEBPG="no"       # is this computer actually running an ebpg?
                      # if so, we can do extra checking

echo
echo "Freebeam v$VERSION"
echo -n "                 "
check_extents -v
echo -n "                 "
check_subfield_size -v
echo -n "                 "
convexify -v
echo -n "                 "
darn -v
echo -n "                 "
flatten -v
echo -n "                 "
flatten_with_layout -v 2&>1 > /tmp/fwl_version
cat /tmp/fwl_version
rm /tmp/fwl_version
echo -n "                 "
fracture_feeder.py -v
echo -n "                 "
freeview -v
echo -n "                 "
gpfout -v
echo -n "                 "
gpoly2gds -v
echo -n "                 "
gpoly2txt -v
echo -n "                 "
kill.py -v
echo -n "                 "
remove_overlaps -v
echo -n "                 "
slice -v
echo -n "                 "
grid_snap -v
echo -n "                 "
splitter -v
echo -n "                 "
subfield_size -v
echo

if [ "$1" == "-v" ]; then exit; fi;
    
rm -f fullstop.txt try_again.txt


if [ $THISISEBPG == "yes" ] &&  ! command -v pg ; then
    echo 
    echo "This system does not seem to be an EBPG."
    echo "If this copy of freebeam is on a separate server,"
    echo "you should set THISISEBPG to 'no' at the top"
    echo "of freebeam.sh"
    echo
    exit
fi


if [ "$infile" == "" ]; then
    echo
    echo "You must define a number of symbols before starting freebeam."
    echo "Look at the example /public/freebeam_example.sh"
    echo
    exit
fi


if [ "$outfile" == "" ]; then outfile="$infile" ; fi
outfile=${outfile//' '/'_'}


# set some reasonable defaults:

pss=$lsb    # pss is no longer set by the user

if [ "$llx"     == "" ];  then llx=0       ; fi 
if [ "$lly"     == "" ];  then lly=0       ; fi
if [ "$urx"     == "" ];  then urx=0       ; fi
if [ "$ury"     == "" ];  then ury=0       ; fi
if [ "$crllx"   == "" ];  then crllx=0     ; fi 
if [ "$crlly"   == "" ];  then crlly=0     ; fi
if [ "$crurx"   == "" ];  then crurx=0     ; fi
if [ "$crury"   == "" ];  then crury=0     ; fi
if [ "$xsub"    == "" ];  then xsub=0      ; fi
if [ "$ysub"    == "" ];  then ysub=0      ; fi
if [ "$voltage" == "" ];  then voltage=100 ; fi
if [ "$split"   == "" ];  then split="yes" ; fi
if [ "$nsplits" == "" ];  then nsplits=0   ; fi
if [ "$purge"   == "" ];  then purge="yes" ; fi
if [ "$quiet"   == "" ];  then quiet="y"   ; fi
if [ "$angle"   == "" ];  then angle=0     ; fi
if [ "$array"   == "" ];  then array="no"  ; fi
if [ "$alreadyflat"   == "" ]; then alreadyflat="no"   ; fi
if [ "$layoutflatten" == "" ]; then layoutflatten="no" ; fi
if [ "$polygons"      == "" ]; then polygons="yes" ; fi


if [ "$path" != "" ]; then
    if [ ! -f $path ]; then
        echo
        echo "ERROR: unable to find path file $path"
        echo "Freebeam stopped."
        echo
        exit
    fi
fi
        

# these parameters are required:

if [ "$layer"   == "" ]; then echo "Layers have not been specified."                     ; exit ; fi
if [ "$topcell" == "" ]; then echo "Top cell has not been specified."                    ; exit ; fi
if [ "$bss"     == "" ]; then echo "Beam step size (bss) has not been specified."        ; exit ; fi
if [ "$lsb"     == "" ]; then echo "Least significant bit (lsb) has not been specified." ; exit ; fi
if [ "$xblock"  == "" ]; then echo "X block size (xblock) has not been specified."       ; exit ; fi
if [ "$yblock"  == "" ]; then echo "Y block size (yblock)  has not been specified."      ; exit ; fi


if [ $THISISEBPG == "yes" ] && [ `pg get opt_upg` != 1 ] && [ "$polygons" == "yes" ]; then
    echo
    echo "GPF polygons will work only on UPG pattern generators."
    echo "It looks like your system has the older GPG pattern generator."
    echo "Set polygons='no' to generate the older-style GPF trapezia."
    echo "Freebeam stopped."
    exit
fi

                            
if [ -f tmp ]; then 
    echo
    echo "Freebeam needs to use the name tmp for a directory."
    echo "You already have a file called tmp."
    read -p "Do you want to delete tmp now? (y/n) > " ans
    if [ ${ans:-"c"} == "y" ]; then
	rm -rf tmp
    else
	echo
	echo "Ok, tmp is still there."
	echo "Freebeam stopped."
	echo
	exit
    fi
    echo 
fi

if [ -d tmp ]; then
    echo
    echo "Freebeam needs to use the name tmp for a directory."
    echo "You already have a directory named tmp."
    echo 
    if [ "$1" == "-y" ]; then
        echo "wait ..."
	rm -rf tmp
        echo "And now you do not."
    else
        read -p "Do you want to delete tmp now? (y/n) > " ans
        if [ ${ans:-"c"} == "y" ]; then
            echo "wait ..."
	    rm -rf tmp
            echo "ok, tmp has been deleted"
        else
	    echo
	    echo "Ok, tmp is still there."
	    echo "Freebeam stopped."
	    echo
	    exit
        fi
    fi
    echo 
fi

rm "$outfile.gpf"
mkdir tmp
mkdir tmp/done
mkdir tmp/done/tmp
echo $VERSION > tmp/version

if [ "$remove_overlaps" == "yes" ] || [ "$remove_overlaps" == "y" ] || [ "$remove_overlaps" == "Y" ]; then
    removeoverlaps="true"
else
    removeoverlaps="false"
fi

if [ "$polygons" == "no" ] || [ "$polygons" == "n" ]; then
    traps="traps"
else
    traps="notraps"
fi

if [ "$dosefile" != "" ] && [ "$removeoverlaps" == "true" ]; then
    echo
    echo "ERROR"
    echo
    echo "You have enabled overlap removal, but you have"
    echo "also specified a table (a file) of dose assignments."
    echo "Overlap removal deletes all layer and datatype values,"
    echo "and so you cannot use dose assignments and overlap"
    echo "removal at the same time. Set 'remove_overlaps' to \"no\""
    echo "or set 'dosefile' to \"\"".
    echo
    echo
    touch fullstop.txt
    exit
fi

if [ "$dosefile" != "" ]; then
    if [ -f $dosefile ]; then
        echo
        echo "Using dose file $dosefile"
        echo
    else
        echo 
        echo "Unable to find dose file $dosefile"
        echo "Please try again."
        echo
        touch fullstop.txt
        exit
    fi
fi


if [ "$nsplits" == "" ]; then nsplits=0 ; fi

if [ "$layer" == ""  ]; then layer="-1" ; fi
if [ "$layer" == "*" ]; then layer="-1" ; fi

if [ "$xsub" == "" ]; then xsub=0 ; fi
if [ "$ysub" == "" ]; then ysub=0 ; fi

if [ "$xsub" == "0" ]; then
    subfield_size $xblock $bss $lsb $pss
    xsub=`subfield_size $xblock $bss $lsb $pss`
else
    check_subfield_size $xsub $bss $lsb $pss
fi

if [ -f fullstop.txt ]; then exit ; fi

if [ "$ysub" == "0" ]; then
    subfield_size $yblock $bss $lsb $pss
    ysub=`subfield_size $yblock $bss $lsb $pss`
else
    check_subfield_size $ysub $bss $lsb $pss
fi

if [ -f fullstop.txt ]; then exit ; fi

#if [ "$removeoverlaps" == "true" ]; then
#    check_subfield_size $xsub $bss $lsb $pss $xblock
#    check_subfield_size $ysub $bss $lsb $pss $yblock
#fi

if [ -f fullstop.txt ]; then exit ; fi


# lsb=`dc -e "6 k $lsb 1000 / p"`
lsb=`python3 -c "print( $lsb / 1000.0 )"`

# are the pattern extents on the bss grid?
# They should be, even when using auto-extents.
# check_extents here will not check that the
# pattern exent is less than the given values.
# We will do this after flattening.

if [ "$crllx" == "0" ] && [ "$crlly" == "0" ] && [ "$crurx" == "0" ] && [ "$crury" == "0" ]; then
    crop="no"
else
    crop="yes"
fi

check_extents $bss $llx $lly $urx $ury

if [ "$crop" == "yes" ]; then
    check_extents $bss $crllx $crlly $crurx $crury
fi

if [ -f fullstop.txt ]; then exit ; fi

echo
echo "---------------------------------------------------------------"
echo "                                                       FREEBEAM"
echo "---------------------------------------------------------------"
echo
echo "----------------------------------------------- flatten"
echo
echo
echo "use the \"Kill Freebeam\" button to stop"
echo

if [ "$layoutflatten" != "no" ]; then
    flatten_with_layout "$infile.$suffix" "$topcell" "tmp/$infile""_flat.$suffix"
    if [ -f fullstop.txt ]; then exit ; fi
    infile="tmp/$infile""_flat"
    fi

if [ "$angle" == "" ]; then angle=0 ; fi

# flatten with angle=0 now. Rotate later, after removing holes with 'darn'
# if the gds file is already flat, the user can set alreadyflat=yes

if [ "$angle" == "0" ] && [ "$alreadyflat" == "yes" ]; then 
    echo "cp $infile.$suffix tmp/flat.gds"
    cp "$infile.$suffix" tmp/flat.gds
else
    flatten "$infile.$suffix" "$topcell" -yesbox -nodump -silent -layer $layer 0 tmp/flat.gds

    if [ "$crop" == "no" ]; then
        check_extents $bss $llx $lly $urx $ury
    else
        check_extents $bss $llx $lly $urx $ury
        echo $crllx $crlly $crurx $crury > tmp/pattern_extents.dat
        check_extents $bss $crllx $crlly $crurx $crury
    fi
fi

if [ -f fullstop.txt ]; then exit ; fi


# Do we need to use dataunit_fix when removing overlaps with Layout?
# We might get lots of weird self-crossing polygons. We don't know yet.
#
# if [ "$removeoverlaps" == "true" ]; then 
#     dataunit_fix tmp/flat.gds tmp/flat1nm.gds
#     mv tmp/flat1nm.gds tmp/flat.gds
# fi

# Apply overlap removal to the whole, uncut pattern.

if [ "$removeoverlaps" == "true" ]; then
    echo
    echo "----------------------------------------------- overlap removal"
    echo "This might take a long time."
    echo
    remove_overlaps tmp/flat.gds tmp/noverlap.gds
    mv tmp/noverlap.gds tmp/flat.gds
    layer="0"
fi

echo
echo "----------------------------------------------- splitter"
echo
echo
echo "use the \"Kill Freebeam\" button to stop"
echo
echo

if [ "$split" == "no"  ]; then split="nosplit" ; fi
if [ "$split" == "yes" ]; then split="yesplit" ; fi


if [ $angle != 0 ]; then
    if [ "$llx" == "0" ] && [ "$lly" == "0" ] && [ "$urx" == "0" ] && [ "$ury" == "0" ]; then
        echo
        echo "We need explicit data extents when rotating a pattern."
        echo "Please specify the ll and ur corners when rotating."
        echo
        touch fullstop.txt
        exit
    else
        rotate_extents $bss $llx $lly $urx $ury $angle tmp/rotate_extents.dat
        read llx lly urx ury cx cy < tmp/rotate_extents.dat
        cp tmp/rotate_extents.dat tmp/extents.dat
        rotate_extents $bss $crllx $crlly $crurx $crury $angle tmp/rotate_extents.dat
        read crllx crlly crurx crury crcx crcy < tmp/rotate_extents.dat
    fi
fi

if [ "$array" == ""    ]; then array="noarray"  ; fi
if [ "$array" == "no"  ]; then array="noarray"  ; fi
if [ "$array" == "yes" ]; then array="yesarray" ; fi


splitter tmp/flat.gds $pss $layer $xblock $yblock $llx $lly $urx $ury $crllx $crlly $crurx $crury $xsub $ysub $angle $split $lsb $nsplits $array $bss $traps


echo
echo "----------------------------------------------- fracturing"
echo
echo
echo "use the \"Kill Freebeam\" button to stop"
echo
echo

fracture_feeder.py


if [ -f fullstop.txt ]; then echo "failure in fracture_feeder.py" ; exit ; fi

if [ -f tmp/sliced.gds ]; then rm -f tmp/sliced.gds ; fi

if [ ! -f tmp/sliced_part_0.gds ]; then
    echo "ERROR: no sliced gds files were produced in the fracture step" 
    touch fullstop.txt
    exit
fi

cat tmp/sliced_* >> tmp/sliced.gds

echo
echo
echo "----------------------------------------------- gpfout"
echo

if [ -f fullstop.txt ]; then exit ; fi

size=`du -m tmp/sliced.gds | cut -f1`
if [ $size -gt $MAXGDSFILESIZE ]; then
    touch fullstop.txt
    echo
    echo "ERROR: output file size exceeds limit of $MAXGDSFILESIZE MB."
    echo "       You should chop up the pattern, or choose a different layer."
    echo
    exit 
fi

if [ ! -f $dosefile ];    then dosefile="nodose" ; fi
if [ "$dosefile" == "" ]; then dosefile="nodose" ; fi
if [ ! -f $path ];        then path="nopath" ; fi
if [ "$path" == "" ];     then path="nopath" ; fi


gpfout tmp/sliced.gds "$outfile" $llx $lly $urx $ury $xblock $yblock $xsub $ysub $bss $lsb $voltage $dosefile $path $traps


if [ -f fullstop.txt ]; then exit ; fi


echo
echo "----------------------------------------------- gpfout done"
echo
date

# If you want stats on Freebeam use, you could do this:
# echo `date` $infile.$suffix >> /home/pg/bin/freebeam/freebeam_use.log

if [ "$purge" != "no" ]; then
    rm -rf tmp
fi


if command -v pg > /dev/null ; then
    echo
    echo "Checking for updates..."
    echo "                         use [ctrl] C if this takes too long."
    echo 
    rm -f latest_freebeam_version
    wget https://github.com/ebpgfunpak/freebeam/blob/main/version?raw=true -O latest_freebeam_version
    read latest_version < latest_freebeam_version
    rm -f latest_freebeam_version

    if [ "$latest_version" == "" ]; then
        echo
        echo "Unable to connect to github.com to check for updates."
        echo
        exit
    fi

    if [ `python3 -c "print( $VERSION < $latest_version )"` == "True" ]; then
        echo
        echo "***************************************************"
        echo " A new version of Freebeam is available.           "
        echo " Use update_freebeam.sh to install the new version."
        echo "***************************************************"
        echo
    fi
fi


echo 
echo "You might want to cview or freeview $outfile.gpf"
echo







