#!/bin/bash

infile="yale"  # your input file name, without the suffix
suffix="gds"
                  # use topcell="?" to use the first cell in the file - handy for flat files
topcell="3x"
                  # Use layer="5" to convert layer 5 only.
                  # Layers will be merged together when listed like layer="0,1,2,4,8" 
                  # (with commas and without spaces). To convert all layers, use 
                  # layer="-1" or "*" or ""

outfile="$infile" # gpf file name. 

layer="*"

bss=5             # beam step size in nm, must be a multiple of the lsb of course

lsb=1             # Main field LSB in nm, Usually 1 nm on a wide-field ebpg
                  # but it might be 0.5nm on a small-field machine,
                  # unless you are using an unusually small field for 
                  # a high-accuracy PBG or waveguide pattern.
                  # Note that setting the placement accuracy, pss, is no longer an option.
                  # pss = lsb always. There is no reason to change this.

xblock=500        # block size, um = size of the writing area, which is
yblock=500        # anything smaller than the maximum field size. 
                  # Some people say "field size" instead of "block size,"
                  # which is confusing. Technically the "field" is the maximum deflection.

xsub=0            # subfield size, um. Set them to zero to let freebeam
ysub=0            # calculate the best subfield for you.


llx=0             # Pattern extents, um. Set them to zero for automatic extents.
lly=0             # But watch out: automated chip extents are a bad idea in general.
urx=500           # Pattern extents MUST include the entire pattern.
ury=500           # Use the cropping window (below) to select a smaller part of the pattern. 
                  
crllx=0           # cropping window. set to zero for no cropping.
crlly=0
crurx=0
crury=0

dosefile=""       # (optional) file of <layer> <datatype> <relative-dose>

                  # Removing overlaps can be very time consuming.
                  # It's smarter to design your patterns without overlaps.
                  # Removing overlaps just proves you are bad designer.
                  # Also note that you cannot have dose assignments
                  # and overlap removal at the same time.
                
remove_overlaps="no"        # use Layout to remove overlaps

polygons="no"      # Polygons produce smoother patterns than trapezia.
                   # For UPG systems with Firebird, choose polygons="yes".
                   # For older GPG systems or newer machines without Firebird, choose "no".

# parameters you probably will not change:

voltage=100        # voltage in kV
split="yes"        # split into multiple files/threads? probably yes.
nsplits=0          # chop the input into this many pieces. Set to 0 for auto.
purge="yes"        # delete temporary files when finished? probably yes.
alreadyflat="no"   # if a big pattern is already flat, this saves time.
quiet="y"          # 'y' means don't ask about deleting the tmp directory
angle=0            # Do you want to rotate the pattern? Well that's strange.
array="no"         # array="yes" saves space, especially if you have some very large pads
                   #     but you must disable arrays if you want to sort fields or shapes 
                   #     with ybrsort later.
path=""            # write the gpf blocks (fields) in the order indicated by this path.
                   #     This is the same as using "gpfsort -path" but it's a lot faster.
                   #     The path file can have the same format as that used by gpfsort.
layoutflatten="no" # flatten first with Layout - useful for eliminating weird gds
                   #     features like mirrored and rotated arrays and similar
                   #     goofy crap that no sane person would use.



# Things you definitely do not want to change:

                  # ---- now run freebeam... note the log file

. freebeam.sh -$quiet   2>&1 | tee $outfile.log

if [ -f fullstop.txt ]; then exit ; fi

echo
echo "----------------------------------------------- cview $outfile.gpf"

if [ "$polygons" == "yes" ]; then
    echo
    echo "Use freeview to display the writing order."
    echo
fi

if [ -f $outfile.gpf ]; then
    if command -v cview > /dev/null ; then
        cview $outfile.gpf
    elif command -v freeview > /dev/null ; then
        freeview $outfile.gpf &
        sleep 2
        echo
        echo ">>>>>>"
        echo ">>>>>> Click on [show all shapes] "
        echo ">>>>>>"
        echo
    fi
fi    

# if cview is buggy, here is an alternative:
# rm -f $outfile-gpf.gds
# gpoly2gds $outfile.gpf $outfile-gpf.gds $llx $lly $urx $ury
# layout $outfile-gpf.gds 
  








