// WE SHOULD ADD SUPPORT FOR MSF MULTIPLES
// WHICH CAN CREATE EFFECTIVELY SMALLER RELATIVE DOSES

#include <stdio.h>
#include "gpf_pub.h"
#include "polyq.h"
#include "libgds.h"
#include "libgpfree.h"

#define MAXVERTEX    512

#define TRUE  1
#define FALSE 0
#define BOOL  unsigned char
#define ERROR_LAYER 255
#define NUMDOSELAYERS 254
#define SUBDACBITS   (14)

// v1.4  Include msf in output

// v2.0  Accepts trapezia in addition to polygons

// v3.1  The Beams library is no longer required
//       gpf_read_header()  replaced by gpfree_read_header()
//       gpf_read_indexf()  replaced by gpfree_read_index()
//       gpf_read_union()   replaced by gpfree_read_union()
//       Note that we still need the Beams header files.

// v3.2  Do not include beams.h
//       New version of gpf.h

// v3.3  Replaced gpf.h with gpf_pub.h


#define VERSION 3.3

// It might be better to use the pattern extent
// from the gpf header, assuming the header is ok.
// For now we set the extents by the pattern.

int total=0;

void 
usage()
    {
    printf( "\ngpoly2txt v%1.1f\n", VERSION );
    printf( "\nUsage: gpoly2txt infile.gpf outfile [lower_left x y (um)] [upper-right x y (um)] \n" );
    printf( "\nConvert gpf polygons and trapezia to text. Output coordinates are microns. \n");
    printf( "You can view the file outfile.txt with freeview.\n" );
    printf( "Include the lower-left and upper-right points if you want to recreate \n" );
    printf( "the original design coordinates. Clocks and field boundaries are also included.\n" );
    printf( "Invalid polygons and field boundaries will appear on layer %d.\n", ERROR_LAYER );
    printf( "The text output will be in outfile.txt. Note that the argument outfile does not end in .txt\n" );
    printf( "Pattern and block extents are written to outfile_extents.txt\n\n" );
    exit(0);
    }

//----------------------------------------------------------------------------------------

void
flag( struct gds_itemtype *g, dxy ll, dxy ur, dxy offset, char *message, double dbu_um )
    {
    static dxy d, size;
    d.x = g->x[0] * dbu_um; 
    d.y = g->y[0] * dbu_um;
    size.x = ur.x - ll.x;
    size.y = ur.y - ll.y;
    printf( "\n ERROR: %s near %1.3lf, %1.3lf um  size %1.4lf, %1.4lf um  offset %1.3lf, %1.3lf um", 
	    message, d.x, d.y, size.x, size.y, offset.x, offset.y );
    g->layer = ERROR_LAYER;
    }

//----------------------------------------------------------------------------------------

double
clock2dose( gpf_uint_t freq_msw, gpf_uint_t freq_lsw )
    {
    static unsigned long horloge;
    horloge = (freq_msw << 16) + freq_lsw;
    return( 2147483648.0 / horloge );
    }

double
oneclock2dose( gpf_uint_t freq )
    {
    return( 2147483648.0 / freq );
    }

int
dose2layer( double dose, double mindose, double maxdose )
    {
    return( (int) round( NUMDOSELAYERS * (dose - mindose) / (maxdose - mindose) ) );
    }


void
write_clock( FILE *out, gpf_uint_t freq_msw, gpf_uint_t freq_lsw )
    {
    static double dose;
    dose = clock2dose( freq_msw, freq_lsw );
    if ( dose <= 0.0 ) bail( "illegal clock frequency or something weird in write_clock" );
    fprintf( out, "CLOCK   %f\n", 1.0/dose );
    }

//-------------------------------------------------------------------------------------------------------------


BOOL
samepoint( xy a, xy b )
    {
    return( (a.x == b.x) && (a.y == b.y) );   // or maybe it should be < bss away?  
    }


void
remove_duplicate_vertices( polygon *p )
    {
    static int n, nm1, i;
    static xy   vi, vi1, v0, vnm1;
    static BOOL escape;

    n   = p->len;
    nm1 = n - 1;
    i   = 0;

    while ( i < nm1 )
	{
	vi  = vertex( p, i );
	vi1 = vertex( p, i+1 );
	escape = FALSE;
	while ( ! escape && (i < nm1) && samepoint( vi, vi1 ) )
	    {
	    escape = delete_vertex( p, i+1 );
	    n = p->len;
	    nm1 = n - 1;
	    vi  = vertex( p, i );
	    if ( i == nm1 )
		vi1 = vertex( p, 0 );
	    else 
		vi1 = vertex( p, i+1 );
	    }
	i += 1;
	}

    if ( n < 3 ) return;

    v0   = vertex( p, 0 );

    vnm1 = vertex( p, nm1 );

    if ( samepoint( v0, vnm1 ) ) p->len -= 1;

    }

//----------------------------------------------------------------------------------------

void
write_txt_polygon( FILE *out, struct gds_itemtype *poly, polygon *raw, dxy subfs, dxy offset, double dbu_um  )
    {
    static xy  ur, ll;
    static int i, oldlayer;
    static dxy dur, dll;

    if ( ! raw->active ) return;
    
    oldlayer = poly->layer;
    ur.x = poly->x[0];
    ur.y = poly->y[0];
    ll.x = ur.x;
    ll.y = ur.y;

    for ( i=1; i < poly->n; i++ )
	{
	if ( poly->x[i] < ll.x ) ll.x = poly->x[i];
        if ( poly->y[i] < ll.y ) ll.y = poly->y[i];
	if ( poly->x[i] > ur.x ) ur.x = poly->x[i];
        if ( poly->y[i] > ur.y ) ur.y = poly->y[i];
	}
    
    dll.x = ll.x * dbu_um;
    dll.y = ll.y * dbu_um;
    dur.x = ur.x * dbu_um;
    dur.y = ur.y * dbu_um;
    
    // The polygon might become degenerate after all this floating point math,
    // so we should check the raw numbers instead.

    if ( raw->ll.x == raw->ur.x )   flag( poly, dll, dur, offset, "polygon is degenerate in x", dbu_um );
    if ( raw->ll.y == raw->ur.y )   flag( poly, dll, dur, offset, "polygon is degenerate in y", dbu_um );

    if ( offset.x < 0.0 ) flag( poly, dll, dur, offset, "offset x < 0", dbu_um );
    if ( offset.y < 0.0 ) flag( poly, dll, dur, offset, "offset y < 0", dbu_um );

    //if ( (dur.x - dll.x + offset.x) > subfs.x ) flag( poly, dll, dur, offset, "polygon exceeds x subfield size", dbu_um );
    //if ( (dur.y - dll.y + offset.y) > subfs.y ) flag( poly, dll, dur, offset, "polygon exceeds y subfield size", dbu_um );

    fprintf( out, "NVERT   %d\n", poly->n );
    
    for ( i=0; i < poly->n; i++ )
        {
        fprintf( out, "        %f %f\n", dbu_um * poly->x[i], dbu_um * poly->y[i] );
        }

    total++;
    }

//----------------------------------------------------------------------------------------

void
write_block_outline( FILE *out, dxy stage, dxy block, dxy shift )
    {
    static dxy blockll, blockur;
    
    blockll.x = shift.x + stage.x - block.x / 2.0;
    blockll.y = shift.y + stage.y - block.y / 2.0;
    blockur.x = shift.x + stage.x + block.x / 2.0;
    blockur.y = shift.y + stage.y + block.y / 2.0;
        
    fprintf( out, "BLOCK   %lf %lf %lf %lf\n", blockll.x, blockll.y, blockur.x, blockur.y );
    //fprintf( extents,   "%lf %lf %lf %lf\n", blockll.x, blockll.y, blockur.x, blockur.y );
    }

//----------------------------------------------------------------------------------------

void
write_txt_array( FILE *out, struct gds_itemtype *poly, polygon *raw, dxy d1, dxy d2, int n1, int n2, double dbu_um, dxy subfs, dxy offset )
    {
    static int i, j, v, inst, max;
    static xy  id1, id2;
    static struct gds_itemtype *p;
    static BOOL init=FALSE;

    if ( ! init )
	{
	init = TRUE;
        p = (struct gds_itemtype *) malloc( sizeof( struct gds_itemtype ) );
        p->x = (int *) malloc( MAXVERTEX * sizeof( int ) ); 
        p->y = (int *) malloc( MAXVERTEX * sizeof( int ) );
	}

    if ( ! raw->active ) return;
    
    p->layer = poly->layer;
    p->dt    = poly->dt;

    id1.x = (int) round( d1.x / dbu_um );
    id1.y = (int) round( d1.y / dbu_um );
    id2.x = (int) round( d2.x / dbu_um );
    id2.y = (int) round( d2.y / dbu_um );

    p->n = poly->n;

    // Diagonal arrays are so stupid
    // and have never been used, ever, 
    // since the beginning of time.
    // And yet here we are:

    max = (n1 + 1) * (n2 + 1);
    i = 0;
    j = 0;
    for ( inst=0; inst < max; inst++ )
	{
        for ( v=0; v < p->n; v++ )
	    {
	    p->x[v] = poly->x[v] + id1.x * i + id2.x * j ;
	    p->y[v] = poly->y[v] + id1.y * i + id2.y * j ;
	    }
        write_txt_polygon( out, p, raw, subfs, offset, dbu_um );	
        i++;
        if ( i > n1 ) { i=0; j++; }
        if ( j > n2 ) j = 0;
        }

    }    // end write_txt_array

//----------------------------------------------------------------------------------------

void
trap2poly( struct gds_itemtype *gdspoly, polygon *raw, gpf_primitive_activate_t *activate, int base, int height,
           int shiftlt, int shiftrb, dxy main, double mainmid, dxy offset, dxy stage, dxy shift, dxy sfr )
    {
    static long count;
    static gpf_union_t activaterecord;
    static xy   a, b, c, d;
    static dxy  a_um, b_um, c_um, d_um;
    static int  subits;
    static float fsubits, fheight, fbase, fshiftrb, fshiftlt;
    static BOOL init=FALSE;

    // sfr == dbu_um

    if ( ! init )
        {
        init = TRUE;
        subits = pow( 2, SUBDACBITS );  // 2^14 = 16384
        fsubits = subits;
        }

    /*
    activaterecord.activate = *activate;
    printf( "\n\nBase       %d\n", base );
    printf( "Height     %d\n", height );
    printf( "Shift lt   %d\n", shiftlt );
    printf( "Shift rb   %d\n", shiftrb );
    gpf_print_union( stdout, &activaterecord, &count );
    */

    fheight = height;
    fbase   = base;
    if ( activate->rbNeg ) shiftrb = -shiftrb;
    if ( activate->ltNeg ) shiftlt = -shiftlt;
    fshiftlt = shiftlt;
    fshiftrb = shiftrb;

    if ( activate->OrientationY )
        {
        a.x = round( offset.x );
        a.y = round( offset.y );

        if ( activate->rb90 )
            {
            b.x = a.x + base;
            b.y = a.y;
            }
        else
            {
            b.x = a.x + base;
            b.y = a.y + round( -fshiftrb * fbase / fsubits );
            }

        if ( activate->lt90 )
            {
            c.x = b.x;
            c.y = a.y + height;
            }
        else
            {
            c.x = b.x;
            c.y = a.y + height + round( -fshiftlt * fbase / fsubits );
            }

        d.x = a.x;
        d.y = a.y + height;
        
        }
    else
        {
        a.x = round( offset.x );
        a.y = round( offset.y );
        b.x = a.x + base;
        b.y = a.y;

        if ( activate->rb90 )
            {
            c.x = b.x;
            c.y = b.y + height;
            }
        else
            {
            c.x = b.x + round( fshiftrb * fheight / fsubits );
            c.y = b.y + height;
            }

        if ( activate->lt90 )
            {
            d.x = a.x;
            d.y = a.y + height;
            }
        else
            {
            d.x = a.x + round( fshiftlt * fheight / fsubits );
            d.y = a.y + height;
            }
        }

    zero_polygon( raw );                
    append_polygon_xy( raw, a );
    append_polygon_xy( raw, b );
    append_polygon_xy( raw, c );
    append_polygon_xy( raw, d );

    remove_duplicate_vertices( raw );

    if ( raw->len < 3 )
        {
        raw->active = FALSE;
        return;
        }

    a = vertex( raw, 0 );   
    b = vertex( raw, 1 );
    c = vertex( raw, 2 );
    
    a_um.x = a.x * sfr.x + main.x - mainmid + offset.x + stage.x + shift.x;   // um
    a_um.y = a.y * sfr.y + main.y - mainmid + offset.y + stage.y + shift.y;
    b_um.x = b.x * sfr.x + main.x - mainmid + offset.x + stage.x + shift.x;  
    b_um.y = b.y * sfr.y + main.y - mainmid + offset.y + stage.y + shift.y;
    c_um.x = c.x * sfr.x + main.x - mainmid + offset.x + stage.x + shift.x;  
    c_um.y = c.y * sfr.y + main.y - mainmid + offset.y + stage.y + shift.y;

    gdspoly->x[0] = (int) round( a_um.x / sfr.x );   // sfr == dbu_um
    gdspoly->y[0] = (int) round( a_um.y / sfr.y );
    gdspoly->x[1] = (int) round( b_um.x / sfr.x ); 
    gdspoly->y[1] = (int) round( b_um.y / sfr.y );
    gdspoly->x[2] = (int) round( c_um.x / sfr.x ); 
    gdspoly->y[2] = (int) round( c_um.y / sfr.y );

    gdspoly->n = 3;
    
    if ( raw->len == 4 )
        {
        d = vertex( raw, 3 );
        d_um.x = d.x * sfr.x + main.x - mainmid + offset.x + stage.x + shift.x;   // um
        d_um.y = d.y * sfr.y + main.y - mainmid + offset.y + stage.y + shift.y;
        gdspoly->x[3] = (int) round( d_um.x / sfr.x ); 
        gdspoly->y[3] = (int) round( d_um.y / sfr.y );
        gdspoly->n = 4;
        }

    }

//----------------------------------------------------------------------------------------

void
set_pattern_extents( struct gds_itemtype *gdspoly, dxy *ur, dxy *ll )
    {
    static int n, i;
    static BOOL init=FALSE;

    n = gdspoly->n;

    if ( n == 0 ) return;
    
    if ( ! init )
        {
        init = TRUE;
        ll->x = gdspoly->x[0];
        ll->y = gdspoly->y[0];
        ur->x = ll->x;
        ur->y = ll->y;
        }

    for ( i=0; i < n; i++ )
        {
        if ( gdspoly->x[i] < ll->x ) ll->x = gdspoly->x[i];
        if ( gdspoly->y[i] < ll->y ) ll->y = gdspoly->y[i];
        if ( gdspoly->x[i] > ur->x ) ur->x = gdspoly->x[i];
        if ( gdspoly->y[i] > ur->y ) ur->y = gdspoly->y[i];
        }
    
    }

//----------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------

int main( int argc, char *argv[] )
    {

    FILE *in, *out, *extents;

    gpf_header_t             *header;
    gpf_header_user_t        *user;
    gpf_ulong_t              *pos;
    gpf_uint_t                freq_msw, freq_lsw;
    gpf_union_t              *glob;
    gpf_indexf_t             *field_index;
    gpf_primitive_24_t       *thing;
    gpf_primitive_16_t       *thing2;
    gpf_primitive_coord_t    *vertex;
    gpf_primitive_activate_t *activate;
    struct gds_itemtype      *gdspoly;
    polygon                  *raw;
    double                    mainmid, dbu_um, dbu_uu, mindose, maxdose, dose, shape_smsf;
    float                     dbu_um_f, dbu_uu_f;
    long                      shapecnt;
    dxy                       sfr, mfr, stage, main, offset, vert, d1, d2;
    dxy                       block, msf, smsf, subfs, ll, ur, shift;
    dxy                       patsize, patll, patur;
    int                       n1, n2, blocknum;
    int                       i, v, p, numprims, main_bits, sub_bits, numv;
    int                       base, height, shiftlt, shiftrb, prev_base, prev_height, prev_shiftlt, prev_shiftrb, nprims;
    char                     *infile, *outfile, outfiletxt[256], extentxt[256];
    char                     *converter;
    BOOL                      first, end_field;

    
    if ( argc == 2 && strcmp( argv[1], "-v" ) == 0 )
        {
        printf( "gpoly2txt v%1.1f\n", VERSION );
        exit(0);
        }

    
    if ( argc < 3 ) usage();

    infile  = argv[1];
    outfile = argv[2];

    if ( strlen(infile)  > 256 ) bail( "Invalid input file name." );
    if ( strlen(outfile) > 256 ) bail( "Invalid output file name." );

    if ( ! strstr( infile, ".gpf" ) ) bail( "Input file name should end in .gpf" );
    if (   strstr( outfile, "."   ) ) bail( "Output file name should not have a suffix." );
         
    ll.x = 0;
    ll.y = 0;
    ur.x = 0;
    ur.y = 0;
    shift.x = 0;
    shift.y = 0;

    if ( argc == 7 )
	{
	sscanf( argv[3], "%lf", &ll.x );
	sscanf( argv[4], "%lf", &ll.y );
	sscanf( argv[5], "%lf", &ur.x );
	sscanf( argv[6], "%lf", &ur.y );
        shift.x = (ur.x + ll.x) / 2.0;
        shift.y = (ur.y + ll.y) / 2.0;
	}

    raw = initialize_polygon( MAXVERTEX, 0, 0, 0 );   // used to store unprocessed ints, so we can look for degens

    in = gpfree_file_open( infile, "r" );
    if ( ! in ) bail( "unable to open gpf file" );
    
    sprintf( outfiletxt, "%s.txt", outfile );
    sprintf( extentxt,   "%s_extents.txt", outfile );
    
    printf( "\nInput  gpf file: %s\n",   infile );
    printf(   "Output txt file: %s\n",   outfiletxt );
    printf(   "Extents file:    %s\n\n", extentxt );

    header   = (gpf_header_t *)       malloc( sizeof( gpf_header_t ) );
    user     = (gpf_header_user_t *)  malloc( sizeof( gpf_header_user_t ) );
    pos      = (gpf_ulong_t *)        malloc( sizeof( gpf_ulong_t ) );
    glob     = (gpf_union_t *)        malloc( sizeof( gpf_union_t ) );

    thing    = (gpf_primitive_24_t *)       glob;
    thing2   = (gpf_primitive_16_t *)       glob;
    vertex   = (gpf_primitive_coord_t *)    glob;
    activate = (gpf_primitive_activate_t *) glob;

    gpfree_read_header( in, header, user );

    //gpf_print_header( stdout, header, user );

    sfr.x     = header->SubFieldResolutionX;      // usually 0.0005 um
    sfr.y     = header->SubFieldResolutionY;
    mfr.x     = header->MainFieldResolutionX;     // usually 0.001 um
    mfr.y     = header->MainFieldResolutionY;
    main_bits = header->NrMainFieldBits;
    sub_bits  = header->NrSubFieldBits;
    msf.x     = header->ResolutionX;              // eg 5
    msf.y     = header->ResolutionY;           
    smsf.x    = header->BeamStepSizeX;            // eg 10
    smsf.y    = header->BeamStepSizeY;
    block.x   = header->MainFieldSizeX * mfr.x * msf.x;   // um
    block.y   = header->MainFieldSizeY * mfr.y * msf.y;   // um
    subfs.x   = header->SubFieldSizeX  * sfr.x * smsf.x;  // um
    subfs.y   = header->SubFieldSizeY  * sfr.y * smsf.y;  // um
    patsize.x = header->PatternSizeX   * mfr.x * msf.x;   // um
    patsize.y = header->PatternSizeY   * mfr.x * msf.x;   // um
    converter = user->ConverterName;

    if ( (strstr( converter, "BEAMER" ) != 0) || (strstr( converter, "beamer" ) != 0) )
        {
        printf( "\n\nConverter: %s\n", converter );
        printf( "\nThis gpf file was created by Beamer. The input file\n" );
        printf( "can contain only polygons, usually created with Freebeam.\n\n" );
        bail( "The gpf file contains non-polygon shapes, since it was created by Beamer." );
        }
    
    dbu_um = sfr.x;     // might be 0.5 nm or even less
    dbu_uu = dbu_um;    // this makes the cad display um units 

    out = fopen( outfiletxt, "w" );
    if ( ! out ) bail( "unable to open output txt file" );


    // WRITE TXT HEADER

    if ( msf.x  != msf.y  ) bail( "MSF in X is not the same as MSF in Y ? I do not understand." );
    if ( smsf.x != smsf.y ) bail( "subfield MSF in X is not the same as subfield MSF in Y ? I do not understand." );
    if ( mfr.x  != mfr.y  ) bail( "mainfield resolution X is not equal to that in Y ? I do not understand." );
    if ( sfr.x  != sfr.y  ) bail( "subfield resolution X is not equal to that in Y ? I do not understand." );
    
    fprintf( out, "MSF     %1.0lf\n", msf.x );
    fprintf( out, "SMSF    %1.0lf\n", smsf.x );
    fprintf( out, "MFR     %lf\n", mfr.x );
    fprintf( out, "SFR     %lf\n", sfr.x );
    fprintf( out, "PATSIZE %lf %lf\n", patsize.x, patsize.y );

             
    //mdbu = msf.x  * mfr.x;   // one database unit
    //sdbu = smsf.x * sfr.x;   // equals this many microns

    gdspoly = (struct gds_itemtype *) malloc( sizeof( struct gds_itemtype ) );
    gdspoly->x = (int *) malloc( MAXVERTEX * sizeof( int ) ); 
    gdspoly->y = (int *) malloc( MAXVERTEX * sizeof( int ) );
    gdspoly->layer = 0;
    gdspoly->dt    = 0;

    mindose = 1.0;
    maxdose = 1.0;

    mainmid   = mfr.x * pow( 2, main_bits - 1 );          // usually 524.288 um

    printf( "\nMain field size (um):       %1.3lf, %1.3lf", block.x, block.y );
    printf( "\nMax subfield size (um):     %1.3lf, %1.3lf", subfs.x, subfs.y );
    printf( "\nMain field resolution (um): %1.5lf", mfr.x );
    printf( "\nSub  field resolution (um): %1.5lf", sfr.x );
    printf( "\nMSF:                        %1.0lf", msf.x );
    printf( "\nNumber of main DAC bits:    %d", main_bits );

    if ( argc == 7 )
	{
	printf( "\nLower left  (um):           %1.3lf, %1.3lf",   ll.x, ll.y );
	printf( "\nUpper right (um):           %1.3lf, %1.3lf\n", ur.x, ur.y );
	}
    else
	printf( "\n" );

    field_index = (gpf_indexf_t *) malloc( header->NrMainFields * sizeof(gpf_indexf_t));

    freq_msw = 0x8000;
    freq_lsw = 0;
    first = TRUE;
    blocknum = 0; 
    printf( "\n" );

    for (i = 0; i < header->NrMainFields; i++)
        {
        printf( "\015 Field %d  ", i ); fflush( stdout );

        gpfree_read_indexf( in, header->IndexOffset, i, &field_index[i] );

        //gpf_print_indexf( stdout, i, &field_index[i] );

        stage.x = field_index[i].FieldToPatternOffsetX;  // um
        stage.y = field_index[i].FieldToPatternOffsetY;
    
        write_block_outline( out, stage, block, shift );

        *pos = field_index[i].StartData;
        numprims = (field_index[i].EndData - field_index[i].StartData) / 4;
        shapecnt = 0;

        main.x   = 0;
        main.y   = 0;
        offset.x = 0;
        offset.y = 0;
        d1.x     = 0;
        d1.y     = 0;
        d2.x     = 0;
        d2.y     = 0;
        n1       = 0;
        n2       = 0;
        prev_base    = 0;
        prev_height  = 0;
        prev_shiftlt = 0;
        prev_shiftrb = 0;
        base         = 0;
        height       = 0;
        shiftlt      = 0;
        shiftrb      = 0;

        end_field = FALSE;

        for ( p=0; p < numprims && ! end_field; p++ )
	    {
            gpfree_read_union( in, pos, glob );   // glob is aka thing, thing2, vertex, and activate

            if      ( thing->IC == GPF_PRIM_MAIN_X_LSW )         main.x     = thing->val * mfr.x;  // um 
            else if ( thing->IC == GPF_PRIM_MAIN_Y_LSW )         main.y     = thing->val * mfr.y;
            else if ( thing->IC == GPF_PRIM_OFFSET_X_LSW )       offset.x   = thing->val * sfr.x;
            else if ( thing->IC == GPF_PRIM_OFFSET_Y_LSW )       offset.y   = thing->val * sfr.y;
            else if ( thing->IC == GPF_PRIM_MAINREPVECT1_X_LSW ) d1.x       = thing->val * mfr.x; 
            else if ( thing->IC == GPF_PRIM_MAINREPVECT2_X_LSW ) d2.x       = thing->val * mfr.x;
            else if ( thing->IC == GPF_PRIM_MAINREPVECT1_Y_LSW ) d1.y       = thing->val * mfr.x; 
            else if ( thing->IC == GPF_PRIM_MAINREPVECT2_Y_LSW ) d2.y       = thing->val * mfr.x;
            else if ( thing->IC == GPF_PRIM_MAINREPVECT1_CNT )   n1         = thing->val;
            else if ( thing->IC == GPF_PRIM_MAINREPVECT2_CNT )   n2         = thing->val;
            else if ( thing->IC == GPF_PRIM_MOVESTAGE )          blocknum   = blocknum + 1;
            else if ( thing->IC == GPF_PRIM_MSF )
                {
                shape_smsf = thing->val;
                fprintf( out, "SMSF    %1.0lf\n", shape_smsf );
                }
            else if ( thing->IC == GPF_PRIM_FREQUENCY_MSW )      freq_msw   = thing->val; 
            else if ( thing->IC == GPF_PRIM_FREQUENCY_LSW )      freq_lsw   = thing->val; 
            else if ( thing->IC == GPF_PRIM_POLYGON ) 
		{
		numv = thing2->val2;                         // number of vertices
                if ( numv > MAXVERTEX ) bail( "too many vertices" );
                gdspoly->n = numv;

                
                write_clock( out, freq_msw, freq_lsw );

                zero_polygon( raw );                         // the unprocessed ints

                for ( v=0; v < numv; v++ )
		    {
		    gpfree_read_union( in, pos, glob );
		    append_polygon( raw, vertex->x, vertex->y );
		    vert.x = vertex->x * sfr.x;                     // remember, vertex is glob
		    vert.y = vertex->y * sfr.y;
		    vert.x = vert.x + main.x - mainmid + offset.x + stage.x + shift.x;   // um
                    vert.y = vert.y + main.y - mainmid + offset.y + stage.y + shift.y;
                    gdspoly->x[v] = (int) round( vert.x / dbu_um );
                    gdspoly->y[v] = (int) round( vert.y / dbu_um );
		    }
                
		gpfree_read_union( in, pos, glob );

		if ( thing->IC != GPF_PRIM_ACTIVATE ) bail( "bad gpf file: missing activate line" );
                if ( activate->ShapeID != GPF_SHAPE_POLYGON ) bail( "bad gpf file: this polygon has the wrong shape ID" );

                set_pattern_extents( gdspoly, &patur, &patll );  // hmm. what about arrays?
                    
                if ( ! (total % 5000) ) printf( "\015 Field %d      polygon %d  ", i, total ); fflush( stdout );

                if ( activate->MainRepVect1 || activate->MainRepVect2 ) 
		    {
		    if ( ! activate->MainRepVect1 ) n1 = 0;
		    if ( ! activate->MainRepVect2 ) n2 = 0;
                    write_txt_array( out, gdspoly, raw, d1, d2, n1, n2, dbu_um, subfs, offset );

                    // printf( "\nshape  =  %1.3f, %1.3f\n", gdspoly->x[0]/1000.0, gdspoly->y[0]/1000.0 );
                    // printf( "d1     =  %1.3f, %1.3f\n", d1.x, d1.y );
                    // printf( "d2     =  %1.3f, %1.3f\n", d2.x, d2.y );
                    // printf( "offset =  %1.3f, %1.3f\n", offset.x, offset.y );
                    // printf( "n      =  %d, %d\n", n1, n2 );
                    
		    }
		else
		    {
		    write_txt_polygon( out, gdspoly, raw, subfs, offset, dbu_um );
		    }

                p = p + numv + 1;
		}
            else if ( thing->IC == GPF_PRIM_BASE_LSW     ||
                      thing->IC == GPF_PRIM_HEIGHT_LSW   ||
                      thing->IC == GPF_PRIM_SHIFT_LT_LSW ||
                      thing->IC == GPF_PRIM_SHIFT_RB_LSW    )
                {
                // Read in this trap including the activate line,
                // then convert the trap into a polygon.
                // Remember the previous values, in case they
                // are set by implication.
                
                base    = prev_base;
                height  = prev_height;
                shiftlt = prev_shiftlt;
                shiftrb = prev_shiftrb;
                nprims  = 0;
                
                while ( thing->IC != GPF_PRIM_ACTIVATE )
                    {
                    if      ( thing->IC == GPF_PRIM_BASE_LSW     ) { base    = thing->val; prev_base    = base;    p++;  }
                    else if ( thing->IC == GPF_PRIM_HEIGHT_LSW   ) { height  = thing->val; prev_height  = height;  p++;  }
                    else if ( thing->IC == GPF_PRIM_SHIFT_LT_LSW ) { shiftlt = thing->val; prev_shiftlt = shiftlt; p++;  }
                    else if ( thing->IC == GPF_PRIM_SHIFT_RB_LSW ) { shiftrb = thing->val; prev_shiftrb = shiftrb; p++;  }
                    else
                        {
                        printf( "\n------------ Unknown primitive inside trapezoid description: %d\n", thing->IC );
                        //gpf_print_union( stdout, glob, &shapecnt );
                        bail( "Trapezoid contains unexpected component." );
                        }
                        
                    gpfree_read_union( in, pos, glob );
                    
                    nprims++;
                    if ( nprims > 20 ) bail( "gpoly2txt is confused about a trap description - too many primitives" );
                    }
                
                if ( activate->ShapeID != GPF_SHAPE_TRAPEZIUM ) bail( "gpoly2txt expected a trap activate line" );
                
                zero_polygon( raw );
                write_clock( out, freq_msw, freq_lsw );

                trap2poly( gdspoly, raw,  activate, base,   height, shiftlt,
                           shiftrb, main, mainmid,  offset, stage,  shift,  sfr );

                set_pattern_extents( gdspoly, &patur, &patll );

                if ( activate->MainRepVect1 || activate->MainRepVect2 ) 
		    {
		    if ( ! activate->MainRepVect1 ) n1 = 0;
		    if ( ! activate->MainRepVect2 ) n2 = 0;
                    write_txt_array( out, gdspoly, raw, d1, d2, n1, n2, dbu_um, subfs, offset );
                    
                    // printf( "\nshape  =  %1.3f, %1.3f\n", gdspoly->x[0]/1000.0, gdspoly->y[0]/1000.0 );
                    // printf( "d1     =  %1.3f, %1.3f\n", d1.x, d1.y );
                    // printf( "d2     =  %1.3f, %1.3f\n", d2.x, d2.y );
                    // printf( "offset =  %1.3f, %1.3f\n", offset.x, offset.y );
                    // printf( "n      =  %d, %d\n", n1, n2 );
                    
		    }
		else
		    {
		    write_txt_polygon( out, gdspoly, raw, subfs, offset, dbu_um );
		    }

                }
            else if ( thing->IC == 0 )
                {
                end_field = TRUE;
                }
            else
		{
		printf( "\n------------ Unknown primitive: %d\n", thing->IC );
		//gpf_print_union( stdout, glob, &shapecnt );
		}

	    }

        }

    float dx, dy, dm, cx, cy;

    // would it be better to use the pattern extent from the gpf header?

    //  printf( "\n\npattern extents:  %9.3lf  %9.3lf  %9.3lf  %9.3lf gds dbu units\n\n", patll.x, patll.y, patur.x, patur.y );

    
    dx = patur.x - patll.x;
    dy = patur.y - patll.y;
    dm = dx;
    if ( dy > dm ) dm = dy;
    dm = 1.3 * (dm / 2.0);
    cx = (patur.x + patll.x) / 2.0;
    cy = (patur.y + patll.y) / 2.0;
    patll.x = cx - dm;
    patur.x = cx + dm;
    patll.y = cy - dm;
    patur.y = cy + dm;
    patll.x = patll.x * dbu_um;
    patll.y = patll.y * dbu_um;
    patur.x = patur.x * dbu_um;
    patur.y = patur.y * dbu_um;
        
    
    extents = fopen( extentxt, "w" );
    if ( ! extents ) bail( "unable to open extents txt file" );
    fprintf( extents, "%lf %lf %lf %lf\n", patll.x, patll.y, patur.x, patur.y ); 
    fclose( extents );


    fprintf( out, "END\n" );

    fclose( in );
    fclose( out );

    printf( "\n\nNumber of shapes: %d\n\n", total );

    }

