// grid_snap.c
//                 Snap to the bss grid, just like poly2trap.

// Small changes for long xy type

#include "libgdsmin.h"
#include "polyq.h"
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>

float bss, bss2;          // beam step size, in bits, and some fraction of that
float several_beam_steps;

#define VERSION 2.0

#define TOOCLOSE( a ) ((fabs((a)) < bss2))

#define PI    3.1415926
#define TWOPI 6.2831853
#define PI2   1.5707963
#define PI32  4.7123890

#define printv noop
#define ifv(a) if(FALSE)

#define MAXMARRY 25000

double area_without_vertex( polygon *p, int x );
void   marry_vertices( queue *tout, BOOL fix );
double pointoline_dist( xy p, xy a, xy b );
double poly_area( polygon *p );
BOOL   self_crossing_polygon( polygon *p );


BOOL ding=FALSE;
int total=0;

void noop(){}


//-------------------------------------------------------------------------------------------------------------

BOOL
dequal( double a, double b )   // are two doubles roughly equal?
    {
    if ( a == b ) return( TRUE );

    if ( a == 0.0 )
        return( fabs(b) < 1.0e-7 );
    else if ( b == 0.0 )
        return( fabs(a) < 1.0e-7 );
    else
        return( fabs( (a - b)/a ) < 1.0e-7 );
    }

//---------------------------------------------------------------------------------------------------

BOOL
nearly_horizontal( xy a, xy b )
    {
    return( (abs(a.y - b.y) < bss) && (abs(a.x - b.x) > several_beam_steps) );  
    }

BOOL
nearly_vertical( xy a, xy b )
    {
    return( (abs(a.x - b.x) < bss) && (abs(a.y - b.y) > several_beam_steps) );  
    }

BOOL
samepoint( xy a, xy b )
    {
    return( (abs(a.x - b.x) < bss) && (abs(a.y - b.y) < bss) );  
    }

BOOL
exact_samepoint( xy a, xy b )
    {
    return( (a.x == b.x) && (a.y == b.y) );
    }

//------------------------------------------------------------------------------------------------------

BOOL
between( int x, int a, int b )       // Is x between a and b?
    {
    if ( x == a ) return( TRUE );
    if ( x == b ) return( TRUE );

    if ( a < b )  
	return( x >= a && x <= b );   
    else
	return( x >= b && x <= a );
    }

//------------------------------------------------------------------------------------------------------


double 
angle( xy a, xy b )   // The positive angle 0..2pi ccw with point a at the origin
    {
    static int quad;
    static double theta;

    if ( samepoint( a, b ) ) return( -1 ); // invalid, but this will avoid 180 heading change

    if ( a.y == b.y )
	{
	if ( b.x > a.x ) 
	    { return( 0.0 ); }
	else
	    { return( PI ); }
	}
    else if ( a.x == b.x )
	{
	if ( b.y > a.y ) 
	    { return( PI2 ); }
	else
	    { return( PI32 ); }
	}

    if ( b.y > a.y && b.x > a.x ) 
	quad = 1;
    else if ( b.y > a.y && b.x < a.x )
	quad = 2;
    else if ( b.y < a.y && b.x < a.x )
	quad = 3;
    else if ( b.y < a.y && b.x > a.x )
	quad = 4;
    else
        bail( "logic error in angle" );

    theta = atan( fabs(b.y - a.y) / fabs( b.x - a.x ) );

    if ( quad == 1 )
	return( theta );
    else if ( quad == 2 )
	return( PI - theta );
    else if ( quad == 3 )
	return( PI + theta );
    else
	return( TWOPI - theta );

    }  // angle


double
heading_change( xy a, xy b, xy c )
    {
    static double t1, t2, hc;
    t1 = angle( a, b );
    t2 = angle( b, c );
    hc = t2 - t1;
    if ( hc > PI ) 
	hc = hc - TWOPI;
    if ( hc < -PI )
	hc = hc + TWOPI;
    if ( dequal( hc, PI ) ) hc = PI;
    if ( fabs(hc) > PI )
	{
	printf( "t1=%1.2lf  t2=%1.2lf  hc=%1.2lf\n", t1, t2, hc );
	bail( "logic error in heading change" );
	}
    return( hc );
    }

//-------------------------------------------------------------------------------------------------------------

BOOL
proche( double x, int i )   // la meme chose, mais seulement un nombre
    {
    static double di, fbss;
    di = i;
    fbss = bss;
    return( fabs( x - di ) < fbss );
    }

//-------------------------------------------------------------------------------------------------------------


BOOL
entre( double x, int a, int b )       // est-ce que x entre a et b?
    {
    static double fa, fb;

    if ( proche( x, a ) ) return( TRUE );
    if ( proche( x, b ) ) return( TRUE );

    fa = a;
    fb = b;

    if ( a < b )  
        return( x >= fa && x <= fb );   
    else
        return( x >= fb && x <= fa );
    }


//-------------------------------------------------------------------------------------------------------------

void
remove_duplicate_vertices( polygon *p )
    {
    static int n, nm1, i;
    static xy   vi, vi1, v0, vnm1;
    static BOOL escape;

    n   = p->len;
    nm1 = n - 1;
    i   = 0;

    while ( i < nm1 )
	{
	vi  = vertex( p, i );
	vi1 = vertex( p, i+1 );
	escape = FALSE;
	while ( ! escape && (i < nm1) && samepoint( vi, vi1 ) )
	    {
	    escape = delete_vertex( p, i+1 );
	    n = p->len;
	    nm1 = n - 1;
	    vi  = vertex( p, i );
	    if ( i == nm1 )
		vi1 = vertex( p, 0 );
	    else 
		vi1 = vertex( p, i+1 );
	    }
	i += 1;
	}

    if ( n < 3 ) return;

    v0   = vertex( p, 0 );

    vnm1 = vertex( p, nm1 );

    if ( samepoint( v0, vnm1 ) ) p->len -= 1;

    }

//-------------remove_duplicate_vertices----------------------------------------------------------------

void
remove_flagpoles( polygon *p, BOOL severe )
    {
    static int n, nm1, i, i1, i2;
    static xy  a, b, c;
    static double tooclose;

    if ( severe ) 
	{
	tooclose = bss;
	}
    else
	{
        tooclose = bss/4.0;
        if ( tooclose < 1.0 ) tooclose = 1.0;
	}

    if ( ! p ) return;
    if ( ! p->active ) return;

    n   = p->len;
    nm1 = n - 1;
    i   = 0;
    i1  = (i + 1) % n;
    i2  = (i + 2) % n;

    while ( i < nm1 )
	{
	a = vertex( p, i  );
	b = vertex( p, i1 );
        c = vertex( p, i2 );

        if ( //dequal( PI, fabs(heading_change( a, b, c )) ) ||  // redundant
            ( pointoline_dist(a,b,c) < tooclose )            || 
	    ( pointoline_dist(c,a,b) < tooclose )                )
	    {
    	    delete_precious_vertex( p, i1 );
	    n = p->len;
	    nm1 = n - 1;
	    i1 = (i + 1) % n;
	    i2 = (i + 2) % n;
	    }
	else
	    {
	    i  = (i + 1) % n;
            i1 = (i + 1) % n;
            i2 = (i + 2) % n;
	    }
	}

    if ( n < 3 ) 
	{
	p->active = FALSE;
	return;
	}
    else 
	{
        a = vertex( p, nm1 );
        b = vertex( p, 0   );
        c = vertex( p, 1   );


        if  ( //dequal( PI, fabs(heading_change( a, b, c )) ) || 
            ( pointoline_dist(a,b,c) < tooclose )             || 
	    ( pointoline_dist(c,a,b) < tooclose )                 ) 
	    {
	    delete_precious_vertex( p, 0 );
	    }

        if ( p->len < 3 ) p->active = FALSE;
	}

    }

//--- end remove_flagpoles ------------------------------------------------------------------------------

void
remove_precise_duplicates( polygon *p )
    {
    static int n, nm1, i;
    static xy   vi, vi1, v0, vnm1;
    static BOOL escape;

    n   = p->len;
    nm1 = n - 1;
    i   = 0;

    while ( i < nm1 )
        {
        vi  = vertex( p, i );
        vi1 = vertex( p, i+1 );
        escape = FALSE;
        while ( ! escape && (i < nm1) && exact_samepoint( vi, vi1 ) )
            {
            escape = delete_vertex( p, i+1 );
            n = p->len;
            nm1 = n - 1;
            vi  = vertex( p, i );
            if ( i == nm1 )
                vi1 = vertex( p, 0 );
            else 
                vi1 = vertex( p, i+1 );
            }
        i += 1;
        }

    if ( n < 3 ) return;

    v0   = vertex( p, 0 );

    vnm1 = vertex( p, nm1 );

    if ( exact_samepoint( v0, vnm1 ) ) p->len -= 1;

    }

//-------------remove_precise_duplicates----------------------------------------------------------------

double
distance( xy p, xy q )
    {
    static double dx, dy, d;
    dx = p.x - q.x;
    dy = p.y - q.y;
    d = sqrt( dx * dx + dy * dy );
    return( d );
    }


double
pointoline_dist( xy p, xy a, xy b )
    {
    static double dx, dy, d, px, py, s, u, xx, yy;

    if ( samepoint( p, a ) )
        return( 0 );
    else if ( samepoint( p, b ) )
        return( 0 );
    else if ( samepoint( a, b ) ) 
        {
        dx = p.x - a.x;
        dy = p.y - a.y;
        d = sqrt( dx * dx + dy * dy );
        return( d );
        }
    else 
        {
        px = b.x - a.x;
        py = b.y - a.y;
        s = px * px + py * py;
        u = ((p.x - a.x) * px + (p.y - a.y) * py) / s;
        if ( u > 1.0 )
            u = 1.0;
        else if ( u < 0.0 )
            u = 0.0;
        xx = a.x + u * px;
        yy = a.y + u * py;
        dx = xx - p.x;
        dy = yy - p.y;
        d = sqrt( dx * dx + dy * dy );
        return( d );
        }

    }

//---------end pointoline_dist-----------------------------------------------------------------------------------

BOOL
colinear( xy a, xy b, xy c, int *middle )
    {
    // If the three points are colinear then return TRUE. The return parameter
    // "middle" will indicate which of the three points is in the middle.

    static BOOL first, second, third, vertical, horizontal, colinear;
    static double m;
    static double abc, bac, cab;

    first  = FALSE;
    second = FALSE;
    third  = FALSE;

    if ( (c.x == a.x) || (c.y == a.y) || (b.x == a.x) || (b.y == a.y) )
        {
        vertical   = (a.x == b.x) && (b.x == c.x); 
        horizontal = (a.y == b.y) && (b.y == c.y);
        colinear   = horizontal || vertical;
        if ( horizontal )
            {
            second = between( b.x, a.x, c.x );
            if ( ! second ) first = between( a.x, b.x, c.x );
            third = (! second) && (! first);
            }
        if ( vertical )
            {
            second = between( b.y, a.y, c.y );
            if ( ! second ) first = between( a.y, b.y, c.y );
            third = (! second) && (! first);
            }
        }
    else 
        {
        abc = pointoline_dist( a, b, c );
        bac = pointoline_dist( b, a, c );
        cab = pointoline_dist( c, a, b );

        colinear = TOOCLOSE( abc ) || TOOCLOSE( bac ) || TOOCLOSE( cab ); 

        if ( colinear )
            {
            second = between( b.x, a.x, c.x );
            if ( ! second ) first = between( a.x, b.x, c.x );
            third = (! second) && (! first);
            }
        }
         
    // It's very difficult to decide which point to delete. Usually, we should delete
    // the middle point. However, if two of the points are very close then we should
    // delete the endpoint, maybe.

    //                 It's more reliable to delete the middle point always,
    //                 but it leads to more nasty gaps.
    //*middle = 1;
    //return( colinear );

    if ( colinear )
        {
        if ( samepoint( a, b ) )
            { 
            *middle = 0;              // delete the first vertex
            //printf( "\ndelete a\n" );
            }
        else if ( samepoint( b, c ) )
            {            
            *middle = 2;              // delete the third vertex
            //printf( "\ndelete c\n" );
            }
        else        
            {
            if ( first ) 
                *middle = 0;
            else if ( second ) 
                *middle = 1;
            else if ( third )
                *middle = 2;
            }
        }
    

    return colinear;
    }

//---------end colinear------------------------------------------------------------------------------------------

void
remove_colinear_vertices( polygon *p, BOOL severe )
    {
    // Much harder than it might seem. Normally we would want to delete 
    // the vertex in the middle, but if the polygon doubles back on itself
    // then this can cause the polygon to become distorted. Therefore
    // we start by removing flag poles before looking at colinearity.
    

    static int i, n, nm1, nm2, j, k, middle;
    static BOOL escape;
    static xy vi, vj, vk;
    
    if ( ! p->active ) return;
    
    remove_flagpoles( p, severe );

    if ( ! severe ) return;

    i = 0;
    middle = -1;
    n = p->len;
    nm2 = n - 2;
    nm1 = n - 1;

    while ( i < nm2 )
        {
        j = i + 1;
        k = j + 1;
        escape = FALSE;

        while ( ! escape && (i < nm2) && colinear( vertex(p,i), vertex(p,j), vertex(p,k), &middle ) )
            {
            if ( middle == 1 )
                {
                escape = delete_vertex( p, j );   // if the vertex is precious then we should escape
                }
            else
                escape = TRUE;

            n = p->len;
            nm1 = n - 1;
            nm2 = n - 2;
            }    
        i = i + 1;
        }

    n = p->len;
    nm1 = n - 1;
    nm2 = n - 2;

    if ( n > 2 )
        {
        if ( colinear( vertex(p,nm2), vertex(p,nm1), vertex(p,0), &middle ) )
            {
            if ( middle == 1 ) delete_vertex( p, nm1 );
            }
        }

    n = p->len;
    nm1 = n - 1;
    nm2 = n - 2;
        
    if ( n > 2 )
        {
        if ( colinear( vertex(p,n-1), vertex(p,0), vertex(p,1), &middle ) )
            { 
            if ( middle == 1 ) delete_vertex( p, 0 );
            }
        }
    }

//-----------end remove_colinear_vertices------------------------------------------------------------------------

void
extent( polygon *p )
    {
    static xy v;
    static int n, i, minx, maxx, miny, maxy, b2;
    static BOOL init=FALSE;

    if ( ! init ) { init=TRUE; b2 = bss; }

    n = p->len;

    if ( n < 3 ) 
        {
        p->active = FALSE;
        p->ll.x = 0;
        p->ll.y = 0;
        p->ur.x = 0;
        p->ur.y = 0;
        return;
        }

    v = vertex( p, 0 );
    minx = v.x; 
    maxx = minx;
    miny = v.y;
    maxy = miny;

    for ( i=1; i < n; i++ )
        {
        v = vertex( p, i );
        if ( v.x < minx ) minx = v.x;
        if ( v.x > maxx ) maxx = v.x;
        if ( v.y < miny ) miny = v.y;
        if ( v.y > maxy ) maxy = v.y;
        }

    p->ll.x = minx - b2;
    p->ll.y = miny - b2;
    p->ur.x = maxx + b2;
    p->ur.y = maxy + b2;
    }

//----------end extent-------------------------------------------------------------------------------------------

polygon *
copy_from_gds_polygon( struct gds_itemtype *item, double dbu_rescale )
    {
    // copy item to poly
    // poly has already been initialized

    static int i;
    static double x, y;
    polygon *poly;

    poly = initialize_polygon( item->n, item->layer, item->dt, FALSE );

    for ( i=0; i < item->n; i++ )
        {
        x = item->x[i] / dbu_rescale;
        y = item->y[i] / dbu_rescale;
        append_polygon( poly, x, y );
        }
    
    //remove_colinear_vertices( poly, FALSE );

    extent( poly );
    return( poly );

    }  // copy_from_gds_polygon


//---------------------------------------------------------------------------------------------------------------

BOOL
on_segment( double x, double y, xy a1, xy a2, xy b1, xy b2 )
    {

    // Is the point x,y somewhere on both line segments a1-a2 and b1-b2 ?

    static BOOL cross;

    // If a1-a2 is the same as b1-b2 then they definitely cross, and the intersection
    // is definitely on the segment. If overlaps are bad, then we should think about
    // that elsewhere.

    if ( (samepoint(a1, b1) && samepoint(a2, b2)) || (samepoint(a1, b2) && samepoint(a2, b1)) ) return( TRUE );


    if ( a1.y == a2.y ) 
        cross = entre( x, a1.x, a2.x ) && entre( y, b1.y, b2.y );
    else if ( b1.y == b2.y )
        cross = entre( x, b1.x, b2.x ) && entre( y, a1.y, a2.y );
    else if ( a1.x == a2.x )
        cross = entre( x, b1.x, b2.x ) && entre( y, a1.y, a2.y );
    else if ( b1.x == b2.x )
        cross = entre( x, a1.x, a2.x ) && entre( y, b1.y, b2.y );
    else
        cross = entre( x, b1.x, b2.x ) && entre( y, b1.y, b2.y ) &&
                entre( x, a1.x, a2.x ) && entre( y, a1.y, a2.y );

    return( cross );

    }  // on_segment

//--------------------------------------------------------------------------------------------

BOOL
in_segment( double x, double y, xy a1, xy a2, xy b1, xy b2 )
    {

    // Is the point x,y somewhere on both line segments a1-a2 and b1-b2 ...
    // BUT not on the end points?

    static BOOL cross;
    static xy   v;

    if ( samepoint(a1, b1) || samepoint(a2, b2) || samepoint(a1, b2) || samepoint(a2, b1) ) return( FALSE );

    v.x = (int) round( x );
    v.y = (int) round( y );

    if ( samepoint(v, a1) || samepoint(v, a2) || samepoint(v, b1) || samepoint(v, b2) ) return( FALSE ); 

    if ( a1.y == a2.y ) 
        cross = entre( x, a1.x, a2.x ) && entre( y, b1.y, b2.y );
    else if ( b1.y == b2.y )
        cross = entre( x, b1.x, b2.x ) && entre( y, a1.y, a2.y );
    else if ( a1.x == a2.x )
        cross = entre( x, b1.x, b2.x ) && entre( y, a1.y, a2.y );
    else if ( b1.x == b2.x )
        cross = entre( x, a1.x, a2.x ) && entre( y, b1.y, b2.y );
    else
        cross = entre( x, b1.x, b2.x ) && entre( y, b1.y, b2.y ) &&
                entre( x, a1.x, a2.x ) && entre( y, a1.y, a2.y );

    return( cross );

    }  // in_segment

//--------------------------------------------------------------------------------------------

int
snap( int x )
    {
    //if ( ! (x % bss) ) return( x );  bss is float now
    x = x / bss;
    x = x * bss;
    return( x );
    }

void
snapoint( xy *v ) 
    {
    v->x = snap( v->x );
    v->y = snap( v->y );
    }

double
circumfrence( polygon *p )
    {
    static int i, ni;
    static double dx, dy, c;
    static xy v, nv;

    c = 0.0;
    for ( i=0; i < p->len; i++ )
	{
	ni = (i + 1) % p->len;
        v  = vertex( p, i );
        nv = vertex( p, ni );        
	dx = v.x - nv.x;
        dy = v.y - nv.y;
        c = c + sqrt( dx * dx + dy * dy );
	}
    return( c );
    }

void
snap_to_grid( polygon *p )
    {
    static int i, ni;
    static xy v, nv, ov, onv;
    static double newc, oldc;

    if ( ! p || ! p->active ) return;
    for ( i=0; i < p->len; i++ )
	{
        v  = vertex( p, i );
        ni = (i+1) % p->len;
        nv = vertex( p, ni );              // if the next vertex
        if ( nearly_vertical( v, nv ) )    // has nearly the same x, then
            {                              // set them equal. 
            oldc  = circumfrence( p );
            ov.x  = v.x;
            onv.x = nv.x;
            nv.x  = v.x;
            set_vertex( p, ni, nv );
            newc = circumfrence( p ); // but be careful-
            if ( newc < oldc )        // pick the one that creates
		{                     // the largest polygon
		v.x  = ov.x;
                nv.x = onv.x;
                v.x  = nv.x;
                set_vertex( p, ni, nv );
                set_vertex( p, i, v );
                if ( self_crossing_polygon( p ) )
		    {
		    v.x  = ov.x;
		    nv.x = onv.x;
		    nv.x = v.x;
		    set_vertex( p, ni, nv );
                    //printf( "." );
		    }
		}
            }
        if ( nearly_horizontal( v, nv ) )
            {                              // set them equal. 
            oldc  = circumfrence( p );
            ov.y  = v.y;
            onv.y = nv.y;
            nv.y  = v.y;
            set_vertex( p, ni, nv );
            newc = circumfrence( p ); // but be careful-
            if ( newc < oldc )        // pick the one that creates
		{                     // the largest polygon
		v.y  = ov.y;
                nv.y = onv.y;
                v.y  = nv.y;
                set_vertex( p, ni, nv );
                set_vertex( p, i, v );
                if ( self_crossing_polygon( p ) )
		    {
		    v.y  = ov.y;
		    nv.y = onv.y;
		    nv.y = v.y;
		    set_vertex( p, ni, nv );
                    //printf( ":" );
		    }
		}
            }


        snapoint( &v );
        set_vertex( p, i, v );
	}
    }

//--------------------------------------------------------------------------------------------

BOOL
not_crossing( polygon *p, int a, int b, int c, int d, xy *intersection, BOOL overlapsbad )
    {
    // Does the segment a-b cross the segment c-d?
    // Return True if they do not cross, and return the 
    // intersection point if they do. 
    
    // If the line a-b is the same as the line c-d then
    // these line segments are still "crossing", because
    // doughnut polygons must be eliminated. The program
    // darn.c should have eliminated these sort of holes, 
    // but we must make sure here.

    // If we are making a slice, then we should NOT consider
    // coincident lines to be crossing. It's ok to slice
    // along one of the sides. Hence the logical overlapsbad.

    static xy
        inter,
        a1, a2, b1, b2;

    static BOOL
        crossing,
        a_is_vertical,
        b_is_vertical,
        ma_equals_mb,
        ba_equals_bb;

    static double
        fb1x, fb1y,
        fb2x, fb2y,
        fa1x, fa1y,
        fa2x, fa2y,
        fx, fy,
        ma, mb, 
        ba, bb;

    static int
        ix, iy;

    a1 = vertex( p, a );
    a2 = vertex( p, b );
    b1 = vertex( p, c );
    b2 = vertex( p, d );

    intersection->x = 0;
    intersection->y = 0;

    // first check that the segments are anywhere near each other

    if ( a1.x < b1.x && a1.x < b2.x && a2.x < b1.x && a2.x < b2.x ) return( TRUE );
    if ( a1.x > b1.x && a1.x > b2.x && a2.x > b1.x && a2.x > b2.x ) return( TRUE );
    if ( a1.y < b1.y && a1.y < b2.y && a2.y < b1.y && a2.y < b2.y ) return( TRUE );
    if ( a1.y > b1.y && a1.y > b2.y && a2.y > b1.y && a2.y > b2.y ) return( TRUE );

    // overlaps are bad when we are looking for self-intersections or doughnut shapes.
    // overlaps are not bad when we are slicing.

    if ( ( samepoint( a1, b1 ) && samepoint( a2, b2 ) ) || ( samepoint( a1, b2 ) && samepoint( a2, b1 ) ) )
        {
        intersection->x = a1.x;
        intersection->y = a1.y;
        if ( overlapsbad )
            {
            //printf( "matching segment - probably a doughnut\n" );
            return( FALSE ); // they are crossing, probably a doughnut shape
            }
        else
            return( TRUE );  // they are the same line segment, but this is ok
        }


    if ( samepoint( a1, b1 ) || samepoint( a1, b2 ) )
        {
        intersection->x = a1.x;
        intersection->y = a1.y;
        if ( overlapsbad )
            {
            //printf( "lines touch at an end point\n" ); 
            return( FALSE ); // => crossing
            }
        else
            return( TRUE );
        }

    if ( samepoint( a2, b1 ) || samepoint( a2, b2 ) )  
        {
        intersection->x = a2.x;
        intersection->y = a2.y;
        if ( overlapsbad )
            {
            //printf( "lines touch at an end point\n" ); 
            return( FALSE );
            }
        else
            return( TRUE );
        }

    a_is_vertical = (a1.x == a2.x); 
    b_is_vertical = (b1.x == b2.x); 

    if ( a_is_vertical && b_is_vertical && ! overlapsbad ) return( TRUE );

    if ( a_is_vertical && b_is_vertical )  // and overlaps are bad...
        {
        //printf( "both line segments are vertical\n" ); 
        if ( a1.x == b1.x )
            {
            if ( between( a1.y, b1.y, b2.y ) )
                {
                //printf( "vertical segments overlap\n" );
                intersection->x = a1.x;
                intersection->y = a1.y;
                return( FALSE );
                }
            else if ( between( a2.y, b1.y, b2.y ) )
                {
                //printf( "vertical segments overlap\n" );
                intersection->x = a2.x;
                intersection->y = a2.y;
                return( FALSE );
                }
            else
                {
                return( TRUE );
                }
            }
        else
            {
            return( TRUE );
            }
        }

    if ( a_is_vertical && ! b_is_vertical )
        {
        //printf( "a is vertical but b is not\n" );

        fb1y = b1.y;
        fb2y = b2.y;
        fb1x = b1.x;
        fb2x = b2.x;

        mb = (fb1y - fb2y) / (fb1x - fb2x);
        bb = fb1y - mb * fb1x;

        fx = a1.x;
        fy = mb * fx + bb;

        // fx,fy are floats which have not snapped to the bss grid.
        // Is that what we want? Yes, probably.

        crossing = on_segment( fx, fy, a1, a2, b1, b2 );

        if ( ! crossing )
            {
            intersection->x = 0;
            intersection->y = 0;
            return( TRUE );
            }
        else
            {
            //printf( "segments a and b cross. a is vertical, b is not\n" );
            intersection->x = a1.x; //snap( a1.x );
            iy = round( fy );
            intersection->y = iy; //snap( iy );
            return( FALSE );
            }
        }
        

    if ( b_is_vertical && ! a_is_vertical )
        {
        //printf( "b is vertical but a is not\n" );

        fa1y = a1.y;
        fa2y = a2.y;
        fa1x = a1.x;
        fa2x = a2.x;

        ma = (fa1y - fa2y) / (fa1x - fa2x);
        ba = fa1y - ma * fa1x;

        fx = b1.x;
        fy = ma * fx + ba;

        // fx,fy are floats which have not snapped to the bss grid.
        // Is that what we want? Yes, probably.

        crossing = on_segment( fx, fy, a1, a2, b1, b2 );

        if ( ! crossing )
            {
            intersection->x = 0;
            intersection->y = 0;
            return( TRUE );
            }
        else
            {
            //printf( "segments a and b cross. b is vertical, a is not\n" );
            intersection->x = b1.x; //snap( b1.x );
            iy = round( fy );
            intersection->y = iy; //snap( iy );
            return( FALSE );
            }
        }

    // neither segment is vertical
        
    fa1x = a1.x;  fa1y = a1.y;
    fa2x = a2.x;  fa2y = a2.y;
    fb1x = b1.x;  fb1y = b1.y;
    fb2x = b2.x;  fb2y = b2.y;

    ma = (fa1y - fa2y) / (fa1x - fa2x);
    mb = (fb1y - fb2y) / (fb1x - fb2x);

    bb = b1.y - mb * b1.x;
    ba = a1.y - ma * a1.x;

    ma_equals_mb = dequal( ma, mb );
    ba_equals_bb = dequal( ba, bb );

    if ( ma_equals_mb && ! ba_equals_bb )
        {
        intersection->x = 0;
        intersection->y = 0;
        return( TRUE );      // they are not crossing
        }

    if ( ma_equals_mb )  // Coincident lines, but do the segments overlap?
        {                // We already considered the case where the segments
                         // are the same, so this is not a doughnut cut.
                         // Also, it is not a slice cut, since slice lines
                         // are always on-axis.

        //printf( "coincident lines\n" );

        if ( ma > 1.0 )
            {
            if ( between( a2.y, b1.y, b2.y ) )
                {
                //printf( "off-axis segments that overlap\n" );
                intersection->x = a2.x;
                intersection->y = a2.y;
                return( FALSE );
                }
            else if ( between( a1.y, b1.y, b2.y ) )
                {
                //printf( "off-axis segments that overlap\n" );
                intersection->x = a1.x;
                intersection->y = a1.y;
                return( FALSE );
                }
            else
                {
                intersection->x = 0;
                intersection->y = 0;
                return( TRUE );
                }
            } 
        else
            {
            if ( between( a2.x, b1.x, b2.x ) )
                {
                //printf( "off-axis segments that overlap\n" );
                intersection->x = a2.x;
                intersection->y = a2.y;
                return( FALSE );
                }
            else if ( between( a1.x, b1.x, b2.x ) )
                {
                //printf( "off-axis segments that overlap\n" );
                intersection->x = a1.x;
                intersection->y = a1.y;
                return( FALSE );
                }
            else
                {
                intersection->x = 0;
                intersection->y = 0;
                return( TRUE );
                }
            }
        }
            
    // Finally, just regular lines not on-axis, not parallel

    fx = (bb - ba) / (ma - mb); 
    fy = ma * fx + ba; 

    crossing = on_segment( fx, fy, a1, a2, b1, b2 ); 

    if ( crossing )
        {
        //printf( "off axis segments that cross\n" );
        //printf( "a = %ld,%ld  %ld,%ld\n", a1.x, a1.y, a2.x, a2.y );
        //printf( "b = %ld,%ld  %ld,%ld\n", b1.x, b1.y, b2.x, b2.y );
        //printf( "intersection = %1.1lf, %1.1lf \n", fx, fy );
        ix = round( fx );
        intersection->x = ix; //snap( ix );
        iy = round( fy );
        intersection->y = iy; //snap( iy );
        return( FALSE );
        }
    else
        {
        intersection->x = 0;
        intersection->y = 0;
        return( TRUE );
        }

    }

//---------end not_crossing -------------------------------------------------------------------------------------

BOOL
clockwise( polygon *p )
    {
    // Determine the sign of a polygon's curl by calculating the net change in heading.
    // Seems like a stupid way of calculating the curl, but it's actually a lot more reliable
    // than doing the cross-product. Why?

    static xy a, b, c;
    static int n, i;
    static double dt, hc;

    if ( ! p ) bail( "null polygon sent to clockwise" );
    if ( ! p->active ) bail( "inactive polygon sent to clockwise" );
    if ( p->len < 3 ) bail( "degenerate polygon sent to clockwise" );

    n = p->len;
    dt = 0.0;

    for ( i=0; i < n; i++ )
	{
	a = vertex( p, i );
	b = vertex( p, (i+1) % n );
	c = vertex( p, (i+2) % n );
        hc = heading_change( a, b, c );
        dt = dt + hc;
	}

    return( dt <= 0.0 );
    }


//-------clockwise-----------------------------------------------------------------------------------------------

void
ccw( polygon *p, BOOL *ccwise, BOOL *unknown, double *area )
    {
 
    // Determine the clock wisdom of a polygon
    // by calculating the area. If the area is
    // small then we cannot trust the curliness
    // of the shape. This does not work so well.
    // There is something wrong with this calculation. 
    // Yes there really is something wrong here.

    static int    n;
    static double c, toosmall;
    static int    i, j;
    static xy     vi, vj;
    static BOOL   init=FALSE;

    if ( ! init ) { init=TRUE; toosmall=bss*bss/2.0; }

    n = p->len;
    c = 0.0;

    for ( i=0; i < n; i++ )
        {
        j = (i + 1) % n;
        vi = vertex( p, i );
        vj = vertex( p, j );
        c = c + vi.x * vj.y - vj.x * vi.y;
        }

    *ccwise = c > 0.0;
    *unknown = fabs(c) < toosmall; 
    *area = fabs( c / 2.0 );
    }



double
poly_area( polygon *p )
    {
    static BOOL ccwise, unknown;
    static double area;
    ccw( p, &ccwise, &unknown, &area );
    return( area );
    }


double
area_without_vertex( polygon *p, int x ) // area if we delete vertex x
    {
    static int    n;
    static double c;
    static int    i, j;
    static xy     vi, vj;

    n = p->len;
    c = 0.0;

    i = 0;
    if ( i == x ) i++;

    while ( i < n )
        {
        j = (i + 1) % n;
        vi = vertex( p, i );
        vj = vertex( p, j );
        c = c + vi.x * vj.y - vj.x * vi.y;
	i++;
        if ( i == x ) i++;
        }

    return( fabs( c / 2.0 ) );
    }

 
double
circumference( polygon *p )
    {                                   
    static int    n;
    static double c, dx, dy;
    static int    i, j;
    static xy     vi, vj;

    n = p->len;
    c = 0.0;

    i = 0;
    while ( i < n )
        {
        j = (i + 1) % n;
        vi = vertex( p, i );
        vj = vertex( p, j );
        dx = vi.x - vj.x;
        dy = vi.y - vj.y;
        c = c + sqrt( dx * dx + dy * dy );
        i++;
        }
    
    return( c );
    }
 

//-----ccw et al---------------------------------------------------------------------------------------

polygon_pair *
split_touchy_polygon( polygon *p, int v, int a, int b )
    {
    // Vertex v touches line segment p[a]--p[b]. Split p into two polygons. Let's be careful
    // to retain the original order so that both resulting polygons have the same
    // clock-wiseiness. This is quite tricky.

    static int     n, i;

    static double  change_in_area, area, 
                   area1, area2,
                   toosmall, toomuchange;

    static BOOL    p1_degenerate, p2_degenerate,
                   ccw_p,  
                   ccw_p1, 
                   ccw_p2, 
	           doughnut, 
	           init=FALSE;

    polygon *p1, *p2;
    polygon_pair *pp;

    xy pv, pa, pb;

    pp = initialize_polygon_pair();

    pv = vertex( p, v );
    printv( "split touchy at %d, %d\n", pv.x, pv.y );

    if ( ! init ) 
        { 
        init=TRUE; 
        toosmall = bss * bss * 0.8; 
        toomuchange = bss * bss * 4.0;
        }

    doughnut = FALSE;

    area = poly_area( p );

    if ( area < toosmall ) 
        {
	//printv( "the area %1.1lf is < %1.1lf\n", area, toosmall );
        pp->a = NULL;
        pp->b = NULL;
        p->active = FALSE;
        return( pp );
        }

    p1 = initialize_polygon( p->len, p->layer, p->dt, p->obese );
    p2 = initialize_polygon( p->len, p->layer, p->dt, p->obese );

    n = p->len; 
    i = v;

    while ( (i != b) && (i != a) )
        {
        append_polygon_xy( p1, vertex( p, i ) );
        i = (i + 1) % n;
        }
    append_polygon_xy( p1, vertex( p, i ) );

    i = v;
    while ( (i != b) && (i != a) )
        {
        append_polygon_xy( p2, vertex( p, i ) );
        i -= 1;
        if ( i == -1 ) i = n-1;
        }

    append_polygon_xy( p2, vertex( p, i ) ); 
    
    extent( p1 );
    extent( p2 );

    p1_degenerate = (p1->len < 3) || (poly_area(p1) < toosmall);
    p2_degenerate = (p2->len < 3) || (poly_area(p2) < toosmall);

    if ( p1_degenerate && p2_degenerate )
        {
        //printv( "p1 and p2 are both degenerate\n" );
        pp->a = NULL;
        pp->b = NULL;
        p->active = FALSE;
        return( pp );
        }
    else if ( p1_degenerate )
        {
        // sometimes causes a fill-in, so we reject them both
        pp->a = NULL;
        pp->b = NULL; //p2;
        return( pp );
        }
    else if ( p2_degenerate )
        {
        // sometimes causes a fill-in, so we reject them both
        pp->a = NULL; //p1;
        pp->b = NULL;
        return( pp );
        }

    area1 = poly_area( p1 );
    area2 = poly_area( p2 );

    ccw_p1 = TRUE;
    ccw_p2 = FALSE;

    if ( p1->active && p2->active )
	{
        ccw_p1 = ! clockwise( p1 );
        ccw_p2 = ! clockwise( p2 );
	}

    if ( area1 < toosmall && area2 < toosmall )
        {
        pp->a = NULL;
        pp->b = NULL;
        p->active = FALSE;
        change_in_area = 0.0;
        }
    else if ( area1 < toosmall ) 
        {
        pp->a = NULL;
        pp->b = p2;
        change_in_area = fabs(area - area1 - area2);  // is this right? or should it be zero?
        }
    else if ( area2 < toosmall )
        {
        pp->a = p1;
        pp->b = NULL;
        change_in_area = fabs(area - area1 - area2); // maybe it should be zero?;
        }
    else
        {
        pp->a = p1;
        pp->b = p2;
        change_in_area = fabs(area - area1 - area2);
        }

    // something is wrong with the way we calculate area
    // so this next check does not work 

    if ( FALSE ) // change_in_area > toomuchange )
        {
        printv( "large change in area - split rejected\n" );
        printv( "        area= %1.2f  area1= %1.2f  area2= %1.2f\n", area, area1, area2 );
        printv( "        change = %1.1f bss^2\n", change_in_area / (bss*bss) );
	pp->a = NULL;
        pp->b = NULL;
        return( pp );
	}

    doughnut = (ccw_p1 == ccw_p2);

    if ( doughnut ) 
	{
	printv( "p1 and p2 have the same curl ... ignored\n" );
        pp->a = NULL; 
        pp->b = NULL;
        return( pp );
	}

    return( pp );

    }

//-------split_touchy_polygon ------------------------------------------------------------------------------------


polygon_pair *
untie_bow( polygon *p, xy ip, int i1, int i2, int j1, int j2 )
    {
    polygon_pair *pp;
    polygon *p1, *p2;
    static int n, i, t;
    
    pp = initialize_polygon_pair();

    p1 = initialize_polygon( p->len, p->layer, p->dt, p->obese );
    p2 = initialize_polygon( p->len, p->layer, p->dt, p->obese );

    n = p->len;

    if ( i2 < i1 )
        {
        t  = i1;
        i1 = i2;
        i2 = t;
        }

    if ( i2 == (n-1) )
        {
        t  = i1;
        i1 = i2;
        i2 = t;
        }

    if ( j2 < j1 )
        {
        t  = j1;
        j1 = j2;
        j2 = t;
        }

    if ( j2 == (n-1) )
        {
        t  = j1;
        j1 = j2;
        j2 = t;
        }

    i = i2;
    while ( (i != j1) && (i != j2) ) 
        {
        append_polygon_xy( p1, vertex( p, i ) );
        i = (i + 1) % n;
        }
    append_polygon_xy( p1, vertex( p, i ) );
    append_polygon_xy( p1, ip );

    i = j2;
    while ( (i != i1) && (i != i2) )
        {
        append_polygon_xy( p2, vertex( p, i ) );
        i = (i + 1) % n;
        }
    append_polygon_xy( p2, vertex( p, i ) );
    append_polygon_xy( p2, ip ); 

    //p1->layer = 7;
    //p2->layer = 7;

    extent( p1 );
    extent( p2 );
    pp->a = p1;
    pp->b = p2;
    return( pp );
    }

//------ untie_bow ------------------------------------------------------------------------------------------


BOOL
find_and_fix_self_crossing_polygon( polygon *p, queue *todo_list )
    {
    // If the polygon is self crossing, this function chops it into two polygons which
    // are pushed onto the todo_list queue. If one or both of those new polygons are
    // self-crossing, then they will be further chopped as the queue is processed.

    static BOOL   good, init=FALSE; 
    static int    n, i, i1, j, j1;
    static xy     ip, vi, vj, vi1, vj1;
    polygon_pair *pp;
    static double area, toosmall;
    
    if ( ! init ) {init=TRUE; toosmall=bss*bss; }
    
    if ( poly_area( p ) < toosmall )
        {
	//printf( "polygon is too small, find/fix deletes it\n" );
        p->active = FALSE;
        return( TRUE );    
        }

    good = TRUE;
    n = p->len;
    if ( n < 4 ) return( FALSE );

    i = 0;
    while ( good && (i < (n-2)) )   // notice that the polygon will be bad only once,
        {                           // at which point it will be split and the pieces 
        j = i + 2;                  // appended to the todo_list.
        while ( good && (j < n))
            {
            j1 = (j+1) % n;
            i1 = i + 1;

            if ( j1 != i )
                good = not_crossing( p, i, i1, j, j1, &ip, TRUE );  // overlapsbad=true since doughnuts are bad

            //if ( ! good ) printf( "n: %d  i: %d   j: %d   ip: %ld,%ld \n", n, i, j, ip.x, ip.y );

            if ( ! good )  
                {
                vi  = vertex( p, i );
                vi1 = vertex( p, i1 );
                vj  = vertex( p, j );
                vj1 = vertex( p, j1 );

                if ( samepoint( ip, vi ) )
                    {
                    pp = split_touchy_polygon( p, i, j, j1 );
                    }
                else if ( samepoint( ip, vi1 ) )
                    { 
                    pp = split_touchy_polygon( p, i1, j, j1 );
                    }
                else if ( samepoint( ip, vj ) )
                    {
                    pp = split_touchy_polygon( p, j, i, i1 );
                    }
                else if ( samepoint( ip, vj1 ) ) 
                    {
                    pp = split_touchy_polygon( p, j1, i, i1 );
                    }
                else
                    {
                    pp = untie_bow( p, ip, i, i1, j, j1 );
                    }

                if ( ! pp->a && ! pp->b && p->active )  // maybe not a failure! maybe just tiny shapes
                    {                                   // maybe it really was not a bowtie
                    //print_polygon( p, 1000 );
                    //printf( "ip: %ld, %ld \n", ip.x, ip.y );
                    }
                else            // success
                    {
                    p->active = FALSE;
                    remove_flagpoles( pp->a, FALSE );
                    remove_flagpoles( pp->b, FALSE );
                    if ( pp->a ) append_queue( todo_list, pp->a );
                    if ( pp->b ) append_queue( todo_list, pp->b );
                    }
                }

            j += 1;
            }
        i += 1;
        }

    return( ! good );
    }

//-----------end find_and_fix_self_crossing_polygon ----------------------------------------------------------

BOOL
self_crossing_polygon( polygon *p )
    {
    static BOOL   good, init=FALSE;
    static int    n, i, i1, j, j1;
    static xy     ip;

    good = TRUE;
    n = p->len;
    if ( n < 4 ) return( FALSE );

    i = 0;
    while ( good && (i < (n-2)) ) 
        {                         
        j = i + 2;                
        while ( good && (j < n))
            {
            j1 = (j+1) % n;
            i1 = i + 1;

            if ( j1 != i )
                good = not_crossing( p, i, i1, j, j1, &ip, FALSE );  // overlapsbad=false

            j += 1;
            }
        i += 1;
        }

    return( ! good );
    }

//-----------end self-crossing polygon------------------------------------------------------------------------

void
copy_to_gds_polygon( polygon *poly, struct gds_itemtype *item )
    {
    // copy poly to item 
    // item has already been allocated 
    // note that we are not snapping here

    static int i;
    static int ix, iy;
    static xy  v;
    static long li;

    if ( item->n < poly->len ) bail( "attempt to copy large poly into small gds item" );

    for ( i=0; i < poly->len; i++ )
        {
        li = i;
        v = vertex( poly, li );
        item->x[i] = v.x;
        item->y[i] = v.y;
        }

    item->layer = poly->layer;
    item->dt    = poly->dt;
    item->n     = poly->len;
    
    }  // copy_to_gds_polygon

//-----------------------------------------------------------------------------------------------------


void write_trap_list( int fdo, queue *list ) 
    {
   
    static polygon     *p;
    static polygon_ref *pr;
    static int          ilayer, idt, ilen;
    static BOOL         init=FALSE;
    static int          maxlen=0;
    static struct gds_itemtype *gdspoly;

    if ( ! init )
        {
        gdspoly = (struct gds_itemtype *) malloc( sizeof( struct gds_itemtype ) );
        init = TRUE;
        }
    
    pr = list->first;    
    
    while ( pr )
        {
        p = pr->p;
        if ( p->active && p->len != 0 )
            {
            ilen = p->len;
            if ( ilen > maxlen )
                {
                gdspoly->x = (int *) malloc( ilen * sizeof( int ) ); 
                gdspoly->y = (int *) malloc( ilen * sizeof( int ) );
                maxlen = ilen;
                }
            gdspoly->n = ilen; 

            copy_to_gds_polygon( p, gdspoly );

            //if ( p->len > 4 ) p->layer = ERRORLAYER; 

            gds_write_boundary( fdo );
            ilayer = p->layer;
            idt    = p->dt;
            gds_write_layer(    fdo, ilayer );
            gds_write_datatype( fdo, idt    );
            gds_write_xy( fdo, gdspoly->x, gdspoly->y, gdspoly->n );
            gds_write_endel( fdo );      
            total++;

            // printf( "." ); fflush( stdout );
            }
        pr = pr->next;
        }

    // note that this function does not close the structure 

    } // write_trap_list

//---------------------------------------------------------------------------------------------------------------


void 
write_polygon( int fdo, polygon *p ) 
    {
   
    static int    ilayer, idt, ilen;
    static BOOL   init=FALSE;
    static int    maxlen=0;
    static struct gds_itemtype *gdspoly;

    if ( ! init )
        {
        gdspoly = (struct gds_itemtype *) malloc( sizeof( struct gds_itemtype ) );
        init = TRUE;
        }
    
    if ( ! p ) return;

    if ( p->active && p->len != 0 )
        {
        ilen = p->len;
        if ( ilen > maxlen )
            {
            gdspoly->x = (int *) malloc( ilen * sizeof( int ) ); 
            gdspoly->y = (int *) malloc( ilen * sizeof( int ) );
            maxlen = ilen;
            }
        gdspoly->n = ilen; 

        copy_to_gds_polygon( p, gdspoly );

        //if ( p->len > 4 ) p->layer = ERRORLAYER; 

        gds_write_boundary( fdo );
        ilayer = p->layer;
        idt    = p->dt;
        gds_write_layer(    fdo, ilayer );
        gds_write_datatype( fdo, idt    );
        gds_write_xy( fdo, gdspoly->x, gdspoly->y, gdspoly->n );
        gds_write_endel( fdo );      

        }

    // note that this function does not close the structure 

    } // write_polygon

//---------------------------------------------------------------------------------------------------------------

BOOL 
is_parallelogram_already( polygon *p )
    {

    // Are two sides parallel and on-axis?
    // Or maybe it's a triangle with one side on-axis.
    // Off-axis parallelograms - nope.
    
    static int  n;
    static BOOL yup, onaxis, horizontal, vertical,
                hor01, hor12, hor23, hor30,
                ver01, ver12, ver23, ver30; 
    static xy   v0, v1, v2, v3;

    n = p->len;

    if ( (n != 3) && (n != 4) ) return( FALSE );

    yup = FALSE;
    v0 = vertex( p, 0 );
    v1 = vertex( p, 1 );
    v2 = vertex( p, 2 );

    if ( n == 3 )
        {  
        onaxis = v0.y == v1.y;
        if ( ! onaxis ) onaxis = (v1.y == v2.y);
        if ( ! onaxis ) onaxis = (v2.y == v0.y); 
        if ( ! onaxis ) onaxis = (v0.x == v1.x);
        if ( ! onaxis ) onaxis = (v1.x == v2.x); 
        if ( ! onaxis ) onaxis = (v2.x == v0.x); 
        yup = onaxis;
        }
    else
        {
        v3 = vertex( p, 3 );
        hor01 = (v0.y == v1.y); 
        hor12 = (v1.y == v2.y); 
        hor23 = (v2.y == v3.y); 
        hor30 = (v3.y == v0.y); 
        horizontal = (hor01 && hor23) || (hor12 && hor30);
        vertical = FALSE;
        if ( ! horizontal )
            {
            ver01 = (v0.x == v1.x); 
            ver12 = (v1.x == v2.x); 
            ver23 = (v2.x == v3.x); 
            ver30 = (v3.x == v0.x); 
            vertical = (ver01 && ver23) || (ver12 && ver30);
            }
        yup = horizontal || vertical;
        }
    return yup;
    }

//-----------end is_parallelogram_already---------------------------------------------------------------------

BOOL
crossing_point( xy a1, xy a2, xy b1, xy b2, xy *intersection )
    {
    // Does the line a1-a2 cross the line b1-b2 ? Return True if they cross, and return the 
    // intersection point if they do. This is the same logic as not_crossing, but
    // we have to repackage the vertices first.

    static BOOL crossing;
    static BOOL init=FALSE;
    static polygon *p;

    if ( ! init ) { init=TRUE; p = initialize_polygon( 4, 0, 0, FALSE ); }

    zero_polygon( p );
    append_polygon_xy( p, a1 );
    append_polygon_xy( p, a2 );
    append_polygon_xy( p, b1 );
    append_polygon_xy( p, b2 );

    crossing = ! not_crossing( p, 0, 1, 2, 3, intersection, TRUE );

    return( crossing );
    }

//----------end crossing_point-----------------------------------------------------------------------------------------

BOOL
they_might_overlap( polygon *p, polygon *q )
    {
    // note that the extents were expanded by +- bss
    //
    if ( p->ur.x < q->ll.x ) return( FALSE );
    if ( p->ll.x > q->ur.x ) return( FALSE );
    if ( p->ur.y < q->ll.y ) return( FALSE );
    if ( p->ll.y > q->ur.y ) return( FALSE );
    return( TRUE );    
    }

//----------------------------------------------------------------------------------------------

void
marry_vertices( queue *tout, BOOL fix )    // look for vertices which are closer than bss and
    {                                      // set them to the same value so they will snap together
    polygon_ref *pr, *qr;                  // Maybe we should call this the "healing" function.
    polygon     *p,  *q, *ptemp;
    xy           v,   w;
    xy           wa,  wb;
    int          i,   j;
    int          dx,  dy,  d, range;
    int          besti, bestj, btemp;
    BOOL         escape, skipi;
    double       cw, cv;


    // look for vertices near each other -----------------------

    //printf( "Snapping vertices...\n" ); fflush(stdout);

 
    pr = tout->first;
    while ( pr && ! pr->p->active ) pr = pr->next;

    while ( pr )
        {
        p = pr->p;
        qr = pr->next;
        while ( qr && ! qr->p->active ) qr = qr->next;
        range = 0;
        while ( qr && (range < MAXMARRY))
            {
            q = qr->p;
            range++;
            if ( they_might_overlap( p, q ) )
                {
                i = 0;
                while ( i < p->len )
                    {
                    v = vertex( p, i );
                    j = 0;
                    while ( j < q->len && i < p->len ) 
                        {
                        w = vertex( q, j );

                        dx = abs( v.x - w.x );
                        dy = abs( v.y - w.y );

                        if ( dx <= bss && dy <= bss ) // kiss the bride
                            {
                            besti = i;
                            bestj = j;
                            v = vertex( p, besti );
                            w = vertex( q, bestj );
                         
                            // If we have already moved one of these vertices
                            // then that one should stay put. If we have already
                            // moved both vertices, then we should do nothing.

                            if ( ! p->precious[besti] && ! q->precious[bestj] )
                                {
                                // should we snap to v or w? 
                                // let's see which one creates the largest circumference

                                assign_vertex( p, besti, w );
                                assign_vertex( q, bestj, w );
                                cw = circumference( p ) + circumference( q );
                                assign_vertex( p, besti, v );
                                assign_vertex( q, bestj, v );
                                cv = circumference( p ) + circumference( q );

                                if ( cw > cv ) 
                                    {
                                    assign_vertex( q, bestj, w );
                                    assign_vertex( p, besti, w );
                                    }

                                p->precious[besti] = TRUE;
                                q->precious[bestj] = TRUE;
                                }
                            else if ( p->precious[besti] )
                                {
                                assign_vertex( q, bestj, v ); 
                                q->precious[bestj] = TRUE;
                                }
                            else if ( q->precious[bestj] )
                                {
                                assign_vertex( p, besti, w );
                                p->precious[besti] = TRUE; 
                                }

                            j = j + 1;  

                            }
                        j++;
                        }
                    i++;
                    }
                }

            qr = qr->next;
            while ( qr && ! qr->p->active ) qr = qr->next;
            }

        pr = pr->next;
        while ( pr && ! pr->p->active ) pr = pr->next;
        }
    


    //---------------------
    // roll through the list again, 
    // eliminating duplicate and colinear vertices.

    pr = tout->first;
    while ( pr && ! pr->p->active ) pr = pr->prev;

    printv( "removing duplicate and colinear vertices...\n" ); fflush(stdout);

    while ( pr )
        {
        p = pr->p;
        remove_duplicate_vertices( p );
        remove_colinear_vertices(  p, FALSE ); 
        snap_to_grid( p );
        if ( fix ) find_and_fix_self_crossing_polygon( p, tout ); //should not be done before darning the holes
        pr = pr->next;
        while ( pr && ! pr->p->active ) pr = pr->next;
        }
        
    }

//-----------------marry_vertices------------------------------------------------------------------




BOOL
near( xy v, xy w )
    {
    // w is near v if it is less than one beam step away

    static double vx, vy, wx, wy, fbss, dx, dy;
    
    vx = v.x;
    vy = v.y;
    wx = w.x;
    wy = w.y;
    dx = fabs( v.x - w.x );
    dy = fabs( v.y - w.y );
    fbss = bss;
    
    return( (sqrt( dx*dx + dy*dy ) < fbss) );
    
    }

// currently used:

void
weed_vertices( queue *tout )     // look for vertices which are closer than bss, then
    {                            // delete any that are too close
    polygon_ref *pr;
    polygon     *p;
    xy           v,   w;
    int          i,   j, range;
    double       dx,  dy;
    int          maxrange=10;

    pr = tout->first;
    while ( pr && ! pr->p->active ) pr = pr->next;

    while ( pr )
        {
        p = pr->p;
        unhit_all( p );
        range = maxrange > (p->len - 1) ? (p->len - 1) : maxrange;
        
        for ( i=0; i < p->len; i++ )
            {
            if ( ! is_hit( p, i ) )
                {
                v = vertex( p, i );
                for ( j=1; j < range; j++ )
                    {
                    if ( ((i+j) % p->len) != i )
                        {
                        w = vertex( p, i+j );
                        if ( near( v, w ) ) hit_vertex( p, i+j );
                        }
                    }
                }
            }

        i = 0;
        while ( i < p->len )
            {
            if ( is_hit( p, i ) )
                {
                unhit_vertex( p, i );
                delete_vertex( p, i );
                }
            else
                {
                i++;
                }
            }

        if ( p->len < 3 ) p->active = FALSE;
        
        pr = pr->next;
        while ( pr && ! pr->p->active ) pr = pr->next;
        }
            
    }

//-----------------weed_vertices------------------------------------------------------------------





//==============================================================================================================

int main( int argc, char *argv[] )
    {

    pid_t     pid;

    char     *ifilename,
	      ofilename[256];

    FILE     *abort;

    BOOL      fix,         // fix self-crossing polygons
              not_eof;

    int       // bss and bss2 are global
              fd,
              n,
              fdo,
              numpolys,
              layer,
              dt,
              i;

    double    bss_nm,
              dbu_uu,
              dbu_m,
              dbu_rescale,
              dbu_um,
              progress,
              was;

    unsigned short 
              count,           
              token;   

    polygon  *poly, *p;        

    polygon_ref *pr;

    struct gds_itemtype *item;

    queue    *tous;

    
    if ( argc == 2 && strcmp( argv[1], "-v" ) == 0 )
        {
        printf( "grid_snap v%1.1f\n", VERSION );
        exit(0);
        }


    if ( argc < 3 )
        {
        printf( "\n\ngrid_snap version %1.1f", VERSION );
        printf( "\nUse: grid_snap infile.gds outfile.gds bss_nm [fix|nofix] \n\n" );
        exit(0);
        }

    pid = getpid();
    ifilename = argv[1]; 

    sprintf( ofilename, "%s", argv[2] );
    sscanf( argv[3], "%lf", &bss_nm );

    fix = TRUE;
    if ( argc > 4 ) fix = (strcmp(argv[4], "nofix") != 0);
    if ( fix )
	printv( "\nalso fix self-crossing polygons\n" );
    else
	printv( "\ndo not fix self-crossing polygons\n" );

    //------- end command parse --------------------------

    int   ch1, ch2;
    char *sn1, *sn2, sn[10]="";

    if ( strncmp( argv[1], "moslice", 7 ) != 0 ) 
        {
        if ( strstr( ifilename, "_part_" ) )
            {
            ch1 = (int) '_';
            sn1 = rindex( ifilename, ch1 );
            ch2 = (int) '.';
            sn2 = rindex( ifilename, ch2 ) - 1;
            if ( sn1 && sn2 )
                {
                strncpy( sn, sn1+1, sn2 - sn1 );
                strcat(  sn, " " );
                //printf( "\015starting grid_snap part %s    ", sn );    
                }
            }
	}
    else
        {
        //printf( "poly2trap %s %s %1.2lf \n", ifilename, ofilename, bss_nm );
        }
        

    //abort = fopen( "fullstop.txt", "r" );
    //if ( abort != NULL ) { printf( "snap stops.\n" ); exit(0); } 
    //fclose( abort );

    //------------------------------------------ open the input file

    fd = open( ifilename, 0 );
    if ( fd <= 0 ) bail( "unable to open input file" );
    item = (struct gds_itemtype *) malloc( sizeof( struct gds_itemtype ) );
    if ( ! item ) bail( "unable to allocate gds item" );

    //------------------------------------------ read the polygons and push them onto the queue tous

    tous = initialize_queue();
    numpolys = 0;
    progress = 0.0;
    not_eof = read( fd, &count, 2 ) > 0;

    while ( not_eof )
        {
        gds_swap2bytes( (BYTE *) &count );
        if ( count >= 4 )
            {
            read( fd, &token, 2 );
            gds_swap2bytes( (BYTE *) &token );

            switch( token )
                {
                case 0x0305:  // UNITS
                          gds_read_units( fd, &dbu_um, &dbu_uu, &dbu_m, 0 );
                          dbu_rescale = 1.0e-9 / dbu_m ;
                          dbu_m  = dbu_m * dbu_rescale;
                          dbu_uu = dbu_m * 1.0e6;
                          dbu_um = dbu_m * 1.0e6; 
                          bss = round( bss_nm / (dbu_m * 1.0e9));   // beam step size in database units
                          if ( bss == 0 ) bss = 1;
                          bss2 = bss * 0.8;            // this is what we mean by "too close"
                          several_beam_steps = 4.0 * bss;
                          if ( bss2 == 0 ) bss2 = 1;
                          printv(   "Beam step, dbu:     %d \n\n", bss );
                          if ( dbu_m != 1e-9 ) bail( "Scaling the database unit should not be happening in grid_snap." );
                          break;
                case 0x2B02:  // printf( "\nPROPATTR means obese \n" ); not used
                          SKIPOVER( fd, count );
                          break;
                case 0x0D02:  //printf( "LAYER \n" ) ;
                          gds_read_layer( fd, count, &item, 0 );
                          layer = item->layer;
                          break;
                case 0x0E02:  //printf( "DATATYPE \n" ) ;
                          gds_read_datatype( fd, count, &item, 0 );
                          dt = item->dt;
                          break;
                case 0x1003:  //printf( "XY \n" ) ;

                          gds_read_xy( fd, count, 0, &item, 0 );

                          item->layer = layer;
                          item->dt = dt;
                          poly = copy_from_gds_polygon( item, dbu_rescale );
                          append_queue( tous, poly );
                          numpolys++;
                          break;
                default:  
                          SKIPOVER( fd, count );
                }    
            }      
        not_eof = read( fd, &count, 2 ) > 0;
        }          

    close(fd);


    // We used to look for vertices which are close to each other,
    // then set them equal so they will not snap apart, aka "marry".
    // We already did this in a separate program. It's 
    // important to marry the vertices before splitting
    // the pattern.

    //printv( "marry the vertices of %d polygons...\n", numpolys );
    //marry_vertices( tous, fix );

    // Instead, we will just do the snapping step found at the end of marry_vertices()
    
    printv( "removing duplicate and colinear vertices...\n" ); fflush(stdout);
    
    weed_vertices( tous );

    pr = tous->first;
    while ( pr && ! pr->p->active ) pr = pr->prev;

    while ( pr )
        {
        p = pr->p;
        remove_duplicate_vertices( p );
        remove_colinear_vertices(  p, FALSE ); 
        snap_to_grid( p );
        if ( fix ) find_and_fix_self_crossing_polygon( p, tous ); 
        pr = pr->next;
        while ( pr && ! pr->p->active ) pr = pr->next;
        }
    

    // open output file --------------------------------------

    fdo = open( ofilename, O_CREAT | O_TRUNC | O_WRONLY, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH );
    if ( fdo <= 0 ) ( "unable to open output gds file" );

    gds_create_lib_correctly( fdo, "traps", dbu_um, dbu_uu );
    gds_write_bgnstr( fdo );
    gds_write_strname( fdo, "traps" );


    // write output file and that's all -----------------------------

    
    write_trap_list( fdo, tous );
    gds_write_endstr( fdo ); 
    gds_write_endlib( fdo );    
    close( fdo );

    printv( "\ndone.\n" );

    }


