// library of polygon functions, with floating point vertices

#include <stdio.h>
#include <stdlib.h>

#ifndef BOOL
#define TRUE 1
#define FALSE 0
#define BOOL unsigned char
#endif

#ifndef BYTE
#define BYTE unsigned short int
#endif

#ifndef xy
typedef struct xy
    {
    double x;
    double y;
    }                     xy;
#endif

#ifndef polygon
typedef struct polygon
    {
    long len;   
    long max;   
    BOOL active;
    struct polygon *next;
    struct polygon *prev;
    long layer;
    long dt;
    xy ll;   // extents
    xy ur;
    xy *v;   // vertices
    BOOL obese;
    BOOL hit;
    BOOL *frap;
    }                     polygon;
#endif

#ifndef polygon_pair
typedef struct polygon_pair
    {
    polygon *a;
    polygon *b;
    }                     polygon_pair;
#endif

#ifndef long_list
typedef struct long_list
    {
    long len;
    long max;
    long *n;
    }                     long_list;
#endif

#ifndef queue
typedef struct queue
    {
    long len;
    polygon *first;
    polygon *last;
    }                     queue;
#endif

#ifndef polygon_ref
typedef struct polygon_ref
    {
    polygon *p;
    struct polygon_ref *next;
    struct polygon_ref *prev;
    }                     polygon_ref;
#endif

#ifndef separate_queue
typedef struct separate_queue
    {
    long len;
    polygon_ref *first;
    polygon_ref *last;
    }                     separate_queue;
#endif

#ifndef queue_ref
typedef struct queue_ref
    {
    separate_queue *q;
    BOOL active;
    struct queue_ref *next;
    struct queue_ref *prev;
    }                     queue_ref;
#endif

#ifndef queue_queue
typedef struct queue_queue
    {
    queue_ref *first;
    queue_ref *last;
    }                     queue_queue;
#endif


#ifndef lll
typedef struct lll
    {
    long n;
    struct lll *prev;
    struct lll *next;
    }                               lll;
#endif

#ifndef p_list
typedef struct p_list
    {
    lll *top;
    lll *bottom;
    }                               p_list;
#endif

#ifndef array2d_lll
typedef struct array2d_lll
    {
    long lenx;
    long leny;
    p_list **element;
    }                               array2d_lll;
#endif 
    
#ifndef array_2dq
typedef struct array_2dq
    {
    xy len;
    separate_queue ***q;
    }                               array_2dq;
#endif

void bail( char *message );

xy vertex( polygon *p, long i );

xy dvertex( polygon *p, long i, int d );

polygon *initialize_polygon( long n, long layer, long dt, BOOL obese );

queue *initialize_queue();

void append_polygon( polygon *p, double x, double y );

void append_polygon_xy( polygon *p, xy v );

void append_queue( queue *q, polygon *p );

polygon *pop_queue( queue *q );

polygon *pop_queue_fifo( queue *q );

long_list *initialize_long_list( long n );

void assign_vertex( polygon *p, long i, xy v );

void delete_vertex( polygon *p, long gone );

void print_polygon( polygon *p, double scale );

polygon *copy_polygon( polygon *p );

polygon *copy_flip_polygon( polygon *p );

void copy_into_polygon( polygon *from, polygon *to );

void reverse_polygon( polygon *p );

void multiply_polygon( polygon *p, double f );

queue *initialize_queue();

array2d_lll *initialize_array2d_lll( long nx, long ny );

void append_array2d( array2d_lll *list, long x, long y, long n );

lll *element_of_array2d( array2d_lll *list, long x, long y );

lll *initialize_lll();

lll *prepend_lll( lll *p, long n );

lll *append_lll( lll *p, long n );

lll *remove_from_lll( lll *p, lll *q );

void print_lll( lll *p );

void print_array2d( array2d_lll *list, long x, long y );

BOOL **initialize_bool( long r, long c );            

array_2dq *initialize_2dq( long nr, long nc );

separate_queue *get_field_list( array_2dq *a, long x, long y );

void append_separate_queue( separate_queue *q, polygon *p );

separate_queue *initialize_separate_queue();

void append_queue_queue( queue_queue *qq, separate_queue *q );

queue_queue *initialize_queue_queue();

separate_queue *prepend_separate_queue( separate_queue *q, polygon *p );

void free_separate_queue( separate_queue *q );

void zero_polygon( polygon *p );

void set_all_false( BOOL **b, long r, long c );

void set_vertex( polygon *p, long i, xy v );
