
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>

                            // Setting BYTESWAP to 1 is appropriate for big-endian Intel processors. 
                            // GDS format was originally used on little-endian, older computers.
#define BYTESWAP 1

#ifndef BYTE
#define BYTE unsigned char
#endif

#ifndef BOOL
#define BOOL  unsigned char
#define TRUE  1
#define FALSE 0
#endif

BYTE  gdsswap;
short gdsword;

#define SKIPOVER( fd, count )  { for ( i=0; i < ((count)-4); i+=2 ) read( (fd), &gdsword, 2 ); }
#define BAILOUT( message )     { printf( "\n\nERROR: %s\n\n", (message) ); fflush(stdout); exit(-1); }
#define ATTENTION( message )   { printf( "\n                            WARNING: %s\n\n", (message) ); fflush(stdout); }
 

struct gds_itemtype
    {                               
    int   n;                         /* number of verticies                                                       */
    int   layer;                     /* layer                                                                     */
    int   dt;                        /* datatype                                                                  */ 
    int  *x;                         /* array of x coordinates or possibly just the reference point X             */
    int  *y;                         /* array of y coordinates or possibly just the reference point Y             */
    BOOL *hit;                       /* used for chopping up polygons                                             */
    struct gds_itemtype *nextitem;       
    };                               
                                       

struct gds_celltype                  /* A GDS library is a linked list of cells.                                  */
    {                                /* A cell is a linked list of items.                                         */
    char *name;                      /* name of the cell                                                          */
    struct gds_itemtype *item;       /* one element of the cell                                                   */
    struct gds_celltype *nextcell;   /* pointer to the next cell, forming a linked list                           */  
    };


char *gds_read_libname( int fd, int count, BOOL verbose );
void  gds_read_header( int fd, int count, BOOL verbose );
void  gds_swap4bytes( BYTE *four  );
void  gds_swap2bytes( BYTE *two );
void  gds_read_header( int fd, int count, BOOL verbose );
void  gds_write_header( int fd );
void  gds_read_bgnlib( int  fd, int count, int *cell_table_size, struct gds_celltype **lib, 
                       struct gds_celltype ***c, struct gds_celltype **cc,  BOOL verbose );
void  gds_write_bgnlib( int fd );
void  gds_read_bgnstr( int fd, int count, int *cell_table_size, int *num_cells, struct gds_celltype ***c, 
                       struct gds_celltype **cc, struct gds_itemtype **ci, BOOL verbose );
void  gds_write_bgnstr( int fd );
void  gds_write_node( int fd );
void  gds_read_endlib( int fd, int count, BOOL verbose );
void  gds_write_endlib( int fd );
void  gds_read_endstr( int fd, int count, BOOL verbose );
void  gds_write_endstr( int fd );
void  gds_write_libname( int fd, char *name );
void  gds_read_strname( int fd, int count, int num_cells, struct gds_celltype ***c, 
                        struct gds_celltype **cc, BOOL verbose );
void  gds_write_strname( int fd, char *name );
void  gds_read_sname( int fd, int count );
void  gds_write_sname( int fd, char *s );
void  gds_read_boundary( int fd, int count, struct gds_itemtype **ci, BOOL verbose );
void  gds_write_boundary( int fd );
void  gds_read_endel( int fd, int count, BOOL verbose );
void  gds_write_endel( int fd );
void  gds_read_layer( int fd, int count, struct gds_itemtype **ci, BOOL verbose );
void  gds_write_layer( int fd, short int layer );
void  gds_read_datatype( int fd, int count, struct gds_itemtype **ci, BOOL verbose );
void  gds_write_datatype( int fd, short int dt );
void  gds_read_xy( int fd, int count, double dbu_um, struct gds_itemtype **ci, BOOL verbose );
void  gds_write_xy( int fd, int *x, int *y, int n );
void  gds_write_long_xy( int fd, long *x, long *y, long n );
void  gds_read_units( int fd, double *Pdbu_um, double *Pdbu_uu, double *Pdbu_m, BOOL verbose );
void  gds_write_units( int fd, double dbu_uu, double dbu_m );
void  gds_create_lib( int fd, char *libname, double dbu_um );    
void  gds_create_lib_correctly( int fd, char *libname, double dbu_um, double dbu_uu );
void  gds_write_obese( int fd );
void  gds_make_next_item( struct gds_itemtype **ci );
