// library of polygon functions, with integer vertices
// This version uses longs for xy coordinates.

// v2  now with mv_polygon

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define TRUE 1
#define FALSE 0
#define BOOL unsigned char

#ifndef BYTE
#define BYTE unsigned short int
#endif

#ifndef xy
typedef struct xy
    {
    long x;
    long y;
    }                     xy;
#endif

#ifndef dxy
typedef struct dxy
    {
    double x;
    double y;
    }                     dxy;
#endif


#ifndef polygon
typedef struct polygon
    {
    int   len;   
    int   max;   
    BOOL  active;
    int   layer;
    int   dt;
    xy    ll;   // extents
    xy    ur;
    xy    *v;   // vertices
    BOOL  obese;
    BOOL *precious;
    BOOL *hit;       // a synonym for precious
    BOOL *hit2;     
    }                     polygon;
#endif


#ifndef polygon_array
typedef struct polygon_array
    {
    polygon **par;
    int       len;
    }                     polygon_array;
#endif

#ifndef polygon_pair
typedef struct polygon_pair
    {
    polygon *a;
    polygon *b;
    }                     polygon_pair;
#endif

#ifndef int_list
typedef struct int_list
    {
    int  len;
    int  max;
    int *n;
    }                     int_list;
#endif


#ifndef polygon_ref
typedef struct polygon_ref
    {
    polygon *p;
    struct polygon_ref *next;
    struct polygon_ref *prev;
    }                     polygon_ref;
#endif


#ifndef queue
typedef struct queue
    {
    int len;
    polygon_ref *first;
    polygon_ref *last;
    }                     queue;
#endif


#ifndef queue_ref
typedef struct queue_ref
    {
    queue *q;
    BOOL active;
    struct queue_ref *next;
    struct queue_ref *prev;
    }                     queue_ref;
#endif

#ifndef queue_queue
typedef struct queue_queue
    {
    queue_ref *first;
    queue_ref *last;
    }                     queue_queue;
#endif

#ifndef ill
typedef struct ill
    {
    int n;
    struct ill *prev;
    struct ill *next;
    }                               ill;
#endif

#ifndef p_list
typedef struct p_list
    {
    ill *top;
    ill *bottom;
    }                               p_list;
#endif

#ifndef array2d_ill
typedef struct array2d_ill
    {
    int lenx;
    int leny;
    p_list **element;
    }                               array2d_ill;
#endif 
    
#ifndef array_2dq
typedef struct array_2dq
    {
    xy len;
    queue ***q;
    }                               array_2dq;
#endif

void bail( char *message );

xy vertex( polygon *p, int i );

dxy dvertex( polygon *p, int i );

polygon *initialize_polygon( int n, int layer, int dt, BOOL obese );

polygon_pair *initialize_polygon_pair();

queue *initialize_queue();

void append_polygon( polygon *p, long x, long y );

void append_polygon_xy( polygon *p, xy v );

void append_polygon_dxy( polygon *p, dxy v );

polygon_ref *append_queue( queue *q, polygon *p );

int_list *initialize_int_list( int n );

void append_int_list( int_list *list, int layer );

BOOL in_int_list( int_list *p, int x );

void assign_vertex( polygon *p, int i, xy v );

BOOL delete_vertex( polygon *p, int gone );

void print_polygon( polygon *p, float scale );

void printn_polygon( polygon *p, float scale, int n );

void mv_polygon( polygon *from, polygon *to );

polygon *copy_polygon( polygon *p );

polygon *copy_flip_polygon( polygon *p );

void reverse_polygon( polygon *p );

void multiply_polygon( polygon *p, float f );

queue *initialize_queue();

array2d_ill *initialize_array2d_ill( int nx, int ny );

void append_array2d( array2d_ill *list, int x, int y, int n );

ill *element_of_array2d( array2d_ill *list, int x, int y );

ill *initialize_ill();

ill *append_ill( ill *p, int n );

ill *remove_from_ill( ill *p, ill *q );

void print_ill( ill *p );

void print_array2d( array2d_ill *list, int x, int y );

BOOL **initialize_bool( int r, int c );            

array_2dq *initialize_2dq( int nr, int nc );

queue *get_field_list( array_2dq *a, int x, int y );

queue *get_field_list_db( array_2dq *a, int x, int y, int location );

queue *initialize_queue();

void append_queue_queue( queue_queue *qq, queue *q );

queue_queue *initialize_queue_queue();

void free_queue( queue *q );

void zero_polygon( polygon *p );

void set_all_false( BOOL **b, int r, int c );

void set_vertex( polygon *p, int i, xy v );

void set_dvertex( polygon *p, int i, dxy v );

polygon *pop_queue_fifo( queue *q );

int insert_vertex( xy v, polygon_ref *pr, int jj );

BOOL insert_vertex_near( xy v, polygon_ref *pr, int jj, int bss );

void unhit_all( polygon *p );

void hit_vertex( polygon *p, int i );

void unhit_vertex( polygon *p, int i );

BOOL delete_vertex( polygon *p, int i );

BOOL delete_precious_vertex( polygon *p, int i );

BOOL polygon_ccw( polygon *p, double *area, int bss );

double polygon_area( polygon *p, int bss );

void delete_queue_member( queue *q, polygon_ref *pr );

void set_extents( polygon *p );

xy previous_vertex( polygon *p, int i );

xy next_vertex( polygon *p, int i );

void free_polygon( polygon *p );

polygon_array *initialize_polygon_array( int np, int n );

void unhit_polygon_array( polygon_array *pa );

void just_copy_polygon( polygon *from, polygon_ref *to );

polygon_ref *initialize_polygon_ref();

void zero_queue( queue *q );

void no_precious( polygon *p );

BOOL is_hit( polygon *p, int i );
