// gpfree
//         Functions for reading and writing Raith gpf files.
//         Just like the Beams library, but with open-source software.
//
//         Here we assume you are using a big-endian machine, such as
//         an Intel x86 or AMD. If you are using a PDP-11 or a VAX,
//         that's wonderful but you're out of luck.

#include "gpf_pub.h"
#include "polyq.h"
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>


//--------------------------------------------------------------------------------------------------------

FILE *
gpfree_file_open( char *name, char *mode )  
    {
    // replaces gpf_file_open( FILE *f, char *name, char *mode );
    // Who knows what that input file pointer was for. Weird.
    
    return( fopen( name, mode ) );
    }


void
gpfree_file_close( FILE *fp )
    {
    // replaces gpf_file_close() which seems pretty silly
    
    fclose( fp );
    }


//--------------------------------------------------------------------------------------------------------

void
gpfree_write_union( FILE *fp, gpf_ulong_t *gpfpos, gpf_union_t *record )
    {
    // replaces gpf_write_union( fp, gpfpos, record );
    // Opening the file does not position the write head
    // below the header. We need to use lseek to skip ahead.

    static char *brecord, temp[4];
    brecord = (char *) record;
    temp[0] = brecord[3];       // The ghost of VMS!
    temp[1] = brecord[2]; 
    temp[2] = brecord[1]; 
    temp[3] = brecord[0]; 
    lseek( fileno(fp), *gpfpos, SEEK_SET );
    write( fileno(fp), temp, 4 );

    // fwrite( temp, 4, 1, fp ); does not work
    
    *gpfpos += 4;

    }

//--------------------------------------------------------------------------------------------------------

gpf_double_t
flip_double( gpf_double_t x )   // The VAX was little-endian, 
    {                           // and so gpf is also. Did gpf start
    static char *b, flip[8];    // on the PDP-11 ? Also little-endian.
    static gpf_double_t *f;
    b = (char *) &x;
    flip[7] = b[0];
    flip[6] = b[1];
    flip[5] = b[2];
    flip[4] = b[3];
    flip[3] = b[4];
    flip[2] = b[5];
    flip[1] = b[6];
    flip[0] = b[7];
    f = (gpf_double_t *) flip;
    return( *f );
    }


gpf_ulong_t
flip_long( gpf_ulong_t x )
    {                       
    static char *b, flip[8];
    static gpf_ulong_t *f;
    b = (char *) &x;
    flip[7] = b[0];
    flip[6] = b[1];
    flip[5] = b[2];
    flip[4] = b[3];
    flip[3] = b[4];
    flip[2] = b[5];
    flip[1] = b[6];
    flip[0] = b[7];
    f = (gpf_ulong_t *) flip;
    return( *f );
    }


gpf_uint_t
flip_int( gpf_uint_t x )
    {
    static char *b, flip[4];
    static gpf_uint_t *f;
    b = (char *) &x;
    flip[3] = b[0];
    flip[2] = b[1];
    flip[1] = b[2];
    flip[0] = b[3];
    f = (gpf_uint_t *) flip;
    return( *f );
    }


//--------------------------------------------------------------------------------------------------------

void gpfree_write_indexf( FILE *fp, gpf_ulong_t gpfpos, gpf_indexf_t *index )
    {
    // replaces
    // long gpf_write_indexf( FILE *fp, gpf_ulong_t start, gpf_ulong_t block, gpf_indexf_t *index );
    // We don't know what a 'block' is, but it's always zero apparently.
    // We don't know what the return value might be. Maybe it's the number of bytes written,
    // but who cares because it's not used anyway.

    index->FieldToPatternOffsetX = flip_double( index->FieldToPatternOffsetX );
    index->FieldToPatternOffsetY = flip_double( index->FieldToPatternOffsetY );
    index->StartData             = flip_long(   index->StartData             );
    index->EndData               = flip_long(   index->EndData               );
    index->Flags                 = flip_long(   index->Flags                 );

    lseek( fileno(fp), gpfpos, SEEK_SET );
    write( fileno(fp), index, sizeof( gpf_indexf_t ) );

    }

//--------------------------------------------------------------------------------------------------------

void
gpfree_write_header( FILE *fp, gpf_header_t *header, gpf_header_user_t *user )
    {
    // replaces gpf_write_header()
    
    header->MainFieldResolutionX = flip_double( header->MainFieldResolutionX );
    header->MainFieldResolutionY = flip_double( header->MainFieldResolutionY );
    header->SubFieldResolutionX  = flip_double( header->SubFieldResolutionX  );
    header->SubFieldResolutionY  = flip_double( header->SubFieldResolutionY  );
    header->ResolutionX          = flip_int(    header->ResolutionX          );
    header->ResolutionY          = flip_int(    header->ResolutionY          );
    header->BeamStepSizeX        = flip_int(    header->BeamStepSizeX        );
    header->BeamStepSizeY        = flip_int(    header->BeamStepSizeY        );
    header->PatternSizeX         = flip_long(   header->PatternSizeX         );
    header->PatternSizeY         = flip_long(   header->PatternSizeY         );
    header->MainFieldSizeX       = flip_int(    header->MainFieldSizeX       );
    header->MainFieldSizeY       = flip_int(    header->MainFieldSizeY       );
    header->SubFieldSizeX        = flip_int(    header->SubFieldSizeX        );
    header->SubFieldSizeY        = flip_int(    header->SubFieldSizeY        );
    header->HighTension          = flip_int(    header->HighTension          );
    header->ShapeTypes           = flip_int(    header->ShapeTypes           );
    header->SubFieldPlacement    = flip_int(    header->SubFieldPlacement    );
    header->NrMainFields         = flip_int(    header->NrMainFields         );
    header->MainFieldPlacement   = flip_int(    header->MainFieldPlacement   );
    header->FractureStyle        = flip_int(    header->FractureStyle        );
    header->NrMainFieldBits      = flip_int(    header->NrMainFieldBits      );
    header->NrSubFieldBits       = flip_int(    header->NrSubFieldBits       );
    header->WorkLevel            = flip_int(    header->WorkLevel            );
    header->MinFreqFactor        = flip_int(    header->MinFreqFactor        );
    header->MaxFreqFactor        = flip_int(    header->MaxFreqFactor        );
    header->NrFreqFactors        = flip_int(    header->NrFreqFactors        );
    header->PixelTime            = flip_long(   header->PixelTime            );
    header->FileSize             = flip_long(   header->FileSize             );
    header->HeaderOffset         = flip_long(   header->HeaderOffset         );
    header->IndexOffset          = flip_long(   header->IndexOffset          );
    header->StructureOffset      = flip_long(   header->StructureOffset      );
    header->SequenceSize         = flip_long(   header->SequenceSize         );
    header->MarkerOffset         = flip_long(   header->MarkerOffset         );
    header->StructureSize        = flip_long(   header->StructureSize        );
    header->MarkerSize           = flip_long(   header->MarkerSize           );
    header->RefFreqFactor        = flip_int(    header->RefFreqFactor        );
    
    lseek( fileno(fp), 0, SEEK_SET );
    write( fileno(fp), header, GPF_HEADER_SIZE );
    lseek( fileno(fp), GPF_HEADER_SIZE, SEEK_SET );
    write( fileno(fp), user, GPF_HEADER_USER_SIZE );
    }


//--------------------------------------------------------------------------------------------------------

//--------------------------------------------------------------------------------------------------------

void
gpfree_read_union( FILE *fp, gpf_ulong_t *gpfpos, gpf_union_t *record )
    {
    // replaces gpf_read_union( fp, gpfpos, record );

    static char *b, temp[4];
    
    lseek( fileno(fp), *gpfpos, SEEK_SET );
    read(  fileno(fp), temp, 4 );

    b = (char *) record;
    
    b[3] = temp[0];    // the ghost of VMS!
    b[2] = temp[1];
    b[1] = temp[2];
    b[0] = temp[3];

    *gpfpos += 4;

    }

//--------------------------------------------------------------------------------------------------------

void gpfree_read_indexf( FILE *fp, gpf_ulong_t start, gpf_ulong_t block, gpf_indexf_t *index )
    {
    // replaces
    // long gpf_read_indexf( FILE *fp, gpf_ulong_t start, gpf_ulong_t block, gpf_indexf_t *index );
    // where block is the field number. In a fit of inconsistency, we are not updating 'start'
    // after each read; rather, we must add block * sizeof(indexf). That's how Raith does it, go figure.

    static gpf_ulong_t gpfpos;

    gpfpos = start + block * sizeof(gpf_indexf_t);
    
    lseek( fileno(fp), gpfpos, SEEK_SET );
    read(  fileno(fp), index, sizeof( gpf_indexf_t ) );

    index->FieldToPatternOffsetX = flip_double( index->FieldToPatternOffsetX );
    index->FieldToPatternOffsetY = flip_double( index->FieldToPatternOffsetY );
    index->StartData             = flip_long(   index->StartData             );
    index->EndData               = flip_long(   index->EndData               );
    index->Flags                 = flip_long(   index->Flags                 );

    }

//--------------------------------------------------------------------------------------------------------

void
gpfree_read_header( FILE *fp, gpf_header_t *header, gpf_header_user_t *user )
    {
    // replaces gpf_read_header()
    
    lseek( fileno(fp), 0, SEEK_SET );
    read(  fileno(fp), header, sizeof( gpf_header_t ) );

    lseek( fileno(fp), GPF_HEADER_SIZE, SEEK_SET );
    read(  fileno(fp), user, sizeof( gpf_header_user_t ) );

    header->MainFieldResolutionX = flip_double( header->MainFieldResolutionX );
    header->MainFieldResolutionY = flip_double( header->MainFieldResolutionY );
    header->SubFieldResolutionX  = flip_double( header->SubFieldResolutionX  );
    header->SubFieldResolutionY  = flip_double( header->SubFieldResolutionY  );
    header->ResolutionX          = flip_int(    header->ResolutionX          );
    header->ResolutionY          = flip_int(    header->ResolutionY          );
    header->BeamStepSizeX        = flip_int(    header->BeamStepSizeX        );
    header->BeamStepSizeY        = flip_int(    header->BeamStepSizeY        );
    header->PatternSizeX         = flip_long(   header->PatternSizeX         );
    header->PatternSizeY         = flip_long(   header->PatternSizeY         );
    header->MainFieldSizeX       = flip_int(    header->MainFieldSizeX       );
    header->MainFieldSizeY       = flip_int(    header->MainFieldSizeY       );
    header->SubFieldSizeX        = flip_int(    header->SubFieldSizeX        );
    header->SubFieldSizeY        = flip_int(    header->SubFieldSizeY        );
    header->HighTension          = flip_int(    header->HighTension          );
    header->ShapeTypes           = flip_int(    header->ShapeTypes           );
    header->SubFieldPlacement    = flip_int(    header->SubFieldPlacement    );
    header->NrMainFields         = flip_int(    header->NrMainFields         );
    header->MainFieldPlacement   = flip_int(    header->MainFieldPlacement   );
    header->FractureStyle        = flip_int(    header->FractureStyle        );
    header->NrMainFieldBits      = flip_int(    header->NrMainFieldBits      );
    header->NrSubFieldBits       = flip_int(    header->NrSubFieldBits       );
    header->WorkLevel            = flip_int(    header->WorkLevel            );
    header->MinFreqFactor        = flip_int(    header->MinFreqFactor        );
    header->MaxFreqFactor        = flip_int(    header->MaxFreqFactor        );
    header->NrFreqFactors        = flip_int(    header->NrFreqFactors        );
    header->PixelTime            = flip_long(   header->PixelTime            );
    header->FileSize             = flip_long(   header->FileSize             );
    header->HeaderOffset         = flip_long(   header->HeaderOffset         );
    header->IndexOffset          = flip_long(   header->IndexOffset          );
    header->StructureOffset      = flip_long(   header->StructureOffset      );
    header->SequenceSize         = flip_long(   header->SequenceSize         );
    header->MarkerOffset         = flip_long(   header->MarkerOffset         );
    header->StructureSize        = flip_long(   header->StructureSize        );
    header->MarkerSize           = flip_long(   header->MarkerSize           );
    header->RefFreqFactor        = flip_int(    header->RefFreqFactor        );

    // printf( "\n\n%s %s\n\n", header->Title, header->Version );
    
    }


//----------------------------------------------------------------------------------------
