// In this version, we have turned off "cut_long_thing", which chops up 
// high aspect ratio shapes.

// In this version we turn off the snap and colinear elimination step
// when finding a cut. Also the calculation
// of the intersection point uses "floor" and proper floating point math.

// v2.0  Uses longer polygons, since we are now using convexify
//       instead of poly2trap. cutline() now allocates arrays that
//       can handle long polygons.

// v2.1  Includes option 'noheader' for gds output

// v2.2  nearly() is now == and we got rid of vertical and horizontal
//       snapping. That snap was good for parallelograms but is 
//       not necessary for polygons. It makes them worse.

// v2.3  cut_triangular_sliver now *does not* snap to grid. This was causing
//       triangles to snap down one pixel. This function now checks for
//       degens, declaring success if the triangle is degenerate.

// v2.4  cut_triangular_sliver is now called only under dire circumstances,
//       not routinely for every triangle. Are 100:1 triangles a problem
//       anymore? Maybe not.

// v2.5  snapping the cuts to the bss grid does improve things a little
//       but we are still getting a lot of out-of-bounds shapes

// v2.6  so many problems with large rectangles

// V2.7  noarray option surppresses rectangle arrays. You need this
//       for overlap removal, since we can't have shapes spanning subfields.

// v2.8  xy type is now long
// v2.9  correction to snapoint()
// v3.0  many small corrections - deleted many unused variables


#define VERSION 3.0

#include "libgdsmin.h"
#include "polyq.h"
#include <string.h>
#include <strings.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>

double bss, bssr, bssm, bss2, bsst2;
int    bssri;
double nearness, presque, tooclose, equality;
BOOL   verbose=FALSE;
pid_t  pid;
int    num_out=0;
char  *ifilename;

#define EQUAL nearly
#define TOOCLOSE( a ) ((fabs((a)) < tooclose))
#define BSSFACTOR 10.0

#define printv noop

#define ERRORLAYER 255
#define MAXMARRY 25000
#define BSSR 1.0
#define BSSM 1.05    

void noop(){}
void remove_colinear_vertices( polygon *p, BOOL severe );
BOOL shape_extents_unknown( polygon *p );
void set_shape_extents( polygon *p );
BOOL between( int a, int b, int c );


//--------------------------------------------------------------------------------------------------

double
frac( double x )
    {
    static double f;
    f = (x - floor( x ));
    if ( fabs( f - 1.0 ) < 1e-6 ) f = 0.0;
    return f;
    }

//---------------------------------------------------------------------------------------------------

BOOL
degenerate( polygon *p )
    {
    return( (p->ll.x == p->ur.x) || (p->ll.y == p->ur.y) );
    }

//---------------------------------------------------------------------------------------------------

BOOL
samepoint( xy a, xy b )
    {
    //return( (abs(a.x - b.x) < bssr) && (abs(a.y - b.y) < bssr) );  

    return( (a.x == b.x) && (a.y == b.y) );
    }

BOOL
exact_samepoint( xy a, xy b )
    {
    return( (a.x == b.x) && (a.y == b.y) );
    }

BOOL
nearly( int a, int b )
    {
    return( a == b );
    }

BOOL
fequal( double a, double b )
    {
    return( fabs( a - b ) < 1e-6 );
    }

//---------------------------------------------------------------------------------------------------         

void
nag( char *message, int *count )
    {
    *count += 1;
    if ( *count < 10 )
        {
        if ( *count == 1 ) printf( "\n\n" );
        printf( "WARNING: %s\n", message );
        }
    else if ( *count == 10 )
        {
        printf( "*** REPEATED WARNING MESSAGES SUPPRESSED\n\n" );
        }
    fflush( stdout );
    }


//--------------------------------------------------------------------------------------------------------

BOOL
dequal( double a, double b )   // are two doubles roughly equal?
    {
    if ( a == b ) return( TRUE );

    if ( a == 0.0 )
        return( fabs(b) < 1.0e-7 );
    else if ( b == 0.0 )
        return( fabs(a) < 1.0e-7 );
    else
        return( fabs( (a - b)/a ) < 1.0e-7 );
    }


BOOL
proche( double x, int i )   
    {
    static double di, fbss;
    di = i;
    fbss = bss;
    return( fabs( x - di ) < fbss );
    }


BOOL
entre( double x, int a, int b )       // est-ce que x entre a et b?
    {
    static double fa, fb;

    if ( proche( x, a ) ) return( TRUE );
    if ( proche( x, b ) ) return( TRUE );

    fa = a;
    fb = b;

    if ( a < b )  
        return( x >= fa && x <= fb );   
    else
        return( x >= fb && x <= fa );
    }


//-------------------------------------------------------------------------------------------------------------


BOOL
flon_segment( double x, double y, xy a1, xy a2, xy b1, xy b2 )
    {

    // Is the point x,y somewhere on both line segments a1-a2 and b1-b2 ?

    static BOOL cross;

    // If a1-a2 is the same as b1-b2 then they definitely cross, and the intersection
    // is definitely on the segment. If overlaps are bad, then we should think about
    // that elsewhere.

    if ( (samepoint(a1, b1) && samepoint(a2, b2)) || (samepoint(a1, b2) && samepoint(a2, b1)) ) return( TRUE );


    if ( a1.y == a2.y ) 
        cross = entre( x, a1.x, a2.x ) && entre( y, b1.y, b2.y );
    else if ( b1.y == b2.y )
        cross = entre( x, b1.x, b2.x ) && entre( y, a1.y, a2.y );
    else if ( a1.x == a2.x )
        cross = entre( x, b1.x, b2.x ) && entre( y, a1.y, a2.y );
    else if ( b1.x == b2.x )
        cross = entre( x, a1.x, a2.x ) && entre( y, b1.y, b2.y );
    else
        cross = entre( x, b1.x, b2.x ) && entre( y, b1.y, b2.y ) &&
                entre( x, a1.x, a2.x ) && entre( y, a1.y, a2.y );

    return( cross );

    }  // flon_segment


//--------------------------------------------------------------------------------------------

BOOL
not_crossing( polygon *p, int a, int b, int c, int d, xy *intersection, BOOL overlapsbad )
    {
    // Does the segment a-b cross the segment c-d?
    // Return True if they do not cross, and return the 
    // intersection point if they do. 
    
    // If the line a-b is the same as the line c-d then
    // these line segments are still "crossing", because
    // doughnut polygons must be eliminated. The program
    // darn.c should have eliminated these sort of holes, 
    // but we must make sure here.

    // If we are making a slice, then we should NOT consider
    // coincident lines to be crossing. It's ok to slice
    // along one of the sides. Hence the logical overlapsbad.

    static xy
        a1, a2, b1, b2;

    static BOOL
        crossing,
        a_is_vertical,
        b_is_vertical,
        ma_equals_mb,
        ba_equals_bb;

    static double
        fb1x, fb1y,
        fb2x, fb2y,
        fa1x, fa1y,
        fa2x, fa2y,
        fx, fy,
        ma, mb, 
        ba, bb;

    static int
        ix, iy;

    a1 = vertex( p, a );
    a2 = vertex( p, b );
    b1 = vertex( p, c );
    b2 = vertex( p, d );

    intersection->x = 0;
    intersection->y = 0;

    // first check that the segments are anywhere near each other

    if ( a1.x < b1.x && a1.x < b2.x && a2.x < b1.x && a2.x < b2.x ) return( TRUE );
    if ( a1.x > b1.x && a1.x > b2.x && a2.x > b1.x && a2.x > b2.x ) return( TRUE );
    if ( a1.y < b1.y && a1.y < b2.y && a2.y < b1.y && a2.y < b2.y ) return( TRUE );
    if ( a1.y > b1.y && a1.y > b2.y && a2.y > b1.y && a2.y > b2.y ) return( TRUE );

    // overlaps are bad when we are looking for self-intersections or doughnut shapes.
    // overlaps are not bad when we are slicing.

    if ( ( samepoint( a1, b1 ) && samepoint( a2, b2 ) ) || ( samepoint( a1, b2 ) && samepoint( a2, b1 ) ) )
        {
        intersection->x = a1.x;
        intersection->y = a1.y;
        if ( overlapsbad )
            {
            //printf( "matching segment - probably a doughnut\n" );
            return( FALSE ); // they are crossing, probably a doughnut shape
            }
        else
            return( TRUE );  // they are the same line segment, but this is ok
        }


    if ( samepoint( a1, b1 ) || samepoint( a1, b2 ) )
        {
        intersection->x = a1.x;
        intersection->y = a1.y;
        if ( overlapsbad )
            {
            //printf( "lines touch at an end point\n" ); 
            return( FALSE ); // => crossing
            }
        else
            return( TRUE );
        }

    if ( samepoint( a2, b1 ) || samepoint( a2, b2 ) )  
        {
        intersection->x = a2.x;
        intersection->y = a2.y;
        if ( overlapsbad )
            {
            //printf( "lines touch at an end point\n" ); 
            return( FALSE );
            }
        else
            return( TRUE );
        }

    a_is_vertical = (a1.x == a2.x); 
    b_is_vertical = (b1.x == b2.x); 

    if ( a_is_vertical && b_is_vertical && ! overlapsbad ) return( TRUE );

    if ( a_is_vertical && b_is_vertical )  // and overlaps are bad...
        {
        //printf( "both line segments are vertical\n" ); 
        if ( a1.x == b1.x )
            {
            if ( between( a1.y, b1.y, b2.y ) )
                {
                //printf( "vertical segments overlap\n" );
                intersection->x = a1.x;
                intersection->y = a1.y;
                return( FALSE );
                }
            else if ( between( a2.y, b1.y, b2.y ) )
                {
                //printf( "vertical segments overlap\n" );
                intersection->x = a2.x;
                intersection->y = a2.y;
                return( FALSE );
                }
            else
                {
                return( TRUE );
                }
            }
        else
            {
            return( TRUE );
            }
        }

    if ( a_is_vertical && ! b_is_vertical )
        {
        //printf( "a is vertical but b is not\n" );

        fb1y = b1.y;
        fb2y = b2.y;
        fb1x = b1.x;
        fb2x = b2.x;

        mb = (fb1y - fb2y) / (fb1x - fb2x);
        bb = fb1y - mb * fb1x;

        fx = a1.x;
        fy = mb * fx + bb;

        // fx,fy are floats which have not snapped to the bss grid.
        // Is that what we want? Yes, probably.

        crossing = flon_segment( fx, fy, a1, a2, b1, b2 );

        if ( ! crossing )
            {
            intersection->x = 0;
            intersection->y = 0;
            return( TRUE );
            }
        else
            {
            //printf( "segments a and b cross. a is vertical, b is not\n" );
            intersection->x = a1.x; //snap( a1.x );
            iy = round( fy );   
            intersection->y = iy; //snap( iy );
            return( FALSE );
            }
        }
        

    if ( b_is_vertical && ! a_is_vertical )
        {
        //printf( "b is vertical but a is not\n" );

        fa1y = a1.y;
        fa2y = a2.y;
        fa1x = a1.x;
        fa2x = a2.x;

        ma = (fa1y - fa2y) / (fa1x - fa2x);
        ba = fa1y - ma * fa1x;

        fx = b1.x;
        fy = ma * fx + ba;

        // fx,fy are floats which have not snapped to the bss grid.
        // Is that what we want? Yes, probably.

        crossing = flon_segment( fx, fy, a1, a2, b1, b2 );

        if ( ! crossing )
            {
            intersection->x = 0;
            intersection->y = 0;
            return( TRUE );
            }
        else
            {
            //printf( "segments a and b cross. b is vertical, a is not\n" );
            intersection->x = b1.x; //snap( b1.x );
            iy = round( fy ); 
            intersection->y = iy; //snap( iy );
            return( FALSE );
            }
        }

    // neither segment is vertical
        
    fa1x = a1.x;  fa1y = a1.y;
    fa2x = a2.x;  fa2y = a2.y;
    fb1x = b1.x;  fb1y = b1.y;
    fb2x = b2.x;  fb2y = b2.y;

    ma = (fa1y - fa2y) / (fa1x - fa2x);
    mb = (fb1y - fb2y) / (fb1x - fb2x);

    bb = b1.y - mb * b1.x;
    ba = a1.y - ma * a1.x;

    ma_equals_mb = dequal( ma, mb );
    ba_equals_bb = dequal( ba, bb );

    if ( ma_equals_mb && ! ba_equals_bb )
        {
        intersection->x = 0;
        intersection->y = 0;
        return( TRUE );      // they are not crossing
        }

    if ( ma_equals_mb )  // Coincident lines, but do the segments overlap?
        {                // We already considered the case where the segments
                         // are the same, so this is not a doughnut cut.
                         // Also, it is not a slice cut, since slice lines
                         // are always on-axis.

        //printf( "coincident lines\n" );

        if ( ma > 1.0 )
            {
            if ( between( a2.y, b1.y, b2.y ) )
                {
                //printf( "off-axis segments that overlap\n" );
                intersection->x = a2.x;
                intersection->y = a2.y;
                return( FALSE );
                }
            else if ( between( a1.y, b1.y, b2.y ) )
                {
                //printf( "off-axis segments that overlap\n" );
                intersection->x = a1.x;
                intersection->y = a1.y;
                return( FALSE );
                }
            else
                {
                intersection->x = 0;
                intersection->y = 0;
                return( TRUE );
                }
            } 
        else
            {
            if ( between( a2.x, b1.x, b2.x ) )
                {
                //printf( "off-axis segments that overlap\n" );
                intersection->x = a2.x;
                intersection->y = a2.y;
                return( FALSE );
                }
            else if ( between( a1.x, b1.x, b2.x ) )
                {
                //printf( "off-axis segments that overlap\n" );
                intersection->x = a1.x;
                intersection->y = a1.y;
                return( FALSE );
                }
            else
                {
                intersection->x = 0;
                intersection->y = 0;
                return( TRUE );
                }
            }
        }
            
    // Finally, just regular lines not on-axis, not parallel

    fx = (bb - ba) / (ma - mb); 
    fy = ma * fx + ba; 

    crossing = flon_segment( fx, fy, a1, a2, b1, b2 ); 

    if ( crossing )
        {
        //printf( "off axis segments that cross\n" );
        //printf( "a = %ld,%ld  %ld,%ld\n", a1.x, a1.y, a2.x, a2.y );
        //printf( "b = %ld,%ld  %ld,%ld\n", b1.x, b1.y, b2.x, b2.y );
        //printf( "intersection = %1.1lf, %1.1lf \n", fx, fy );
        ix = floor( fx );
        intersection->x = ix; //snap( ix );
        iy = floor( fy );
        intersection->y = iy; //snap( iy );
        return( FALSE );
        }
    else
        {
        intersection->x = 0;
        intersection->y = 0;
        return( TRUE );
        }

    }

//---------end not_crossing -------------------------------------------------------------------------------------

BOOL
self_crossing( polygon *p )
    {
    static BOOL   good;
    static int    n, i, i1, j, j1;
    static xy     ip;

    good = TRUE;
    n = p->len;
    if ( n < 4 ) return( FALSE );

    i = 0;
    while ( good && (i < (n-2)) ) 
        {                         
        j = i + 2;                
        while ( good && (j < n))
            {
            j1 = (j+1) % n;
            i1 = i + 1;

            if ( j1 != i )
                good = not_crossing( p, i, i1, j, j1, &ip, FALSE );  // overlapsbad=false

            j += 1;
            }
        i += 1;
        }

    return( ! good );
    }

//-----------end self-crossing polygon------------------------------------------------------------------------


BOOL
toosmall( polygon *p )
    {
    if ( shape_extents_unknown( p ) ) set_shape_extents( p );
    if ( (p->ur.x - p->ll.x) < nearness ) return( TRUE );
    if ( (p->ur.y - p->ll.y) < nearness ) return( TRUE );
    //if ( (p->ur.x - p->ll.x) <= bssm ) return( TRUE );
    //if ( (p->ur.y - p->ll.y) <= bssm ) return( TRUE );
    return( FALSE );
    }

//--------------------------------------------------------------------------------------------------------

BOOL
between( int x, int a, int b )       // Is x between a and b?
    {
    if ( x == a ) return( TRUE );
    if ( x == b ) return( TRUE );

    if ( a < b )  
	return( x >= a && x <= b );   
    else
	return( x >= b && x <= a );
    }

//---------------------------------------------------------------------------------------------------------

BOOL
on_segment( xy p, xy a1, xy a2, xy b1, xy b2 )
    {

    // Is the point p somewhere on both line segments a1-a2 and b1-b2 ?

    static BOOL cross;

    if ( samepoint( p, a1 ) && (samepoint( p, b1 ) || samepoint( p, b2 )))
        return( TRUE );

    if ( samepoint( p, a2 ) && (samepoint( p, b1 ) || samepoint( p, b2 )))
        return( TRUE );

    if ( EQUAL(a1.y, a2.y) ) 
        cross = between( p.x, a1.x, a2.x ) && between( p.y, b1.y, b2.y );
    else if ( EQUAL(b1.y, b2.y) )
        cross = between( p.x, b1.x, b2.x ) && between( p.y, a1.y, a2.y );
    else if ( EQUAL(a1.x, a2.x) )
        cross = between( p.x, b1.x, b2.x ) && between( p.y, a1.y, a2.y );
    else if ( EQUAL(b1.x, b2.x) )
        cross = between( p.x, a1.x, a2.x ) && between( p.y, b1.y, b2.y );
    else
        cross = between( p.x, b1.x, b2.x ) && between( p.y, b1.y, b2.y ) &&
                between( p.x, a1.x, a2.x ) && between( p.y, a1.y, a2.y );

    return( cross );

    }  // on_segment

//----------------------------------------------------------------------------------------------------

double
pointoline_dist( xy p, xy a, xy b )
    {
    static double dx, dy, d, px, py, s, u, xx, yy;

    if ( samepoint( p, a ) )
        return( 0 );
    else if ( samepoint( p, b ) )
        return( 0 );
    else if ( samepoint( a, b ) ) 
        {
        dx = p.x - a.x;
        dy = p.y - a.y;
        d = sqrt( dx * dx + dy * dy );
        return( d );
        }
    else 
        {
        px = b.x - a.x;
        py = b.y - a.y;
        s = px * px + py * py;
        u = ((p.x - a.x) * px + (p.y - a.y) * py) / s;
        if ( u > 1.0 )
            u = 1.0;
        else if ( u < 0.0 )
            u = 0.0;
        xx = a.x + u * px;
        yy = a.y + u * py;
        dx = xx - p.x;
        dy = yy - p.y;
        d = sqrt( dx * dx + dy * dy );
        return( d );
        }

    }

//---------end pointoline_dist------------------------------------------------------------------------------

void
remove_duplicate_vertices( polygon *p )
    {
    static long n, nm1, i;
    static xy   vi, vi1, v0, vnm1;
    
    n   = p->len;
    nm1 = n - 1;
    i   = 0;

    while ( i < nm1 )
        {
        vi  = vertex( p, i );
        vi1 = vertex( p, i+1 );
        while ( (i < nm1) && samepoint( vi, vi1 ) )
            {
            delete_precious_vertex( p, i+1 );
            n = p->len;
            nm1 = n - 1;
            vi  = vertex( p, i );
            if ( i == nm1 )
                vi1 = vertex( p, 0 );
            else 
                vi1 = vertex( p, i+1 );
            }
        i += 1;
        }

    if ( n < 3 ) return;

    v0   = vertex( p, 0 );

    vnm1 = vertex( p, nm1 );

    if ( samepoint( v0, vnm1 ) ) p->len -= 1;

    }

//----------------------------------------------------------------------------------------------
 
double
circumference( polygon *p )
    {                                   
    static int    n;
    static double c, dx, dy;
    static int    i, j;
    static xy     vi, vj;

    n = p->len;
    c = 0.0;

    i = 0;
    while ( i < n )
        {
        j = (i + 1) % n;
        vi = vertex( p, i );
        vj = vertex( p, j );
        dx = vi.x - vj.x;
        dy = vi.y - vj.y;
        c = c + sqrt( dx * dx + dy * dy );
        i++;
        }
    
    return( c );
    }

//------------------------------------------------------------------------------------------------------------

void
snap_to_grid( polygon *p )
    {
    static int i; 
    static xy v;
    static double x, y;

    for ( i=0; i < p->len; i++ )
        {
        v = vertex( p, i );
        x = v.x;
        y = v.y;
        x = bss * floor( x / bss );
        y = bss * floor( y / bss );
        v.x = (int) floor( x );
        v.y = (int) floor( y );        
        assign_vertex( p, i, v );
        }

    //remove_duplicate_vertices( p );

    }

void
snap_to_half_grid( polygon *p )
    {
    static int i; 
    static xy v;
    static double x, y;

    for ( i=0; i < p->len; i++ )
        {
        v = vertex( p, i );
        x = v.x;
        y = v.y;
        x = bss2 * floor( x / bss2 );
        y = bss2 * floor( y / bss2 );
        v.x = (int) floor( x );
        v.y = (int) floor( y );        
        assign_vertex( p, i, v );
        }

    remove_duplicate_vertices( p );

    }

xy
snapoint( xy v )
    {
    static int ibss;
    static xy  vv;
    ibss = (int) floor( bss );
    if ( ibss <= 0 ) ibss = 1;
    if ( bss < 0.5 )
        {
        printf( "\nbss = %1.5lf  ibss = %d \n", bss, ibss );
        bail( "beam step size is too small" );
        }
    if ( ibss > 1 )
        {
        vv.x = ibss * (v.x / ibss);
        vv.y = ibss * (v.y / ibss);
        }
    else
        {
        vv.x = v.x;
        vv.y = v.y;
        }
    return( vv );
    }

void 
snapoint_to( xy *v, int n )
    {
    static double x, y, m;
    x = v->x;
    y = v->y;
    m = n;
    x = n * trunc( x / m );
    y = n * trunc( y / m );
    v->x = (int) trunc( x );
    v->y = (int) trunc( y );
    }

 
//-------------------------------------------------------------------------------------------------

BOOL
they_might_overlap( polygon *p, polygon *q )
    {
    // note that the extents were expanded by +- bss
    //
    if ( shape_extents_unknown( p ) ) set_shape_extents( p );
    if ( p->ur.x < q->ll.x ) return( FALSE );
    if ( p->ll.x > q->ur.x ) return( FALSE );
    if ( p->ur.y < q->ll.y ) return( FALSE );
    if ( p->ll.y > q->ur.y ) return( FALSE );
    return( TRUE );    
    }

//----------------------------------------------------------------------------------------------

BOOL
crossing_point( xy a1, xy a2, xy b1, xy b2, xy *intersection )
// aka find_intersection 
    {
    // Does the line a1-a2 cross the line b1-b2 ? Return True if they cross, and return the 
    // intersection point if they do. 

    // What if the crossing point is right at b1 or b2 ?
    // Then that's a valid cutting point. It's allowed.

    // What if a-b and c-d are colinear?
    // That's not a valid cut, so it's not allowed. Probably.


    static xy
        inter;

    static BOOL
        crossing,
        a_is_vertical,
        b_is_vertical,
	a_is_horizontal,
	b_is_horizontal,
        ma_equals_mb,
        ba_equals_bb;

    static double
        rise, run,
        ma, mb, ba, bb;

    intersection->x = 0;
    intersection->y = 0;

    if ( samepoint( a1, b1 ) || samepoint( a2, b2 ) || samepoint( a1, b2 ) || samepoint( a2, b1 ) )
        {
        return( FALSE );
        }

    a_is_vertical = nearly( a1.x, a2.x );   // nearly means == now, unlike the floating version
    b_is_vertical = nearly( b1.x, b2.x ); 

    a_is_horizontal = nearly( a1.y, a2.y );
    b_is_horizontal = nearly( b1.y, b2.y );

    if ( a_is_vertical && a_is_horizontal ) return( FALSE );
    if ( b_is_vertical && b_is_horizontal ) return( FALSE );

    if ( a_is_vertical && b_is_vertical )   return( FALSE );

    if ( a_is_vertical && ! b_is_vertical )
        {
        rise = b1.y - b2.y;
        run  = b1.x - b2.x;
        mb = rise / run;
        inter.y = (int) round( mb * a1.x + b1.y - mb * b1.x );  
        inter.x = a1.x;

        intersection->x = a1.x;
        intersection->y = inter.y;

        // y is now an integer, but has not snapped to the bss grid.
        // Is that what we want? Yes, probably?

        crossing = on_segment( inter, a1, a2, b1, b2 );

        if ( ! crossing )
            {
            intersection->x = 0;
            intersection->y = 0;
            return( FALSE );
            }
        else
            {
            intersection->x = inter.x;
            intersection->y = inter.y;
            return( TRUE );
            }
        }
        
    if ( b_is_vertical && ! a_is_vertical )
        {
        intersection->x = b1.x;
        rise = (a1.y - a2.y);
        run  = (a1.x - a2.x);
        ma = rise / run; 
        intersection->y = (int) round( ma * b1.x + a1.y - ma * a1.x ); 
        inter.x = intersection->x;
        inter.y = intersection->y;
        
        crossing = on_segment( inter, a1, a2, b1, b2 );

        if ( ! crossing )
            {
            intersection->x = 0;
            intersection->y = 0;
            return( FALSE );
            }
        else
            {
            intersection->x = inter.x;
            intersection->y = inter.y;
            return( TRUE );
            }
        }
        
    rise = (a1.y - a2.y);
    run  = (a1.x - a2.x);
    ma = rise / run;
    rise = (b1.y - b2.y);
    run  = (b1.x - b2.x);
    mb = rise / run;

    bb = b1.y - mb * b1.x;
    ba = a1.y - ma * a1.x;

    ma_equals_mb = dequal( ma, mb );
    ba_equals_bb = dequal( ba, bb );

    if ( ma_equals_mb && ! ba_equals_bb )
        {
        //printf( "equal slopes but not equal intercepts\n" );
        intersection->x = 0;
        intersection->y = 0;
        return( FALSE );
        }
 
    if ( ma_equals_mb )  // coincident lines, which means not a valid cut
        {
        //printf( "coincident lines\n" );
        intersection->x = 0;
        intersection->y = 0;
        return( FALSE );
        }
            
    // Finally, just regular lines not on-axis, not parallel

    double interx;
    interx  = ( (bb - ba) / (ma - mb) ); // this is where round causes gaps but floor causes faulty self-crossing polygons
    inter.y = (int) floor( ma * interx + ba );
    inter.x = (int) floor( interx );      

    crossing = on_segment( inter, a1, a2, b1, b2 );    

    if ( crossing )
        {
        intersection->x = inter.x;
        intersection->y = inter.y;
        //snapoint( intersection );   // Do we need to snap? No, and this causes a lot of problems.
        return( TRUE );
        }
    else
        {
        intersection->x = 0;
        intersection->y = 0;
        return( FALSE );
        }
    
    }

//----------end crossing_point---------------------------------------------------------------


void
extent( polygon *p, int *minx, int *maxx, int *miny, int *maxy )
    {
    static xy v;
    static int n, i;
    
    n = p->len;
    if ( n == 0 ) bail( "extent: null polygon" );

    if ( ! (p->ll.x == 0 && p->ll.y == 0 && p->ur.x == 0 && p->ur.y == 0) )
        {
        *minx = p->ll.x;
        *miny = p->ll.y;
        *maxx = p->ur.x;
        *maxy = p->ur.y;
        return;
        }

    v = vertex( p, 0 );
    *minx = v.x; 
    *maxx = *minx;
    *miny = v.y;
    *maxy = *miny;
    if ( n == 1 ) return;

    for ( i=1; i < n; i++ )
        {
        v = vertex( p, i );
        if ( v.x < *minx ) *minx = v.x;
        if ( v.x > *maxx ) *maxx = v.x;
        if ( v.y < *miny ) *miny = v.y;
        if ( v.y > *maxy ) *maxy = v.y;
        }

    p->ll.x = *minx;
    p->ll.y = *miny;
    p->ur.x = *maxx;
    p->ur.y = *maxy;
    }

//----------end extent---------------------------------------------------------------------------

BOOL
colinear( xy a, xy b, xy c, int *middle )
    {
    // If the three points are colinear then return TRUE. The return parameter
    // "middle" will indicate which of the three points is in the middle.

    static BOOL first, second, third, vertical, horizontal, colinear;
    static double abc, bac, cab;

    first  = FALSE;
    second = FALSE;
    third  = FALSE;

    if ( EQUAL(c.x, a.x) || EQUAL(c.y, a.y) || EQUAL(b.x, a.x) || EQUAL(b.y, a.y) )
        {
        vertical   = EQUAL(a.x, b.x) && EQUAL(b.x, c.x); 
        horizontal = EQUAL(a.y, b.y) && EQUAL(b.y, c.y);
        colinear   = horizontal || vertical;
        if ( horizontal )
            {
            second = between( b.x, a.x, c.x );
            if ( ! second )
                first = between( a.x, b.x, c.x );
            third = (! second) && (! first);
            }
        if ( vertical )
            {
            second = between( b.y, a.y, c.y );
            if ( ! second )
                first = between( a.y, b.y, c.y );
            third = (! second) && (! first);
            }
        }
    else
        {
        //ac_slope = (c.y - a.y) / (c.x - a.x);
        //ab_slope = (b.y - a.y) / (b.x - a.x);
        //colinear = dequal( ac_slope, ab_slope );

        abc = pointoline_dist( a, b, c );
        bac = pointoline_dist( b, a, c );
        cab = pointoline_dist( c, a, b );

        //printf( "\na: %g,%g  b: %g,%g  c: %g,%g", a.x, a.y, b.x, b.y, c.x, c.y );
        //printf( "\nabc: %g  bac: %g  cab: %g  tooclose: %g\n", abc, bac, cab, tooclose );

        colinear = TOOCLOSE( abc ) || TOOCLOSE( bac ) || TOOCLOSE( cab ); 

        if ( colinear )
            {
            second = between( b.x, a.x, c.x );
            if ( ! second )
                first = between( a.x, b.x, c.x );
            third = (! second) && (! first);
            }
        }
         
    // It's very difficult to decide which point to delete. Usually, we should delete
    // the middle point. However, if two of the points are very close then we should
    // delete the endpoint, maybe.

    *middle = -1;

    if ( colinear )
        {
        if ( samepoint( a, b ) )
            { 
            *middle = 0;              // delete the first vertex
            //printf( "\ndelete a\n" );
            }
        else if ( samepoint( b, c ) )
            {            
            *middle = 2;              // delete the third vertex
            //printf( "\ndelete c\n" );
            }
        else        
            {
            if ( first ) 
                *middle = 0;
            else if ( second ) 
                *middle = 1;
            else if ( third )
                *middle = 2;
            }
        }

    return colinear;
    }

//---------end colinear-----------------------------------------------------------------------------


void
remove_colinear_vertices( polygon *p, BOOL severe )
    {
    // Much harder than it might seem. Normally we would want to delete 
    // the vertex in the middle, but if the polygon doubles back on itself
    // then this can cause the polygon to become distorted. If the 
    // parameter "severe" is true, then this function will delete the
    // middle vertex anyway. The caller must decide when this extra
    // effort is necessary.
    
    static int i, n, nm1, nm2, j, k, middle;
    static BOOL escape;
    
    severe = FALSE;       // sometimes 'severe' causes shapes to collapse

    i = 0;
    middle = -1;
    n = p->len;
    nm2 = n - 2;
    nm1 = n - 1;

    while ( i < nm2 )
        {
        j = i + 1;
        k = j + 1;
        escape = FALSE;

        while ( ! escape && (i < nm2) && colinear( vertex(p,i), vertex(p,j), vertex(p,k), &middle ) )
            {
            if ( middle == 1 )
                delete_vertex( p, j );
            else if ( severe )
                {
                if ( middle == 0 )
                    delete_vertex( p, i );
                else
                    delete_vertex( p, k );
                }
            else
                escape = TRUE;

            n = p->len;
            nm1 = n - 1;
            nm2 = n - 2;
            }    
        i = i + 1;
        }

    n = p->len;
    nm1 = n - 1;
    nm2 = n - 2;

    if ( n > 2 )
        {
        if ( colinear( vertex(p,nm2), vertex(p,nm1), vertex(p,0), &middle ) )
            {
            if ( middle == 1 )
                delete_vertex( p, nm1 );
            else if ( severe )
                {
                if ( middle == 0 )
                    delete_vertex( p, nm2 );
                else
                    delete_vertex( p, 0 );
                }
            }
        }

    n = p->len;
    nm1 = n - 1;
    nm2 = n - 2;
        
    if ( n > 2 )
        {
        if ( colinear( vertex(p,n-1), vertex(p,0), vertex(p,1), &middle ) )
            { 
            if ( middle == 1 )
                delete_vertex( p, 0 );
            else if ( severe )
                {
                if ( middle == 1 )
                    delete_vertex( p, n-1 );
                else
                    delete_vertex( p, 1 );
                }
            }
        }
    }

//-----------end remove_colinear_vertices----------------------------------------------------------

BOOL
shape_extents_unknown( polygon *p )
    {
    return( p->ll.x == 0 && p->ll.y == 0 && p->ur.x == 0 && p->ur.y == 0 );
    }


void
set_shape_extents( polygon *p )
    {
    static int  i;
    static xy   v;

    if ( p->len < 3 )
        {
        p->active = FALSE;
        return;
        }

    p->ll = vertex( p, 0 );
    p->ur = vertex( p, 0 );

    for ( i=1; i < p->len; i++ )
        {
        v = vertex( p, i );
        if ( v.x < p->ll.x ) p->ll.x = v.x;
        if ( v.y < p->ll.y ) p->ll.y = v.y;
        if ( v.x > p->ur.x ) p->ur.x = v.x;
        if ( v.y > p->ur.y ) p->ur.y = v.y;
        }

    return;
    }

//------------end set_shape_extents----------------------------------------------------------------

BOOL
polygon_includes_vertex( polygon *a, xy v )
    {
    static BOOL si;
    static int  i;
    static xy   w;

    si = FALSE;
    for ( i=0; i < a->len && ! si; i++ )
        {
        w = vertex( a, i );
        si = samepoint( v, w );
        }
    return( si );
    }


BOOL
identical_polygons( polygon *a, polygon *b )
    {
    // Do the polygons have identical vertices?
    // Oh sure, you can have different polygons
    // with the same vertices, but that's never
    // going to happen.

    static int  i;
    static BOOL oui;
    static xy   v;

    oui = TRUE;  

    for ( i=0; i < a->len && oui; i++ )
        {
        v = vertex( a, i );
        oui = polygon_includes_vertex( b, v );
        }

    return( oui );
    }

//------------end identical_polygons---------------------------------------------------------------

BOOL
far_away( xy point, polygon *trap ) 
    {
    if ( shape_extents_unknown( trap ) ) set_shape_extents( trap );
    return( ! ( point.x < (trap->ur.x + bssri) && 
                point.x > (trap->ll.x - bssri) &&
                point.y < (trap->ur.y + bssri) &&
                point.y > (trap->ll.y - bssri)    )   );
    }

//-------------------------------------------------------------------------------------------------

queue *
cutline( queue *traps, polygon *trap, xy a, xy b, xy d, BOOL impunity, BOOL *success )
    { 
    static int    n;
    static int    minx, maxx, miny, maxy;
    static BOOL   horizontal, crossing, flip, first=TRUE;
    static int    *hitype;
    static int    vertex_hits, side_hits, i;
    static xy     v1, v2, point, *ps;
    static int    maxv;
    polygon       *t1, *t2;


    *success = FALSE;

    if ( ! trap ) bail( "[slice/cutline] null polygon sent to cutline" );

    if ( ! trap->active ) return( traps );

    n = trap->len;

    if ( first )
	{
	first = FALSE;
	maxv = 2 * n;
	hitype = (int *) malloc( maxv * sizeof(int) );
	ps     = (xy *)  malloc( maxv * sizeof(xy)  );
	}

    if ( n > maxv )
	{
	maxv = 2 * n;
	hitype = (int *) malloc( maxv * sizeof(int) );
	ps     = (xy *)  malloc( maxv * sizeof(xy)  );
	}

    if ( n < 3 ) 
        {
        trap->active = FALSE;
        return( traps );
        }

    extent( trap, &minx, &maxx, &miny, &maxy );
    horizontal = EQUAL( a.y, b.y ); 

    if ( horizontal )
        {
        a.y = b.y;
        if ( ! between( a.y, miny, maxy ) ) 
            { 
            return( traps );
            }
        }
    else
        {
	a.x = b.x;
        if ( ! between( a.x, minx, maxx ) ) 
            {
            return( traps );
            }
        }
 
    vertex_hits = 0;
    side_hits   = 0;

    for ( i=0; i < maxv; i++ ) 
        {
	hitype[i] = 0;
	ps[i].x = 0;
	ps[i].y = 0;
	}
    
    // hitype 0 means no intersection
    //        1 means the intersection on this vertex
    //        2 means the intersection is between this vertex and the next


    for ( i=0; i < n; i++ )   
        {
        v1 = vertex( trap, i   );
        v2 = vertex( trap, (i+1) % n );
        crossing = crossing_point( v1, v2, a, b, &point );

        if ( crossing )
            {
            hitype[i] = 2;      // intersection is between
            if ( samepoint( v1, point ) )
                {
                assign_vertex( trap, i, point );
                hitype[i] = 1;  // intersection is on the vertex
                }
            else if ( samepoint( v2, point ) )
                {
                assign_vertex( trap, ((i+1) % n), point );
                hitype[i] = 0;  // unless we've hit the next vertex 
                }
            
            if ( hitype[i] == 1 ) vertex_hits++;
            if ( hitype[i] == 2 ) side_hits++;
            ps[i].x = point.x;
            ps[i].y = point.y;

	    }
        }

    if ( vertex_hits == 0 && side_hits   == 0 ) return( traps );
    if ( vertex_hits == 1 && side_hits   == 0 ) return( traps );
    if ( side_hits   == 1 && vertex_hits == 0 ) return( traps );                   // strange case
    if ( side_hits > 2 || (side_hits == 2 && vertex_hits > 0 ) ) return( traps );  // sometimes helps, sometimes hurts!  //???????


    // Split this trap into two shapes.
    // The new shapes could have as many vertices as the original polygon

    t1 = initialize_polygon( trap->len, trap->layer, trap->dt, trap->obese );
    t2 = initialize_polygon( trap->len, trap->layer, trap->dt, trap->obese );

    // Here's the plan: Run around the original shape, collecting vertices and assigning
    // them to the new shape. If we hit a point where the line crosses, then assign
    // that point to both new shapes and switch the destination shape.
    
    flip = FALSE;

    for ( i=0; i < n; i++ )
        {
        if ( flip )
            {
            append_polygon_xy( t2, vertex( trap, i ) );
            if ( hitype[i] == 2 )
                {
                append_polygon_xy( t2, ps[i] );
                append_polygon_xy( t1, ps[i] );
                flip = ! flip;
                }
            else if ( hitype[i] == 1 )
                {
                append_polygon_xy( t1, ps[i] );
                flip = ! flip;
                }
            }
        else
            {
            append_polygon_xy( t1, vertex( trap, i ) );
            if ( hitype[i] == 2 ) 
                {
                append_polygon_xy( t1, ps[i] );
                append_polygon_xy( t2, ps[i] );
                flip = ! flip;
                }
            else if ( hitype[i] == 1 )
                {
                append_polygon_xy( t2, ps[i] );
                flip = ! flip;
                }
            }
        }

    //snap_to_half_grid( t1 );
    //remove_colinear_vertices( t1, FALSE );
    //snap_to_half_grid( t2 );
    //remove_colinear_vertices( t2, FALSE );

    // Fewer problems are caused if we reject both polygons if one is degenerate.
    // But if we are quarting an obese shape then we slice with impunity. 

    *success = FALSE;

    if ( impunity )
        {
        if ( t1->len > 2 && ! self_crossing( t1 ) )
            {
            set_shape_extents( t1 );
            append_queue( traps, t1 ); 
            *success = TRUE;
            }
        if ( t2->len > 2 && ! self_crossing( t2 ) )
            {
            set_shape_extents( t2 );
            append_queue( traps, t2 );
            *success = TRUE; 
            }
        trap->active = FALSE;     
        }
    else if ( t1->len > 2 && t2->len > 2 && ! self_crossing( t1 ) && ! self_crossing( t2 ) )
        {
        set_shape_extents( t1 );
        set_shape_extents( t2 );

        if ( ! ( identical_polygons( t1, trap ) || identical_polygons( t2, trap ) ) )
            {
            if ( ! toosmall( t1 ) ) append_queue( traps, t1 ); 
            if ( ! toosmall( t2 ) ) append_queue( traps, t2 ); 
            trap->active = FALSE;
            }

        *success = TRUE;     
        }
    else
	{
	//printf( "\nUnable to cut polygon in slice.c ... " );
	//print_polygon( trap, 1 );
	*success = FALSE;
	}

    // If you try to discard one degenerate, keeping the other,
    // then the good one is stuck on the bottom of the queue
    // where it will be again sliced to one degenerate and one 
    // good polygon. The program falls into an infinite loop.
        
    return( traps );
    }

//--------end cutline---------------------------------------------------------------------

int
len( ill *p )
    {
    ill *q;
    int n;
    n = 0;
    q = p;
    while ( q )
        {
        if ( q->n >= 0 ) n++;
        q = q->next;
        }
    return( n );
    }

//--------------------------------------------------------------------------------------------------

#define EGAL( a, b ) ((a) == (b))


BOOL
not_a_large_rectangle( polygon *p, xy d, BOOL subfield_cut )
    {
    // d is the subfield size.
    // If p is bigger than a subfield but smaller than a block,
    // then p could be a "large rectangle" which can be expressed
    // as an array. Maybe. Arrays are very tricky because they
    // might land in the wrong field due to rounding errors. 

    static xy zero, one, two, three;
    int ex, ey;

    if ( ! subfield_cut ) return( TRUE );    // always cut at field boundaries

    if ( p->len != 4 ) return( TRUE );
    if ( (p->ur.x == p->ll.x) || (p->ur.y == p->ll.y) ) set_shape_extents( p );

    ex = p->ur.x - p->ll.x;
    ey = p->ur.y - p->ll.y;

    if ( (ex < d.x) && (ey < d.y)) return( TRUE );   // too small! no cutting or array needed!

    zero  = vertex( p, 0 );
    one   = vertex( p, 1 );
    two   = vertex( p, 2 );
    three = vertex( p, 3 );

    // Is this thing a rectangle? If not, return true, meaning make the cut

    if ( ! (EGAL(zero.y,three.y) || EGAL(zero.x,three.x))) return( TRUE );
    if ( ! (EGAL(zero.y,  one.y) || EGAL(zero.x,  one.x))) return( TRUE );
    if ( ! (EGAL(one.y,   two.y) || EGAL(one.x,   two.x))) return( TRUE );
    if ( ! (EGAL(two.y, three.y) || EGAL(two.x, three.x))) return( TRUE );

    return( FALSE );
    }

//--------------------------------------------------------------------------------------------------

void
horizontal_cutup( queue *traps, xy ll, xy ur, xy block, xy d, BOOL quiet, BOOL this_is_subfield_cut, BOOL noarray )
    {

    // Starting at ll, make horizontal cuts through all the shapes in "traps".

    // block is the block size (block.x, block.y)
    // d is the subfield size (d.x, d.y)

    static int               count;
    static xy                a, b;
    static int               endofblock;
    static polygon_ref       *pr;
    static polygon           *trap;
    static BOOL              success, goforit;
    static int               nnag;
    
    if ( block.y <= 0 ) bail( "horizontal_cutup: y block size <= 0" );

    if ( traps->len == 0 ) nag( "horizontal_cutup: there are no shapes", &nnag );

    /*
    pr = traps->first;
    while ( pr )
        {
        pr->p->hit = FALSE;
        pr = pr->next;
        }
    */

    a.x = ll.x - block.x;  a.y = ll.y;  // a-b is a horizontal line
    b.x = ur.x + block.x;  b.y = a.y;   // starting at the LL corner

    // snap a and b to the output mainfield lsb 
    // we don't know what that is, so let's try snapping to the bss instead

    a = snapoint( a );
    b = snapoint( b );

    endofblock = a.y + block.y;
    count = 0;

    while ( a.y < ur.y ) 
        {
        count++;

        while ( a.y <= endofblock )
            {
            pr = traps->first;
            
            while ( pr )
                {
                trap = pr->p;
                if ( trap->active && a.y < trap->ur.y && a.y > trap->ll.y )
                    {
  		    goforit = noarray || not_a_large_rectangle( trap, d, this_is_subfield_cut ); // field cuts always return TRUE
                    if ( goforit ) traps = cutline( traps, trap, a, b, d, FALSE, &success );
                    }
                pr = pr->next;
                }

            a.y = a.y + d.y;
            b.y = a.y;
	    a = snapoint( a );
	    b = snapoint( b );
            }
       
        if ( a.y > endofblock ) // happens when block size is not multiple of subfield size
            {
            a.y = endofblock;
            b.y = a.y;
            a = snapoint( a );
	    b = snapoint( b );
            }

        endofblock += block.y;
        }
    
    }

//----------end horizontal_cutup-----------------------------------------------------------------------------

void
one_horizontal_cut( queue *traps, xy cut, xy ll, xy ur )
    {

    // Make one cut at cut.y

    static xy                a, b, d;
    static polygon_ref       *pr;
    static polygon           *trap;
    static BOOL              success;
    static int               nnag;

    d.x = 0;
    d.y = 0;

    if ( traps->len == 0 ) nag( "one_horizontal_cut: there are no shapes", &nnag );

    a.x = ll.x;  a.y = cut.y;           // a-b is a horizontal line
    b.x = ur.x;  b.y = cut.y;           
    a.y = b.y;

    pr = traps->first;
            
    while ( pr )
        {
        trap = pr->p;
        if ( a.y < trap->ur.y && a.y > trap->ll.y )
            {
            traps = cutline( traps, trap, a, b, d, FALSE, &success );
            }
        pr = pr->next;
        }
    }

//-----------end one_horizontal_cut---------------------------------------------------------------

void
one_vertical_cut( queue *traps, xy cut, xy ll, xy ur )
    {

    // Make one cut at cut.x

    static xy                a, b, d;
    static polygon_ref       *pr;
    static polygon           *trap;
    static BOOL              success;
    static int               nnag;

    d.x = 0;
    d.y = 0;

    if ( traps->len == 0 ) nag( "one_vertical_cut: there are no shapes", &nnag );

    a.x = cut.x;  a.y = ll.y;           // a-b is a horizontal line
    b.x = cut.x;  b.y = ur.y;           
    a.x = b.x;

    pr = traps->first;
            
    while ( pr )
        {
        trap = pr->p;
        if ( a.x < trap->ur.x && a.x > trap->ll.x )
            {
            traps = cutline( traps, trap, a, b, d, FALSE, &success );
            }
        pr = pr->next;
        }
    }

//-----------end one_vertical_cut---------------------------------------------------------------

void
crop_pattern( queue *traps, xy cropll, xy cropur, xy ll, xy ur )
    {
    static polygon_ref *pr;
    static polygon     *t;

    //printf( "\nCropping..." ); fflush( stdout );
    one_vertical_cut( traps, cropll, ll, ur );
    one_vertical_cut( traps, cropur, ll, ur );
    one_horizontal_cut( traps, cropll, ll, ur );
    one_horizontal_cut( traps, cropur, ll, ur );

    cropll.x -= bss;
    cropll.y -= bss;
    cropur.x += bss;
    cropur.y += bss;
    
    pr = traps->first;
    while ( pr )
        {
        t = pr->p;
        if ( ! (t->ll.x >= cropll.x && t->ur.x <= cropur.x && t->ll.y >= cropll.y && t->ur.y <= cropur.y ) )
            t->active = FALSE;
        pr = pr->next;
        }

    //printf( " done\n\n" ); fflush( stdout );
    }

//------end crop------------------------------------------------------------------------------------------

//-------------------------------------
BOOL
near( int a, int b )
    {
    static int small=50;
    return( abs(a - b) < small ); 
    }

//-------------------------------------

void
vertical_cutup( queue *traps, xy ll, xy ur, xy block, xy d, BOOL quiet, BOOL this_is_subfield_cut, BOOL noarray )
    {

    // Starting at ll, make vertical cuts through all the shapes in "traps".

    // block is the block size (block.x, block.y)
    // d is the subfield size (d.x, d.y)

    static int               count;
    static xy                a, b;
    static int               endofblock;
    static polygon_ref       *pr;
    static polygon           *trap;
    static BOOL              success, goforit;
    static int               nnag;

    if ( block.x <= 0.0 ) bail( "vertical_cutup: x block size <= 0" );

    if ( traps->len == 0 ) nag( "vertical_cutup: there are no shapes", &nnag );

    a.x = ll.x;  a.y = ll.y - block.y;   // a-b is a vertical line
    b.x = ll.x;  b.y = ur.y + block.y;   // starting at the LL corner

    // snap a and b to the output mainfield lsb 
    // we don't know what that is, so let's try snapping to the bss instead

    a = snapoint( a );
    b = snapoint( b );

    endofblock = a.x + block.x;
    count = 0;

    while ( a.x < ur.x ) 
        {
        count++;

        while ( a.x <= endofblock )
            {
            pr = traps->first;
	    //printf( "cut at x = %ld \n", a.x );
            while ( pr )
               {
               trap = pr->p;
               if ( trap->active && a.x < trap->ur.x && a.x > trap->ll.x )
                   {
                   goforit = noarray || not_a_large_rectangle( trap, d, this_is_subfield_cut ); 
                   if ( goforit ) traps = cutline( traps, trap, a, b, d, FALSE, &success );
                   }
               pr = pr->next;
               }
            a.x = a.x + d.x;
            b.x = a.x;
            a = snapoint( a );
	    b = snapoint( b );
            }
       
        if ( a.x > endofblock ) // happens when block size is not multiple of subfield size
            {
            a.x = endofblock;
            b.x = a.x;
	    a = snapoint( a );
	    b = snapoint( b );
            }

        endofblock += block.x;
        }
    
    }

//-----------end vertical_cutup---------------------------------------------------------------


void
copy_from_gds_polygon( struct gds_itemtype *item, polygon *poly )
    {
    // copy item to poly
    // poly has already been initialized

    static int i;

    // the append function checks for overflow

    for ( i=0; i < item->n; i++ )
        {
        append_polygon( poly, item->x[i], item->y[i] );
        }

    //snap_to_grid( poly );  we will do this after everything is read in, so we can marry the vertices.
    
    }  // copy_from_gds_polygon

//-------------------------------------------------------------------------------------------------

void
copy_to_gds_polygon( polygon *poly, struct gds_itemtype *item )
    {
    // copy poly to item 
    // item has already been allocated 

    static int i;
    static xy  v;

    if ( item->n < poly->len ) bail( "attempt to copy large poly into small gds item" );

    for ( i=0; i < poly->len; i++ )
        {
        v = vertex( poly, i );
        item->x[i] = v.x;
        item->y[i] = v.y;
        }

    item->layer = poly->layer;
    item->dt    = poly->dt;
    item->n     = poly->len;
    
    }  // copy_to_gds_polygon

//--------------------------------------------------------------------------------------------------

BOOL 
not_a_parallelogram( polygon *p )
    {

    // Are two sides parallel and on-axis?
    // Or maybe it's a triangle with one side on-axis.
    // Off-axis parallelograms - nope.
    // Maybe it's just too big. How did that happen?

    static int    n;
    static BOOL   yup, onaxis, horizontal, vertical,
                  hor01, hor12, hor23, hor30,
                  ver01, ver12, ver23, ver30; 
    static xy     v0, v1, v2, v3;

    n = p->len;

    if ( (n != 3) && (n != 4) ) return( TRUE );

    yup = FALSE;      
    v0 = vertex( p, 0 );
    v1 = vertex( p, 1 );
    v2 = vertex( p, 2 );

    if ( n == 3 )
        {  
        onaxis = EQUAL(v0.y, v1.y);
        if ( ! onaxis ) onaxis = EQUAL(v1.y, v2.y);
        if ( ! onaxis ) onaxis = EQUAL(v2.y, v0.y); 
        if ( ! onaxis ) onaxis = EQUAL(v0.x, v1.x); 
        if ( ! onaxis ) onaxis = EQUAL(v1.x, v2.x); 
        if ( ! onaxis ) onaxis = EQUAL(v2.x, v0.x); 
        yup = onaxis;
        }
    else
        {
        v3 = vertex( p, 3 );
        hor01 = EQUAL(v0.y, v1.y); 
        hor12 = EQUAL(v1.y, v2.y); 
        hor23 = EQUAL(v2.y, v3.y); 
        hor30 = EQUAL(v3.y, v0.y); 
        horizontal = (hor01 && hor23) || (hor12 && hor30);
        vertical = FALSE;
        if ( ! horizontal )
            {
            ver01 = EQUAL(v0.x, v1.x); 
            ver12 = EQUAL(v1.x, v2.x); 
            ver23 = EQUAL(v2.x, v3.x); 
            ver30 = EQUAL(v3.x, v0.x); 
            vertical = (ver01 && ver23) || (ver12 && ver30);
            }
        yup = horizontal || vertical;  // yes it's a parallelogram
        }

    return( ! yup );
    }

//-----------end not_a_parallelogram------------------------------------------------------------------------------

BOOL
obese_shape( polygon *p, xy d, xy block, BOOL noarray )
    {
    static int minx, maxx, miny, maxy;

    extent( p, &minx, &maxx, &miny, &maxy );

    if ( (fabs(maxx - minx) > d.x) || (fabs(maxy - miny) > d.y) )
        {
        if ( noarray )
            return( TRUE );
        else        
	    return( not_a_large_rectangle( p, d, TRUE ) );
        }
    else
        {
	return( FALSE );
        }

    }


//----------------------------------------------------------------------------------------------------------

BOOL
obese_array( polygon *p, xy d, xy block )
    {

    // big rectangles can be instantiated as arrays in gpfout, but we should still check
    // that they are smaller than a block, because that happens rarely
 
    static int minx, maxx, miny, maxy;

    extent( p, &minx, &maxx, &miny, &maxy );

    return( (fabs(maxx - minx) > block.x) || (fabs(maxy - miny) > block.y) );
    }


//----------------------------------------------------------------------------------------------------------

void
cut_triangular_sliver( polygon *p, BOOL *success, int *iter, double dbu_um )
    {
    // Given a long, narrow triangle, this function chops by moving 
    // the pointy end to half-way along the long side.

    static xy     v0, v1, v2, h1, vr, vm;
    static double d0, d1, d2, dx, dy, dmin, dmax, h, theta, rise, run, ar;
    static int    small, right, mid, ih, hcos, hsin;

    if ( p->len != 3 )
        {
        printf( "\nERROR: cut_triangular_sliver has been sent something other than a triangle.\n" );
        *success = FALSE;       
        return;
        }
    
    //snap_to_grid( p );   often causes triangles to shift down one pixel

    if ( degenerate( p ) )
	{
	*success = TRUE;
	p->active = FALSE;
	return;
	}

    *success = TRUE;
    rise = p->ur.y - p->ll.y; 
    run  = p->ur.x - p->ll.x;
    if ( (rise <= bss) || (run <= bss) ) { p->active = FALSE; return; }
    ar = fabs( run / rise );
    if ( ar < 1.0 ) ar = 1.0 / ar;
    if ( ar < 100.0 ) return;
    *success = FALSE;

    right = -1;
    mid   = -1;

    v0 = vertex( p, 0 );
    v1 = vertex( p, 1 );
    v2 = vertex( p, 2 );
    dx = v0.x - v1.x;
    dy = v0.y - v1.y;
    d0 = sqrt( dx * dx + dy * dy );
    dx = v1.x - v2.x;
    dy = v1.y - v2.y;
    d1 = sqrt( dx * dx + dy * dy );
    dx = v2.x - v0.x;
    dy = v2.y - v0.y;
    d2 = sqrt( dx * dx + dy * dy );
    
    small = 0;  // 0-1 is the smallest side?
    dmin = d0;

    if ( d1 < dmin )  { small = 1; dmin = d1; }
    if ( d2 < dmin )  { small = 2; dmin = d2; }
    if ( small == 0 ) 
        { mid  = 2; right = 1; }             // left=0 was once included
    else if ( small == 1 ) 
        { mid = 0; right = 2; }              // left=1
    else if ( small == 2 ) 
        { mid = 1; right = 0; }              // left=2

    dmax = d0;
    if ( d1 > dmax ) dmax = d1;
    if ( d2 > dmax ) dmax = d2;

    if ( EQUAL( dmin, 0 ) )  // degenerate triangle
	{
	p->active = FALSE;
        return;
	}

    h = dmax - 0.4 * bss * dmax / dmin; //  h is the length at which the pointy end is 0.4 bss wide
     
    ih = (int) round( h );

    if ( h < bss ) 
        {
        p->active = FALSE;
	*success = TRUE;
        //printf( "\nThrowing away long triangular sliver.\n" );
        //print_polygon( p, 10000 );
        return;
        }

    //vl = vertex( p, left );
    vr = vertex( p, right );
    vm = vertex( p, mid );

    if ( EQUAL( vm.x, vr.x ) )
        {
        h1.x = vm.x;
        if ( vm.y > vr.y )
            h1.y = vr.y + ih;
        else
            h1.y = vr.y - ih;
        }
    else if ( EQUAL( vm.y, vr.y ) )
        {
        h1.y = vm.y;
        if ( vm.x > vr.x )
            h1.x = vr.x + ih;
        else
            h1.x = vr.x - ih;
        }
    else
        {
        rise = vm.y - vr.y;
        run  = vm.x - vr.x;

        theta = fabs( atan( rise / run ) );
        
        hcos = (int) round( h * cos( theta ) );
	hsin = (int) round( h * sin( theta ) );

        if ( vm.x < vr.x && vm.y > vr.y )       // quadrant 2
            {
            h1.x = vr.x - hcos;
            h1.y = vr.y + hsin;
            }
        else if ( vm.x < vr.x && vm.y < vr.y )  // quadrant 3
            {
            h1.x = vr.x - hcos;
            h1.y = vr.y - hsin;
            }
        else if ( vm.x > vr.x && vm.y < vr.y )  // quadrant 4
            {
            h1.x = vr.x + hcos;
            h1.y = vr.y - hsin;
            }
        else                                    // quadrant 1
            {
            h1.x = vr.x + hcos;
            h1.y = vr.y + hsin;
            }
        }

    set_vertex( p, mid, h1 );

    *success = TRUE;
    rise = p->ur.y - p->ll.y; 
    run  = p->ur.x - p->ll.x;
    if ( (rise <= bss) || (run <= bss) ) { p->active = FALSE; return; }
    ar = fabs( run / rise );
    if ( ar < 1.0 ) ar = 1.0 / ar;
    if ( ar < 100.0 ) return;
    *success = FALSE;

    (*iter)++;

    if ( *iter < 4 ) 
	cut_triangular_sliver( p, success, iter, dbu_um );
    else
	{
	printf( "\n--------------------- Unable to cut long triangle.\n" );
	print_polygon( p, 1.0/dbu_um );
	}

    return;
    }



//----------------------------------------------------------------------------------------------------------------

void
quarter( polygon *p, queue *traps, xy ll, xy ur, xy subfield, double dbu_um )
    {
    
    // no more mister nice guy - cut the bloody shape with a rusty spoon
    
    static int         minx, maxx, miny, maxy, dx, dy, midx, midy, iter;
    static double      ar, rise, run;
    static xy          a, b, c, d, v;
    static BOOL        success, fatx, faty;
    static double      oldnearness;

    if ( ! p->active ) return;

    d.x = 0.0;
    d.y = 0.0;

    extent( p, &minx, &maxx, &miny, &maxy );

    // first get rid of little slivers

    oldnearness = nearness;
    nearness = 1.5 * bss;
    presque = nearness;
    printv( "\nQuartering shape and removing colinear vertices with impunity.\n" );
    remove_colinear_vertices( p, TRUE );
    remove_duplicate_vertices( p );
    nearness = 0.5 * bss; // 0.5 * bss;
    presque = nearness;

    if ( p->len < 3 )
        {
        p->active = FALSE;
        return;
        }

    if ( fabs(maxx - minx) < oldnearness || fabs(maxy - miny) < oldnearness )
        {
	//printf( "\nthis shape is too thin\n" );
        p->active = FALSE;
        return;
        }
  
    midx = (maxx + minx) / 2;
    midy = (maxy + miny) / 2;

    dx = (maxx - minx) / 5; 
    dy = (maxy - miny) / 5;
    if ( dx == 0 ) dx = 2;
    if ( dy == 0 ) dy = 2;

    fatx = fabs( maxx - minx ) > subfield.x;
    faty = fabs( maxy - miny ) > subfield.y;

    minx -= dx;
    maxx += dx;
    miny -= dy;
    maxy += dy;    

    success = TRUE;

    if ( fatx ) 
        {
        a.x = midx;
        a.y = miny;
        b.x = midx;
        b.y = maxy;

        traps = cutline( traps, p, a, b, d, TRUE, &success );   // true means throw away degenerates

        if ( ! success && p->active )
            {
            rise = maxy - miny;
            run  = maxx - minx;
            ar = fabs( run / rise );
            if ( ar < 1.0 ) ar = 1.0 / ar;

            if ( ar > 10 ) 
                {
                iter = 0;
                if ( p->len == 3 ) 
		    cut_triangular_sliver( p, &success, &iter, dbu_um );

                if ( ! success && p->active )
                    {
                    nearness = nearness / 2.0;
                    run = 0.75 * (maxx - minx);
                    midx = minx + (int) round( run );
                    a.x = midx;
                    a.y = miny;
                    b.x = midx;
                    b.y = maxy;
                    traps = cutline( traps, p, a, b, d, TRUE, &success );
                    nearness = nearness * 2.0;
                    if ( ! success )
		        {
                        printf( "\nvertical quartering slice unsuccessful.\n" );
                        printf( "This is a long narrow thing. dx= %1.3lf  dy= %1.3lf beam steps\n",
                            fabs(maxx-minx)/bss, fabs(maxy-miny)/bss );
                        p->active = FALSE;
                        v = vertex( p, 0 );
                        printf( "Look in the gds file at %ld, %ld \n", v.x, v.y );
                        c.x = (ur.x + ll.x) / 2.0;
                        c.y = (ur.y + ll.y) / 2.0;
                        printf( "or in the gpf file at   %ld, %ld \n", (v.x-c.x), (v.y-c.y) );
		        }
                    }
                }
            else if ( ! success )
                {
                nearness = nearness / 2.0;
		run = 0.75 * (maxx - minx);
                midx = minx + (int) round( run );
                a.x = midx;
                a.y = miny;
                b.x = midx;
                b.y = maxy;
                traps = cutline( traps, p, a, b, d, TRUE, &success );
                nearness = nearness * 2.0;
                if ( ! success )
                    {  
                    printf( "\nvertical quartering slice unsuccessful.\n" );
                    printf( "This is not a narrow shape, so that's especially bad.\n" );
                    p->active = FALSE;
                    v = vertex( p, 0 );
                    printf( "Look in the gds file at %ld, %ld \n", v.x, v.y );
                    c.x = (ur.x + ll.x) / 2;
                    c.y = (ur.y + ll.y) / 2;
                    printf( "or in the gpf file at   %ld, %ld \n", (v.x-c.x), (v.y-c.y) );
                    print_polygon( p, 1000 );
                    }
                }
            }        
        }

    if ( faty )
        {
        a.x = minx;
        a.y = midy;
        b.x = maxx;
        b.y = midy;
       
        traps = cutline( traps, p, a, b, d, TRUE, &success );   

        if ( ! success && p->active )
            {
            ar = fabs((maxx - minx) / (maxy - miny));
            if ( ar < 1.0 ) ar = 1.0 / ar;

            if ( ar > 10 ) 
                {
                iter = 0;
                if ( p->len == 3 ) 
                    cut_triangular_sliver( p, &success, &iter, dbu_um );

                if ( ! success && p->active )
                    {
                    nearness = nearness/2.0;
		    presque  = nearness;
                    midy = miny + 0.4 * (maxy - miny);
                    a.x = minx;
                    a.y = midy;
                    b.x = maxx;
                    b.y = midy;
                    traps = cutline( traps, p, a, b, d, TRUE, &success );  

                    if ( ! success )
		        {
                        midy = miny + 0.6 * (maxy - miny);
                        a.x = minx;
                        a.y = midy;
                        b.x = maxx;
                        b.y = midy;
                        traps = cutline( traps, p, a, b, d, TRUE, &success );  
			if ( ! success )
			    {
                            //printf( "\n     Horizontal quartering slice unsuccessful.\n" );
                            //printf( "     This is a long narrow thing. dx= %1.3lf  dy= %1.3lf beam steps \n",
                            //    fabs(maxx-minx)/bss, fabs(maxy-miny)/bss );
                            p->active = FALSE;
                            //v = vertex( p, 0 );
                            //printf( "     Look in the gds file %s at %ld, %ld \n", ifilename, v.x, v.y );
                            //c.x = (ur.x + ll.x) / 2;
                            //c.y = (ur.y + ll.y) / 2;
                            //printf( "     or in the gpf file at   %ld, %ld \n", (v.x-c.x), (v.y-c.y) );
			    //printf( "     cut line is %ld, %ld to %ld, %ld\n", a.x, a.y, b.x, b.y );
			    }
                        }
                    }
                }
            else if ( ! success )
                {
                nearness = nearness / 2.0;
		run = 0.75 * (maxy - miny);
                midy = miny + (int) round( run );
                a.x = minx;
                a.y = midy;
                b.x = maxx;
                b.y = midy;
                traps = cutline( traps, p, a, b, d, TRUE, &success );  
                nearness = nearness * 2.0;
                if ( ! success )
                    { 
                    printf( "\nhorizontal quartering slice unsuccessful.\n" );
                    printf( "This is not a narrow shape, so that's especially bad.\n" );
                    p->active = FALSE;
                    v = vertex( p, 0 );
                    printf( "Look in the gds file at %ld, %ld \n", v.x, v.y );
                    c.x = (ur.x + ll.x) / 2;
                    c.y = (ur.y + ll.y) / 2;
                    printf( "or in the gpf file at   %ld, %ld \n", (v.x-c.x), (v.y-c.y) );
                    print_polygon( p, 1000 );
                    }
                }
            }
        }

    nearness = oldnearness;
    presque = nearness;

    }

//--------end quarter-------------------------------------------------------------------------------------------------

void
cut_long_thing( polygon *p, queue *traps, double dbu_um )
    {
    // Just like 'quarter' but more gentle. The intention is to
    // cut in half anything with a large aspect ratio.
    
    static int         minx, maxx, miny, maxy, dx, dy, midx, midy, iter;
    static double      ar;
    static xy          a, b, d;
    static BOOL        success, fatx, faty;

    if ( ! p->active ) return;
    if ( shape_extents_unknown( p ) ) set_shape_extents( p );

    d.x = 0.0;
    d.y = 0.0;

    extent( p, &minx, &maxx, &miny, &maxy );

    if ( p->len < 3 )
        {
        p->active = FALSE;
        return;
        }

    if ( fabs(maxx - minx) < nearness || fabs(maxy - miny) < nearness )
        {
	//printf( "\nthis shape is too thin\n" );
        p->active = FALSE;
        return;
        }
  
    midx = (maxx + minx) / 2;
    midy = (maxy + miny) / 2;

    dx = (maxx - minx) / 5; 
    dy = (maxy - miny) / 5;
    if ( dx == 0 ) dx = 2;
    if ( dy == 0 ) dy = 2;

    ar = fabs( maxx - minx ) / fabs( maxy - miny );

    fatx = ar > 5.0;
    faty = ar < 0.2;

    minx -= dx;
    maxx += dx;
    miny -= dy;
    maxy += dy;    

    success = TRUE;

    if ( fatx )
        {
        a.x = midx;
        a.y = miny;
        b.x = midx;
        b.y = maxy;
        iter = 0;
        if ( p->len == 3 ) 
            cut_triangular_sliver( p, &success, &iter, dbu_um );
	else
            traps = cutline( traps, p, a, b, d, TRUE, &success );   // true means throw away degenerates
        }
    else if ( faty )
        {
        a.x = minx;
        a.y = midy;
        b.x = maxx;
        b.y = midy;
       
        iter = 0;
        if ( p->len == 3 ) 
            cut_triangular_sliver( p, &success, &iter, dbu_um );
	else
            traps = cutline( traps, p, a, b, d, TRUE, &success );   
        }
    }


//-------------end cut_long_thing---------------------------------------------------------------------------------------------

void 
write_trap_list( int fdredo, queue *traps, BOOL justone, 
                 int *nout, xy d, BOOL verbose, xy ll, xy ur, xy block, double dbu_um, BOOL noarray ) 
    {

    static polygon *p;
    static polygon_ref *pr, *traplist;
    static int ilayer, idt, ilen;
    static struct gds_itemtype *gdspoly;
    static BOOL init=FALSE;
    static BOOL obese, obese_arr;
    static int maxlen=0;
    static int n;


    traplist = traps->first;
    pr = traplist;

    if ( ! init )
        {
        gdspoly = (struct gds_itemtype *) malloc( sizeof( struct gds_itemtype ) );
        init = TRUE;
        }
    
    n = 0;
    *nout = 0;

    while ( pr )
        {
        p = pr->p;

        if ( p->active && p->len != 0 )
            {
	    // iter = 0;
            // if ( p->len == 3 ) cut_triangular_sliver( p, &success, &iter, dbu_um );

            obese     = obese_shape( p, d, block, noarray );
            obese_arr = obese_array( p, d, block );

            if ( obese ) 
                {
                quarter( p, traps, ll, ur, d, dbu_um );
                obese = FALSE;
                }
            else if ( obese_arr )
                {
                quarter( p, traps, ll, ur, block, dbu_um );
                obese = FALSE;
                }
                        
            if ( p->active )                                    // shitty cropping
                p->active = p->ll.x  >=  ll.x && 
                            p->ll.y  >=  ll.y && 
                            p->ur.x  <=  ur.x && 
                            p->ur.y  <=  ur.y;
            }
        pr = pr->next;
        }


    //-------------------

    pr = traplist;
    while ( pr )
        {
        p = pr->p;

        if ( p->active && p->len != 0 )
            {
            ilen = p->len;
            if ( ilen > maxlen )
                {
                gdspoly->x = (int *) malloc( ilen * sizeof( int ) ); 
                gdspoly->y = (int *) malloc( ilen * sizeof( int ) );
                maxlen = ilen;
                }
            gdspoly->n = ilen; 

            copy_to_gds_polygon( p, gdspoly );

            ilayer = p->layer;
            idt    = p->dt;

            if ( p->active )    // it might be inactive if we had to quarter the shape
                {
                n++;
                (*nout)++;
                gds_write_boundary( fdredo );
                gds_write_layer(    fdredo, ilayer );
                gds_write_datatype( fdredo, idt    );
                if ( obese ) gds_write_obese( fdredo );
                gds_write_xy(       fdredo, gdspoly->x, gdspoly->y, gdspoly->n );
                gds_write_endel(    fdredo );
                num_out++;      
                }

            }
        pr = pr->next;
        if ( justone ) pr = NULL;
        }

    printv( "\n**************************** write_trap_list wrote %d shapes\n", n );

    // note that this function does not close the structure 

    } // write_trap_list

//--------------------------------------------------------------------------------------------------

int
pardonne( double x1, double x2 )   // je suis desole 
    {    
    static long x;

    if ( x1 < 0 && x2 < 0 ) bail( "negative field index, slice/pardonne" );
    if ( x1 < 0 ) return( floor( x2 ) );
    if ( x2 < 0 ) return( floor( x1 ) );
    
    if ( frac( x2 ) < 1e-6 )
        {
        x = floor( x1 );
        }
    else if ( frac( x1 ) < 1e-6 )
        {
        x = floor( x2 );
        }
    else
        {
        if ( frac(x2) > (1.0 - frac(x1)) )
            {
            x = floor( x2 );
            }
        else
            {
            x = floor( x1 );
	    }
        }
    return x;

    }

//---------

array_2dq *
field_sort( queue *shape, xy ll_mfr, xy ur_mfr, xy block_mfr, int ny, int nx )     
    {
    int            x, y, x2, y2;
    array_2dq     *field;
    polygon       *s;
    polygon_ref   *pr;
    double         d, b, fx, fy, fx2, fy2;
    queue         *fieldq;
    BOOL           outofbounds;

    if ( ! shape ) bail( "field sort: null shape pointer" );

    if ( ll_mfr.x >= ur_mfr.x || ll_mfr.y >= ur_mfr.y ) 
	{
        printf( "\n\nLL = %ld, %ld\n", ll_mfr.x, ll_mfr.y );
	printf( "UR = %ld, %ld\n", ur_mfr.x, ur_mfr.y );
	bail( "field_sort: invalid bounds" );
        field = initialize_2dq( ny, nx );
        return( field );
	}

    field = initialize_2dq( ny, nx );

    pr = shape->first;

    if ( ! pr ) bail( "field sort: null pointer at top" );

    while ( pr )
        {
        s = pr->p;
        if ( shape_extents_unknown( s ) ) set_shape_extents( s );

	x  = (s->ll.x - ll_mfr.x) / block_mfr.x;
	y  = (s->ll.y - ll_mfr.y) / block_mfr.y;
        x2 = (s->ur.x - 1 - ll_mfr.x) / block_mfr.x;
        y2 = (s->ur.y - 1 - ll_mfr.y) / block_mfr.y;

        outofbounds = x < 0 || x2 < 0 || y < 0 || y2 < 0 || x >= nx || x2 >= nx || y >= ny || y2 >= ny;

        if ( x2 == x && y2 == y && ! outofbounds )
            {
            fieldq = get_field_list_db( field, x, y, 1 );
            append_queue( fieldq, s );
            }
        else if ( ! outofbounds )        // a shape spans two fields
            {
            d   = (s->ll.x - ll_mfr.x);
            b   = block_mfr.x; 
            fx  = d / b;
            d   = (s->ll.y - ll_mfr.y);
            b   = block_mfr.y;
            fy  = d / b;
            d   = (s->ur.x - 1 - ll_mfr.x);
            b   = block_mfr.x;
            fx2 =  d / b;
            d   = (s->ur.y - 1 - ll_mfr.y);
            b   = block_mfr.y;
            fy2 = d / b;
            x = pardonne( fx, fx2 );
            y = pardonne( fy, fy2 );
            fieldq = get_field_list_db( field, x, y, 2 ); 
            append_queue( fieldq, s );
            }

        pr = pr->next;
        }

    return field;

    }  // field_sort

//--------------------------------------------------------------------------------------------------

void
extent_filter( queue *traps, xy ll, xy ur )
    {
    polygon *p;
    polygon_ref *pr;

    pr = traps->first;
    while ( pr )
        {
        p = pr->p;
        if ( p->ll.x < ll.x || p->ll.y < ll.y || p->ur.x > ur.x || p->ur.y > ur.y )
            {
            p->active = FALSE;                  // not emphatic enough. ok this time with feeling.
            delete_queue_member( traps, pr );   // should not be necessary but someone is not paying attention
            }
        pr = pr->next;
        }
    }

//==================================================================================================

int main( int argc, char *argv[] )
    {
    char    
            *finalname,
             sn[10]="",
            *sn1, 
            *sn2;

    xy       ll, 
             ur,  
             b,   
             d,   
	     e,
             bll, bur,
             llde, urde,
             cropll,
	     cropur;

    double
	     dx, dy, 
             bss_nm,
             bss_um,
             dbu_um,
             dbu_uu,
             dbu_m,
             llx, lly,
             urx, ury,
             bx, by;


    BOOL     abort,
             first=TRUE,
             not_eof,
             got_one,
             autoextent,
             noheader,
             noarray,
             crop;

    int      ifile,
             mofile,
             i, 
             ch1, ch2,
             nout=0;

    long     dt,    
             layer,
             len,
             numpolys,
             r, c,
             nrows,
             ncols;

    queue    
             *thisblock,
             *traps;

    short    count,           
             token;           

    struct gds_itemtype 
             *item;

    polygon  *poly;

    FILE    *extents;

    array_2dq
             *field;


    if ( argc == 2 && strcmp( argv[1], "-v" ) == 0 )
        {
        printf( "slice v%1.1f\n", VERSION );
        exit(0);
        }


    
    pid = getpid();

    if ( argc < 12 )
        {
        printf( "\nslice  version %1.1f\n", VERSION );
        printf( "\nUse: slice infile.gds outfile.gds LLx LLy URx URy xBlock yBlock \n" );
        printf(   "     subfield_size_x subfield_size_y beam_step_size \n" );
        printf(   "     [crop_llx] [crop_lly] [crop_urx] [crom_ury] [noheader] [noarray] \n\n" );
        printf(   "     All units are microns except for the beam step in nm.\n" );
        printf(   "     If the LL and UR coordinates are all zero, and extents.dat does not exist,\n" );
        printf(   "     then chip extents will be calculated automatically. The cropping window \n" );
        printf(   "     must be inside the pattern extent window. The field size specified in \n" );
        printf(   "     gpfout will be a multiple of the subfield size. Use 'noheader' when files  \n" );
        printf(   "     must be later concatenated. Use 'noarray' if the pattern will later be \n" );
        printf(   "     run through overlap removal.\n" );
        exit(0);
        }

    ifilename = argv[1];
    finalname = argv[2];

    sscanf( argv[3],  "%lf", &llx );
    sscanf( argv[4],  "%lf", &lly );
    sscanf( argv[5],  "%lf", &urx );
    sscanf( argv[6],  "%lf", &ury );
    sscanf( argv[7],  "%lf", &bx ); 
    sscanf( argv[8],  "%lf", &by ); 
    sscanf( argv[9],  "%lf", &dx ); 
    sscanf( argv[10], "%lf", &dy ); 
    sscanf( argv[11], "%lf", &bss_nm );

    double cropllx, croplly, cropurx, cropury;

    crop = FALSE;

    if ( argc > 15 )
        {
        sscanf( argv[12], "%lf", &cropllx );  // in um
        sscanf( argv[13], "%lf", &croplly );
        sscanf( argv[14], "%lf", &cropurx );
        sscanf( argv[15], "%lf", &cropury );
        }

    noheader = FALSE;

    if ( argc > 16 )
	noheader = (strcmp(argv[16], "noheader") == 0);

    if ( strstr( ifilename, "_part_" ) )
        {
        ch1 = (int) '_';
        sn1 = rindex( ifilename, ch1 );
        ch2 = (int) '.';
        sn2 = rindex( ifilename, ch2 ) - 1;
        if ( sn1 && sn2 )
            {
            strncpy( sn, sn1+1, sn2 - sn1 );
            strcat(  sn, " " );
            //printf( "starting slice part %s    \n", sn );  
            }
        }
    else
        {
        //printf( "slice %s %s\n", ifilename, finalname );
        }

    noarray = FALSE;
    if ( argc > 17 )
        noarray = (strcmp(argv[17], "noarray") == 0);

    bss_um = bss_nm / 1000.0;
    bss = bss_nm;               // no, we need this in gds database units!
    bssr = bss * BSSR;          // but don't worry, because we recalculate
    bssri = floor( bssr );      // these things below, after reading
    bssm = bss * BSSM;          // the database units

    // The big problem is long, narrow triangles. If we set nearness to 0.5 then
    // we get caught in an infinite loop when the sliver is persistent. There
    // should be a better way to prevent this looping failure. For now, we'll
    // set nearness to 0.6 and see how it goes.

    // We have to set nearness to be at least 0.8 so that horizontal and vertical
    // cuts will really look straight. But then that screws up other things?

    nearness = 0.6 * bss;  
    equality = 1.0 * bss;
    presque = nearness;
    tooclose = nearness / 2.0;

    bss2 = bss / BSSFACTOR;
    bsst2 = bss * 1.5;

    autoextent = fequal( llx, 0 ) && fequal( lly, 0 ) && fequal( urx, 0 ) && fequal( ury, 0 );

    abort = FALSE;

    if ( frac( dx*1000.0 / bss_nm ) > 1e-7 ) 
        {
        printf( "\nERROR: X subfield size is not a multiple of the beam step size." );
        abort = TRUE;
        }

    if ( frac( dy*1000.0 / bss_nm ) > 1e-7 ) 
        {
        printf( "\nERROR: Y subfield size is not a multiple of the beam step size." );
        abort = TRUE;
        }

    if ( frac( bx*1000.0 / bss_nm ) > 1e-7 )
        {
        printf( "\nERROR: X block size is not a multiple of the beam step size.\n" );
        printf(   "       bx = %lf um     frac: %lf\n", (bx), frac( bx*1000.0 / bss_nm ) );
        abort = TRUE;
        }

    if ( frac( by*1000.0 / bss_nm) > 1e-7 )
        {
        printf( "\nERROR: Y block size is not a multiple of the beam step size.\n" );
        printf(   "       by = %lf     frac: %lf\n", (by), frac( by*1000.0 / bss_nm) );
        abort = TRUE;
        }

    if ( abort ) bail( "\nToo bad. Try again.\n" ); 

    dt    = 0;
    layer = 0;

    numpolys    = 0;
    llde.x      = 0;
    llde.y      = 0;
    urde.x      = 0;
    urde.y      = 0;
    
    traps = initialize_queue();

    ifile = open( ifilename, 0 );
    if ( ifile <= 0 ) bail( "unable to open input file" );
    
    item = (struct gds_itemtype *) malloc( sizeof( struct gds_itemtype ) );
    if ( ! item ) bail( "unable to allocate gds item" );


    // Main polygon reading loop-----------------------------------
    //

    
    not_eof = read( ifile, &count, 2 ) > 0;

    while ( not_eof )
        {
        got_one = FALSE;
        
        while ( not_eof && ! got_one )
            {
            gds_swap2bytes( (BYTE *) &count );
            if ( count >= 4 )
                {
                read( ifile, &token, 2 );
                gds_swap2bytes( (BYTE *) &token );
            
                switch( token )
                    {
                    case 0x0305:  //printf( "UNITS \n" );  fflush(stdout);
                              gds_read_units( ifile, &dbu_um, &dbu_uu, &dbu_m, 0 );
                              bss = bss_um / dbu_um;    // eg 0.01 / 0.001
                              bssr = bss * BSSR;
                              bssm = bss * BSSM;
                              bss2 = bss / BSSFACTOR;

			      // now let's also add...

                              bssri = floor( bssr );  
                              nearness = 0.6 * bss;  
                              equality = 1.0 * bss;
                              presque = nearness;
                              tooclose = nearness / 2.0;


			      ll.x = (int) round( llx / dbu_um );
			      ll.y = (int) round( lly / dbu_um );
			      ur.x = (int) round( urx / dbu_um );
			      ur.y = (int) round( ury / dbu_um );
			      b.x  = (int) round( bx  / dbu_um );
			      b.y  = (int) round( by  / dbu_um );
			      d.x  = (int) round( dx / dbu_um );
			      d.y  = (int) round( dy / dbu_um );

			      cropll.x = (int) round( cropllx / dbu_um );
			      cropll.y = (int) round( croplly / dbu_um );
			      cropur.x = (int) round( cropurx / dbu_um );
			      cropur.y = (int) round( cropury / dbu_um );
			      snapoint_to( &cropll, 10 );
			      snapoint_to( &cropur, 10 );
			      crop = ! (cropll.x == 0 && cropll.y == 0 && cropur.x == 0 && cropur.y == 0);
                              
			      snapoint_to( &ll, 10 );
			      snapoint_to( &ur, 10 );
			      snapoint_to( &b,  10 );
			      snapoint_to( &d,  10 );

			      /*
                              printf( "lower left:  %1.4lf %1.4lf um = %ld %ld \n", llx, lly, ll.x, ll.y );
                              printf( "upper right: %1.4lf %1.4lf um = %ld %ld \n", urx, ury, ur.x, ur.y );
                              printf( "block:       %1.4lf %1.4lf um = %ld %ld \n", bx,  by,  b.x,  b.y  ); 
                              printf( "subfield:    %1.4lf %1.4lf um = %ld %ld \n", dx,  dy,  d.y,  d.y  ); 
                              printf( "crop:        %1.4lf %1.4lf to %1.2lf %1.2lf um\n", 
                                       cropllx, croplly, cropurx, cropury );
                              printf( "         =   %ld %ld to %ld %ld \n", cropll.x, cropll.y, cropur.x, cropur.y );
                              printf( "beam step:   %1.2lf nm\n", bss_nm );
                              if ( ! crop ) printf( "no cropping\n" );
			      bail( "stop for diagnostics" );
			      */

			      break;
                    case 0x0D02:  //printf( "LAYER \n" );  fflush(stdout);
                              gds_read_layer( ifile, count, &item, 0 );
                              layer = item->layer;
                              break;
                    case 0x0E02:  //printf( "DATATYPE \n" ); fflush(stdout);
                              gds_read_datatype( ifile, count, &item, 0 );
                              dt = item->dt;
                              break;
                    case 0x1003:  //printf( "XY \n" ) ; fflush(stdout);
                              numpolys++;
                              gds_read_xy( ifile, count, 0, &item, 0 );
                              item->layer = layer;
                              item->dt = dt;
                              got_one = TRUE;
                              len = item->n;
                              poly = initialize_polygon( len, layer, dt, FALSE );
                              copy_from_gds_polygon( item, poly );
                              remove_colinear_vertices( poly, FALSE );
                              set_shape_extents( poly );
                              //------------------------------------
                              append_queue( traps, poly );  
                              //------------------------------------
                              if ( first )
                                  {
                                  llde.x = poly->ll.x;  llde.y = poly->ll.y;
                                  urde.x = poly->ur.x;  urde.y = poly->ur.y;
                                  first = FALSE;
                                  }
                              else
                                  {
                                  if ( poly->ll.x < llde.x ) llde.x = poly->ll.x;
                                  if ( poly->ll.y < llde.y ) llde.y = poly->ll.y;
                                  if ( poly->ur.x > urde.x ) urde.x = poly->ur.x;
                                  if ( poly->ur.y > urde.y ) urde.y = poly->ur.y;
                                  }
                              
                              break;
                    default:      //
                              SKIPOVER( ifile, count );
                    }    
                }      
            not_eof = read( ifile, &count, 2 ) > 0;
            }          

        }  // while not eof

    close( ifile );

			      
    if ( autoextent )
        {
	extents = fopen( "tmp/extents.dat", "r" );
        if ( extents )
            {
            fscanf( extents,  "%lf %lf %lf %lf\n", &llx, &lly, &urx, &ury );
	    ll.x = (int) round( llx / dbu_um );
	    ll.y = (int) round( lly / dbu_um );
	    ur.x = (int) round( urx / dbu_um );
	    ur.y = (int) round( ury / dbu_um );
            snapoint_to( &ll, 10 );
	    snapoint_to( &ur, 10 );
            }
        else
            {
            ll.x = llde.x;  ll.y = llde.y;
            ur.x = urde.x;  ur.y = urde.y;
            e.x = abs( ur.x - ll.x );
            e.y = abs( ur.y - ll.y );
            ll.x = ( ll.x - e.x/10 ) - 1;
            ll.y = ( ll.y - e.y/10 ) - 1;
            ur.x = 1.0 + ( ur.x + e.x/10 );
            ur.y = 1.0 + ( ur.y + e.y/10 );
            if ( ur.x == ll.x ) ur.x = ll.x + 1;
            if ( ur.y == ll.y ) ur.y = ll.y + 1;
            snapoint_to( &ll, 10 );
	    snapoint_to( &ur, 10 );
            //printf(   "Lower left:     (%1d, %1d) um\n", ll.x, ll.y );
            //printf(   "Upper right:    (%1d, %1d) um\n\n", ur.x, ur.y );
            extents = fopen( "tmp/extents.dat", "w" );
            if ( ! extents ) bail( "unable to open tmp/extents.dat for output" );
            fprintf( extents, "%ld %ld %ld %ld\n", ll.x, ll.y, ur.x, ur.y );
            }
        fclose( extents );
        }

    extent_filter( traps, ll, ur );

    //printv( "\nMarry the vertices..." ); fflush( stdout );
    //marry_vertices( traps );  

    //---------- slice and dice -----------

    mofile = open( finalname, O_CREAT | O_TRUNC | O_WRONLY, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH );
    if ( mofile <= 0 ) bail( "unable to open output file" );

    if ( ! noheader )
	{
        gds_create_lib_correctly( mofile, "traps", dbu_um, dbu_uu );
        gds_write_bgnstr( mofile );
        gds_write_strname( mofile, "traps" );
	}

    if ( traps->len > 0 )
	{

        if ( crop )
            {
            //printf( "\nCropping window: (%d,%d) to (%d,%d)\n", cropll.x, cropll.y, cropur.x, cropur.y );
            if ( ll.x > cropll.x || ll.y > cropll.y || ur.x < cropur.x || ur.y < cropur.y )
                bail( "cropping window is outside of pattern extent" );
            }

        if ( crop ) crop_pattern( traps, cropll, cropur, ll, ur );

        ncols = 2 + abs(ur.x - ll.x)/b.x ;
        nrows = 2 + abs(ur.y - ll.y)/b.y ;

        printv( "ll: %ld, %ld   ur: %ld, %ld  block: %ld, %ld  rows: %d  cols: %d\n\n", ll.x, ll.y, ur.x, ur.y, b.x, b.y, nrows, ncols );

        printv( "Slicing %d shapes in %s ...\n", traps->len, ifilename ); 
        
        horizontal_cutup( traps, ll, ur, b, b, FALSE, FALSE, noarray );

        printv( "Dicing %d shapes in %s ...\n", traps->len, ifilename );

        vertical_cutup( traps, ll, ur, b, b, FALSE, FALSE, noarray );

        printv( "Sorting %d shapes in %s ... \n", traps->len, ifilename );

        field = field_sort( traps, ll, ur, b, nrows, ncols );     

        for ( c=0; c < ncols; c++ )
            {
            for ( r=0; r < nrows; r++ )
                {
                thisblock = get_field_list_db( field, c, r, 3 );

                bll.x = ll.x + c * b.x;
                bll.y = ll.y + r * b.y;
                bur.x = bll.x + 2 * b.x;  // bluh
                bur.y = bll.y + 2 * b.y;  // meh

                if ( thisblock->len > 0 ) 
                    {
                    printv( "row %d  col %d  [%d]\n", r, c, thisblock->len ); 
                    horizontal_cutup( thisblock, bll, bur, d, d, TRUE, TRUE, noarray );
                    vertical_cutup(   thisblock, bll, bur, d, d, TRUE, TRUE, noarray );
                    write_trap_list( mofile, thisblock, FALSE, &nout, d, FALSE, ll, ur, b, dbu_um, noarray );
                    }
                
                }
            }
        }

    // not a good idea to put in endlib if we are using cat later
    //
    //gds_write_endstr( mofile ); 
    //gds_write_endlib( mofile );    
    close( mofile );

    //printf( "Number of shapes written: %d\n", num_out );

    // and the script does this next:

    // poly2trap mofilename finalname bss_nm -1

    // and then guillotine, and then poly2trap again. sigh.

    //printf( "Slicing %s done.\n", ifilename  );

    }

